/*****************************************************************************/
/*
                                   Net.c

The networking essentials are based on DEC TCP/IP Services for OpenVMS (aka
UCX). The QIO interface was obviously chosen because of its asynchronous,
non-blocking capabilities. Although some functionalilty could have been
implemented using the BSD sockets abstraction and only the asynchronous I/O
done using the QIO I opted for working out the QIOs. It wasn't too difficult
and avoids the (slight) extra overhead of the sockets layer. The MadGoat
NETLIB port follows the basic outline of the DEC TCP/IP functionality. I
deliberately differentiated the UCX and NETLIB interfaces. It wasn't much
extra work and helps keep obvious the fact of the two APIs. NETLIB is a very
nice package! Wished I'd known about it three years ago. My thanks Matthew.

With resource wait explicitly enabled all sys$qio()s should wait until
resources become available. The only execption is ASTLM, a fundamental
requirement for this server. If SS$_EXQUOTA is returned from a sys$qio() (i.e.
without wait for completion, and therefore depending on AST delivery) then
EXIT THE SERVER with error status!

Some underlying NETLIB supporting TCP/IP packages do not support host-name
resolution at AST delivery level, which this module uses, and hence CANNOT
SUPPORT DNS LOOKUP from v4.4 onwards! (See the NETLIB documentation.)

The 'ServiceStruct' structure has two groups of fields and performs two basic
roles. First, it provides a linked list which can be traversed to determine
from the destination IP address of a request which host was specified in a
(multi-homed) request. Second, for each unique port specified for the same
server the client components of the structure are used for the accept()ing of
requests.  Note that there may be more 'service' structures in the list than
actual sockets created (and client components used) because only one is
allocated against a given port even though the services may specify multiple
host names using that port.

The server identity is derived from the 'ServerHostName' (gethostname()) name
and the 'ServerPort' (which is the first port specified as a service, or the
port specified in the configuration or /PORT= qualifier).


VERSION HISTORY
---------------
06-DEC-97  MGD  resolving a suspected inconsistent AST delivery situation
                by requiring all $QIO()s with an AST routine to ensure any
                queueing errors, etc. are reported via the AST routine
25-OCT-97  MGD  changes to MsgFor() function
06-SEP-97  MGD  multi-homed hosts and multi-port services
30-AUG-97  MGD  bugfix; get server host name before starting logging
                (woops, problem introduced with NETLIB)
09-AUG-97  MGD  message database
23-JUL-97  MGD  HTTPD v4.3, MultiNet dropped, using the NETLIB Library
01-FEB-97  MGD  HTTPd version 4
27-SEP-96  MGD  add dummy read for early error reporting
12-APR-96  MGD  observed Multinet disconnection/zero-byte behaviour
                (request now aborts if Multinet returns zero bytes)
03-JAN-96  MGD  support for both DEC TCP/IP Services and TGV MultiNet
01-DEC-95  MGD  NetWriteBuffered() for improving network I/O
20-DEC-94  MGD  multi-threaded version
20-JUN-94  MGD  single-threaded version
*/
/*****************************************************************************/

/* standard C header files */
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

/* VMS related header files */
#include <descrip.h>
#include <dvidef.h>
#include <iodef.h>
#include <psldef.h>
#include <ssdef.h>
#include <stsdef.h>

/* application-related header files */
#include "wasd.h"
#include "msg.h"
#include "net.h"
#include "error.h"

/******************/
/* global storage */
/******************/

int  QueueLength = 5,
     ServiceCount = 0,
     ServerHostNameLength;

char  ServerHostName [128],
      ServerHostPort [128],
      Services [256];

struct ServiceStruct  *ServiceListHead,
                      *ServiceListTail;

char  ErrorServiceList [] = "Service list confusing.";

#ifdef IP_UCX

$DESCRIPTOR (InetDeviceDsc, "UCX$DEVICE");

int  OptionEnabled = 1;

struct {
   int  l_onoff;
   int  l_linger;
} LingerSeconds = { 1, 10 };

struct {
   unsigned short  Length;
   unsigned short  Parameter;
   char  *Address;
} ReuseAddress =
   { sizeof(OptionEnabled), UCX$C_REUSEADDR, &OptionEnabled },
  ReuseAddressSocketOption =
   { sizeof(ReuseAddress), UCX$C_SOCKOPT, &ReuseAddress },
  LingerOnClose =
   { sizeof(LingerSeconds), UCX$C_LINGER, &LingerSeconds },
  LingerOnCloseOption =
   { sizeof(LingerOnClose), UCX$C_SOCKOPT, &LingerOnClose };

unsigned short  TcpSocket [2] =
   { UCX$C_TCP, INET_PROTYP$C_STREAM };

#endif /* IP_UCX */


#ifdef IP_NETLIB

unsigned long  NetLibAddressListSize = 16,
               NetLibAddressListCount,
               NetLibSocketType = NETLIB_K_TYPE_STREAM,
               NetLibSocketFamily = NETLIB_K_AF_INET,
               NetLibLevel = NETLIB_K_LEVEL_SOCKET,
               NetLibReUseOption = NETLIB_K_OPTION_REUSEADDR,
               NetLibShutdown = NETLIB_K_SHUTDOWN_RECEIVER,
               NetLibOptionOn = 1,
               NetLibSizeofINADDRDEF = sizeof(struct INADDRDEF);
               NetLibSizeofSINDEF = sizeof(struct SINDEF);
               NetLibSizeofOptionOn = sizeof(unsigned long);

#endif /* IP_NETLIB */


/********************/
/* external storage */
/********************/

#ifdef DBUG
extern boolean Debug;
#else
#define Debug 0 
#endif

extern boolean  ControlExitRequested;
extern boolean  ControlRestartRequested;
extern int  CurrentConnectCount;
extern int  MaxConcurrentConnections;
extern int  OutputBufferSize;
extern int  ServerPort;
extern char  CliLogFileName[];
extern char  ControlBuffer[];
extern char  ErrorSanityCheck[];
extern char  Utility[];
extern char  *AcceptHostsPtr;
extern char  *RejectHostsPtr;
extern struct AccountingStruct  Accounting;
extern struct ConfigStruct  Config;
extern struct MsgStruct  Msgs;

/*****************************************************************************/
/*
Get local host (server) name.
*/ 

NetGetServerHostName ()

{
   register char  *cptr;

   int  status;
   unsigned short  Length;

#ifdef IP_NETLIB
   $DESCRIPTOR (ServerHostNameDsc, ServerHostName);
#endif

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetGetServerHostName()\n");

#ifdef IP_UCX

   if (gethostname (ServerHostName, sizeof(ServerHostName)) != 0)
      ErrorExitVmsStatus (0, strerror(errno), FI_LI);

#endif

#ifdef IP_NETLIB

   status = netlib_get_hostname (&ServerHostNameDsc, &Length);
   if (Debug) fprintf (stdout, "netlib_get_hostname() %%X%08.08X\n", status);
   if (VMSnok (status))
      ErrorExitVmsStatus (status, "netlib_get_hostname()", FI_LI);
   ServerHostName[Length] = '\0';

#endif

   /* looks better if its all in lower case (host name is sometimes upper) */
   for (cptr = ServerHostName; *cptr; cptr++) *cptr = tolower(*cptr);
   ServerHostNameLength = cptr - ServerHostName;
   if (Debug) fprintf (stdout, "|%s|\n", ServerHostName);
}

/*****************************************************************************/
/*
Parse the 'Service's parameter to detmine which host(s)/port(s) services need
to be provided for.
*/ 

NetGetService ()

{
   register char  *cptr, *sptr;
   register struct ServiceStruct *svptr;

   boolean  SetServerPort;
   int  status,
        PortNumber;
   char  HostName [128];

#ifdef IP_UCX
   struct in_addr  ServerAddress;
#endif

#ifdef IP_NETLIB
   unsigned short  Length;
   unsigned long  NetLibSocket;
   struct  AnIOsb  IOsb;
   $DESCRIPTOR (HostNameDsc, HostName);
   $DESCRIPTOR (InternetAddressDsc, "");
#endif

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetGetService()\n");

   SetServerPort = false;

   if (Services[0])
      cptr = Services;
   else
   if (Config.ServicePtr == NULL)
      sprintf (cptr = Services, "%s:%d", ServerHostName, ServerPort);
   else
      cptr = Config.ServicePtr;

   while (*cptr)
   {
      while (*cptr && (ISLWS(*cptr) || *cptr == ',')) cptr++;
      if (!*cptr) break;

      HostName[0] = '\0';
      PortNumber = 0;

      if (Debug) fprintf (stdout, "|%s|\n", cptr);
      if (isalpha(*cptr))
      {
         sptr = HostName;
         while (*cptr && !ISLWS(*cptr) && *cptr != ',' && *cptr != ':')
            *sptr++ = *cptr++;
         *sptr = '\0';
         if (*cptr == ':')
         {
            cptr++;
            PortNumber = atol(cptr);
            while (*cptr && isdigit(*cptr)) cptr++;
         }
      }
      else
      if (isdigit(*cptr))
      {
         PortNumber = atol(cptr);
         while (*cptr && isdigit(*cptr)) cptr++;
      }
      else
         ErrorExitVmsStatus (status, ErrorServiceList, FI_LI);

      if (!HostName[0]) strcpy (HostName, ServerHostName);
      if (!PortNumber) PortNumber = 80;
      if (Debug) fprintf (stdout, "|%s:%d|\n", HostName, PortNumber);

      /* allocate memory for a new service structure */
      if ((svptr = calloc (1, sizeof (struct ServiceStruct))) == NULL)
         ErrorExitVmsStatus (status, "calloc()", FI_LI);
      if (Debug) fprintf (stdout, "svptr: %d\n", svptr);

      /* add it to the linked list */
      if (ServiceListHead == NULL)
         ServiceListHead = ServiceListTail = svptr;
      else
      {
         ServiceListTail->NextPtr = svptr;
         ServiceListTail = svptr;
      }
      svptr->NextPtr = NULL;
      ServiceCount++;

      svptr->ConnectCount = 0;

#ifdef IP_UCX

      if ((svptr->HostEntryPtr = gethostbyname (HostName)) == NULL)
      {
         fprintf (stdout, "%%%s-I-SERVICE, %s:%d\n",
                  Utility, HostName, PortNumber);
         ErrorExitVmsStatus (0, strerror(errno), FI_LI);
      }

      strcpy (svptr->ServerHostName, svptr->HostEntryPtr->h_name);
      memcpy (&svptr->ServerInternetAddress32bit,
              svptr->HostEntryPtr->h_addr, 4);
      memcpy (&ServerAddress, svptr->HostEntryPtr->h_addr, 4);
      strcpy (svptr->ServerInternetAddress, inet_ntoa (ServerAddress));

#endif

#ifdef IP_NETLIB

      /* NETLIB needs a socket to do these, create a temporary one */
      if (Debug) fprintf (stdout, "netlib_socket()\n");
      status = netlib_socket (&NetLibSocket, &NetLibSocketType,
                              &NetLibSocketFamily);
      if (Debug) fprintf (stdout, "netlib_socket() %%X%08.08X\n", status);
      if (VMSnok (status))
         ErrorExitVmsStatus (status, "netlib_socket()", FI_LI);

      /* turn the name into a binary address */
      HostNameDsc.dsc$a_pointer = HostName;
      HostNameDsc.dsc$w_length = strlen(HostName);

      status = netlib_name_to_address (&NetLibSocket, 0, &HostNameDsc,
                                       &svptr->NetLibAddressList,
                                       &NetLibAddressListSize,
                                       &NetLibAddressListCount,
                                       &IOsb, 0, 0);
      if (Debug)
         fprintf (stdout,
            "netlib_name_to_address() %%X%08.08X IOsb %%X%08.08X\n",
            status, IOsb.Status);
      if (VMSok (status) && VMSnok (IOsb.Status))
         status = IOsb.Status;
      if (VMSnok (status))
      {
         fprintf (stdout, "%%%s-I-SERVICE, %s:%d\n",
                  Utility, HostName, PortNumber);
         ErrorExitVmsStatus (status, "netlib_name_to_address()", FI_LI);
      }

      svptr->ServerInternetAddress32bit =
         svptr->NetLibAddressList[0].inaddr_l_addr;

      /* convert the binary address into a string */
      InternetAddressDsc.dsc$a_pointer = svptr->ServerInternetAddress;
      InternetAddressDsc.dsc$w_length = sizeof(svptr->ServerInternetAddress)-1;
      status = netlib_addrtostr (&svptr->NetLibAddressList[0],
                                 &InternetAddressDsc, &Length);

      if (Debug) fprintf (stdout, "netlib_addrtostr() %%X%08.08X\n", status);
      if (VMSnok (status))
         ErrorExitVmsStatus (status, "netlib_addrtostr()", FI_LI);
      svptr->ServerInternetAddress[Length] = '\0';
      if (Debug) fprintf (stdout, "|%s|\n", svptr->ServerInternetAddress);

      /* transform abbreviated host name to full ("host" into "host.domain") */
      HostNameDsc.dsc$w_length = sizeof(svptr->ServerHostName)-1;
      HostNameDsc.dsc$a_pointer = &svptr->ServerHostName;

      Length = 0;
      status = netlib_address_to_name (&NetLibSocket, 0,
                  &svptr->NetLibAddressList[0].inaddr_l_addr,
                  &NetLibSizeofINADDRDEF, &HostNameDsc, &Length, &IOsb, 0, 0);
      if (Debug)
         fprintf (stdout,
            "netlib_address_to_name() %%X%08.08X IOsb %%X%08.08X\n",
            status, IOsb.Status);
      svptr->ServerHostName[Length] = '\0';

      /* get rid of the temporary socket */
      status = netlib_close (&NetLibSocket);
      if (Debug) fprintf (stdout, "netlib_close() %%X%08.08X\n", status);

#endif

      svptr->ServerPort = PortNumber;
      sprintf (svptr->ServerPortString, "%d", svptr->ServerPort);

      /* looks better if its all in lower case (sometimes upper) */
      for (sptr = svptr->ServerHostName; *sptr; sptr++) *sptr = tolower(*sptr);

      sprintf (svptr->ServerHostPort, "%s:%d",
               svptr->ServerHostName, svptr->ServerPort);
      svptr->ServerHostNameLength = strlen(svptr->ServerHostName);

      if (Debug)
         fprintf (stdout, "%08.08X|%s|%s|%d|%s|\n",
                  svptr->ServerInternetAddress32bit,
                  svptr->ServerInternetAddress, svptr->ServerHostName,
                  svptr->ServerPort, svptr->ServerHostPort);

      /* process name, etc., generated from the primary port */
      if (!SetServerPort) SetServerPort = ServerPort = svptr->ServerPort;
   }

   if (!ServiceCount)
      ErrorExitVmsStatus (0, "No service specified!", FI_LI);

   /* the host and port for the "official" server */
   sprintf (ServerHostPort, "%s:%d", ServerHostName, ServerPort);
   if (Debug) fprintf (stdout, "|%s|\n", ServerHostPort);
}

/*****************************************************************************/
/*
It is only necessary to create one socket for each port port on the same
server, even if that port is specified against a different (multi-home) host
name because the accept()ing socket can be interrogated to determine which
host had been specified.  Then in a infinite loop accept connections from
clients.
*/ 

NetCreateService ()

{
   register char  *cptr, *sptr;
   register struct ServiceStruct  *svptr, *tsvptr;

   int  status,
        PortNumber;
   unsigned short  Length;
   
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetCreateService()\n");

   /* disable AST's until we've created the listening sockets */
   sys$setast (0);

   /* need SYSPRV to bind to a well-known port */
   EnableSysPrv ();

   for (svptr = ServiceListHead; svptr != NULL; svptr = svptr->NextPtr)
   {
      fprintf (stdout, "%%%s-I-SERVICE, %s\n", Utility, svptr->ServerHostPort);

      /**************************************/
      /* check for previous channel on port */
      /**************************************/

      for (tsvptr = ServiceListHead; tsvptr != NULL; tsvptr = tsvptr->NextPtr)
      {
         /* skip over this entry */
         if (svptr == tsvptr)
         {
            /* got this far in the list, can't be a socket for this port! */
            tsvptr = NULL;
            break;
         }
         /* break if already a socket created for this port */
         if (svptr->ServerPort == tsvptr->ServerPort) break;
      }
      /* if another host already on the same port then just continue */
      if (tsvptr != NULL) continue;

      /*************************/
      /* create service socket */
      /*************************/

#ifdef IP_UCX

      /* assign a channel to the internet template device */
      if (VMSnok (status =
          sys$assign (&InetDeviceDsc, &svptr->ServerChannel, 0, 0)))
         ErrorExitVmsStatus (status, "sys$assign()", FI_LI);

      /* make the channel a TCP, connection-oriented socket */

      if (Debug) fprintf (stdout, "sys$qiow() IO$_SETMODE\n");
      status = sys$qiow (0, svptr->ServerChannel, IO$_SETMODE,
                         &svptr->ServerIOsb, 0, 0,
                         &TcpSocket, 0, 0, 0, &ReuseAddressSocketOption, 0);

      if (Debug)
         fprintf (stdout, "sys$qiow() %%X%08.08X IOsb %%X%08.08X\n",
                  status, svptr->ServerIOsb.Status);
      if (VMSok (status) && VMSnok (svptr->ServerIOsb.Status))
         status = svptr->ServerIOsb.Status;
      if (VMSnok (status))
         ErrorExitVmsStatus (status, "sys$qiow()", FI_LI);

#endif

#ifdef IP_NETLIB

      if (Debug) fprintf (stdout, "netlib_socket()\n");
      status = netlib_socket (&svptr->ServerNetLibSocket, &NetLibSocketType,
                              &NetLibSocketFamily);
      if (Debug) fprintf (stdout, "netlib_socket() %%X%08.08X\n", status);
      if (VMSnok (status))
         ErrorExitVmsStatus (status, "netlib_socket()", FI_LI);

      status = netlib_setsockopt (&svptr->ServerNetLibSocket, &NetLibLevel,
                                  &NetLibReUseOption,
                                  &NetLibOptionOn, &NetLibSizeofOptionOn,
                                  &svptr->ServerIOsb, 0, 0);
      if (Debug)
         fprintf (stdout, "netlib_setsockopt() %%X%08.08X IOsb %%X%08.08X\n",
                  status, svptr->ServerIOsb.Status);
      if (VMSok (status) && VMSnok (svptr->ServerIOsb.Status))
         status = svptr->ServerIOsb.Status;
      if (VMSnok (status))
         ErrorExitVmsStatus (status, "netlib_setsockopt()", FI_LI);

#endif

      /**************************/
      /* bind the server socket */
      /**************************/

#ifdef IP_UCX

      svptr->ServerSocketNameItem.Length = sizeof(svptr->ServerSocketName);
      svptr->ServerSocketNameItem.Address = &svptr->ServerSocketName;

      svptr->ServerSocketName.sin_family = AF_INET;
      svptr->ServerSocketName.sin_port = htons (svptr->ServerPort);
      svptr->ServerSocketName.sin_addr.s_addr = INADDR_ANY;

      if (Debug) fprintf (stdout, "sys$qiow() IO$_SETMODE\n");
      status = sys$qiow (0, svptr->ServerChannel, IO$_SETMODE,
                         &svptr->ServerIOsb, 0, 0,
                         0, 0, &svptr->ServerSocketNameItem, QueueLength, 0, 0);
      if (Debug)
         fprintf (stdout, "sys$qiow() %%X%08.08X IOsb %%X%08.08X\n",
                  status, svptr->ServerIOsb.Status);
      if (VMSok (status) && VMSnok (svptr->ServerIOsb.Status))
         status = svptr->ServerIOsb.Status;
      if (VMSnok (status))
         ErrorExitVmsStatus (status, "sys$qiow()", FI_LI);

#endif

#ifdef IP_NETLIB

      svptr->ServerNetLibName.sin_w_family = NETLIB_K_AF_INET;
      svptr->ServerNetLibName.sin_w_port =
         netlib_hton_word (&svptr->ServerPort);
      svptr->ServerNetLibName.sin_x_addr.inaddr_l_addr = 0;  /* INADDR_ANY */

      if (Debug) fprintf (stdout, "netlib_bind()\n");
      status = netlib_bind (&svptr->ServerNetLibSocket,
                            &svptr->ServerNetLibName,
                            &NetLibSizeofSINDEF, &svptr->ServerIOsb, 0, 0);
      if (Debug)
         fprintf (stdout, "netlib_bind %%X%08.08X IOsb %%X%08.08X\n",
                  status, svptr->ServerIOsb.Status);
      if (VMSok (status) && VMSnok (svptr->ServerIOsb.Status))
         status = svptr->ServerIOsb.Status;
      if (VMSnok (status))
         ErrorExitVmsStatus (status, "netlib_bind()", FI_LI);

      if (Debug) fprintf (stdout, "netlib_listen()\n");
      status = netlib_listen (&svptr->ServerNetLibSocket, &QueueLength,
                              &svptr->ServerIOsb, 0, 0);
      if (Debug)
         fprintf (stdout, "netlib_listen %%X%08.08X IOsb %%X%08.08X\n",
                  status, svptr->ServerIOsb.Status);
      if (VMSok (status) && VMSnok (svptr->ServerIOsb.Status))
         status = svptr->ServerIOsb.Status;
      if (VMSnok (status))
         ErrorExitVmsStatus (status, "netlib_listen()", FI_LI);

#endif

      /****************************/
      /* listen for first request */
      /****************************/

      NetAccept (svptr);
   }

   /************************/
   /* end of services list */
   /************************/

   /* turn off SYSPRV after binding to well-known port */
   DisableSysPrv ();

   /* reenable user-mode ASTs now we've set up our listening sockets */
   sys$setast (1);

   sys$hiber ();
}

/*****************************************************************************/
/*
Queue an accept() to the listening server socket.
*/ 

NetAccept (struct ServiceStruct *svptr)

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetAccept() %d\n", svptr);

#ifdef IP_UCX

   /* assign a channel to the internet template device */
   if (VMSnok (status =
       sys$assign (&InetDeviceDsc, &svptr->ClientChannel, 0, 0)))
      ErrorExitVmsStatus (status, "sys$assign()", FI_LI);

   svptr->ClientSocketNameItem.Length = sizeof(svptr->ClientSocketName);
   svptr->ClientSocketNameItem.Address = &svptr->ClientSocketName;
   svptr->ClientSocketNameItem.LengthPtr = &svptr->ClientSocketNameLength;

   if (Debug) fprintf (stdout, "sys$qio() IO$_ACCESS | IO$M_ACCEPT\n");
   status = sys$qio (0, svptr->ServerChannel, IO$_ACCESS | IO$M_ACCEPT,
                     &svptr->ServerIOsb, &NetAcceptAst, svptr,
                     0, 0, &svptr->ClientSocketNameItem, &svptr->ClientChannel,
                     &LingerOnCloseOption, 0);
   if (Debug) fprintf (stdout, "sys$qio() %%X%08.08X\n", status);
   if (VMSnok (status)) 
      ErrorExitVmsStatus (status, "sys$qio()", FI_LI);

#endif

#ifdef IP_NETLIB

   if (Debug) fprintf (stdout, "netlib_accept()\n");
   status = netlib_accept (&svptr->ServerNetLibSocket,
                           &svptr->ClientNetLibSocket,
                           &svptr->ClientNetLibName, &NetLibSizeofSINDEF,
                           &svptr->NetLibReturnedLength,
                           &svptr->ServerIOsb, &NetAcceptAst, svptr);
   if (Debug)
      fprintf (stdout, "netlib_accept() %%X%08.08X IOsb %%X%08.08X\n",
               status, svptr->ServerIOsb.Status);
   if (VMSok (status) && VMSnok (svptr->ServerIOsb.Status))
      status = svptr->ServerIOsb.Status;
   if (VMSnok (status))
      ErrorExitVmsStatus (status, "netlib_accept()", FI_LI);

   /* option LINGER doesn't seem to be an option :^) ... perhaps default? */

#endif

}

/*****************************************************************************/
/*
A connection has been accept()ed on the specified server socket.
*/ 

NetAcceptAst (struct ServiceStruct *svptr)

{
   register struct RequestStruct  *rqptr;
   register struct ServiceStruct  *tsvptr;

   int  status,
        ServerSocketNameLength;
   unsigned short  PortNumber;
   char  ServerInternetAddress [32];
   struct AnIOsb  IOsb;

#ifdef IP_UCX
   int  SocketNameLength;
   struct sockaddr_in  SocketName;
   struct {
      unsigned int  Length;
      char  *Address;
      unsigned int  *LengthPtr;
   } SocketNameItem;
#endif

#ifdef IP_NETLIB
   static $DESCRIPTOR (InternetAddressDsc, "");
   unsigned short  Length;
   unsigned long  NetLibReturnedLength;
   struct SINDEF  NetLibName;
#endif

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetAcceptAst() %d\n", svptr);

#ifdef IP_UCX
   if (Debug)
      fprintf (stdout, "sys$qio() IOsb %%X%08.08X\n",
               svptr->ServerIOsb.Status);
#endif

#ifdef IP_NETLIB
   if (Debug)
      fprintf (stdout, "netlib_accept() IOsb %%X%08.08X\n",
               svptr->ServerIOsb.Status);
#endif
      
   if (svptr->ServerIOsb.Status == SS$_ABORT && 
       ControlExitRequested || ControlRestartRequested)
   {
      /* server exiting or restarting, just return */
      return;
   }

   if (VMSnok (svptr->ServerIOsb.Status)) 
   {
      /* most often network/system shutting down */
      ErrorExitVmsStatus (svptr->ServerIOsb.Status, "accept()", FI_LI);
   }

   Accounting.ConnectCount++;

#ifdef IP_UCX

   svptr->ClientInternetAddress32bit =
      svptr->ClientSocketName.sin_addr.S_un.S_addr;
   strcpy (svptr->ClientInternetAddress,
           inet_ntoa (svptr->ClientSocketName.sin_addr));
   if (Debug) fprintf (stdout, "|%s|\n", svptr->ClientInternetAddress);

   if (ServiceCount == 1)
   {
      /* "they can have any color they want, provided it's black" */
      tsvptr = ServiceListHead;
   }
   else
   {
      /************************************/
      /* multiple services, get host/port */
      /************************************/

      SocketNameItem.Length = sizeof(SocketName);
      SocketNameItem.Address = &SocketName;
      SocketNameItem.LengthPtr = &SocketNameLength;

      if (Debug) fprintf (stdout, "sys$qiow() IO$_SENSEMODE\n");
      status = sys$qiow (0, svptr->ClientChannel, IO$_SENSEMODE,
                         &IOsb, 0, 0, 0, 0, &SocketNameItem, 0, 0, 0);
      if (Debug)
         fprintf (stdout, "sys$qiow() %%X%08.08X IOsb %%X%08.08X\n",
                  status, IOsb.Status);
      if (VMSok (status) && VMSnok (IOsb.Status))
         status = IOsb.Status;
      if (VMSnok (status))
      {
         /* hmmm, forget it, ready for next connection */
         NetAccept (svptr);
         return;
      }

      PortNumber = ntohs(SocketName.sin_port);

      if (Debug)
         fprintf (stdout, "|%s|%d|\n",
                  inet_ntoa (SocketName.sin_addr), PortNumber);

      for (tsvptr = ServiceListHead; tsvptr != NULL; tsvptr = tsvptr->NextPtr)
      {
         if (SocketName.sin_addr.s_addr ==
             tsvptr->ServerInternetAddress32bit &&
             PortNumber == tsvptr->ServerPort) break;
      }
      /* if none matches then 'tsvptr' is set to NULL */
   }

#endif

#ifdef IP_NETLIB

   InternetAddressDsc.dsc$a_pointer = svptr->ClientInternetAddress;
   InternetAddressDsc.dsc$w_length = sizeof(svptr->ClientInternetAddress)-1;

   svptr->ClientInternetAddress32bit =
      svptr->ClientNetLibName.sin_x_addr.inaddr_l_addr;

   status = netlib_addrtostr (&svptr->ClientNetLibName.sin_x_addr.inaddr_l_addr,
                              &InternetAddressDsc, &Length);
   if (Debug) fprintf (stdout, "netlib_addrtostr() %%X%08.08X\n", status);
   if (VMSnok (status))
      ErrorExitVmsStatus (status, "netlib_addrtostr()", FI_LI);

   svptr->ClientInternetAddress[Length] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", svptr->ClientInternetAddress);

   if (ServiceCount == 1)
   {
      /* ditto */
      tsvptr = ServiceListHead;
   }
   else
   {
      /************************************/
      /* multiple services, get host/port */
      /************************************/

      if (Debug) fprintf (stdout, "netlib_getsockname()\n");
      status = netlib_getsockname (&svptr->ClientNetLibSocket,
                                   &NetLibName, &NetLibSizeofSINDEF,
                                   &NetLibReturnedLength, &IOsb, 0, 0);
      if (Debug)
         fprintf (stdout, "netlib_getsockname %%X%08.08X IOsb %%X%08.08X\n",
                  status, IOsb.Status);
      if (VMSok (status) && VMSnok (IOsb.Status))
         status = IOsb.Status;
      if (VMSnok (status))
      {
         /* hmmm, forget it, ready for next connection */
         NetAccept (svptr);
         return;
      }

      PortNumber = netlib_ntoh_word (&NetLibName.sin_w_port);

      if (Debug)
      {
         char  AddressString [32];
         unsigned short  Length;
         $DESCRIPTOR (AddressStringDsc, AddressString);

         netlib_addrtostr (&NetLibName.sin_x_addr, &AddressStringDsc, &Length);
         AddressString[Length] = '\0';
         fprintf (stdout, "|%s|%d|\n", AddressString, PortNumber);
      }

      for (tsvptr = ServiceListHead; tsvptr != NULL; tsvptr = tsvptr->NextPtr)
      {
         if (NetLibName.sin_x_addr.inaddr_l_addr ==
             tsvptr->ServerInternetAddress32bit &&
             PortNumber == tsvptr->ServerPort) break;
      }
      /* if none matches then 'tsvptr' is set to NULL */
   }

#endif

   if (Debug && tsvptr != NULL)
      fprintf (stdout, "|%s|%s:%d|\n",
               tsvptr->ServerInternetAddress,
               tsvptr->ServerHostName, tsvptr->ServerPort);

   if (tsvptr == NULL)
      svptr->ConnectCount++;
   else
      tsvptr->ConnectCount++;

   if (Config.DnsLookup)
      NetHostLookup (svptr);
   else
      strcpy (svptr->ClientHostName, svptr->ClientInternetAddress);
   if (Debug) fprintf (stdout, "|%s|\n", svptr->ClientHostName);

   if (CurrentConnectCount >= MaxConcurrentConnections)
   {
      /************/
      /* too busy */
      /************/

      Accounting.ConnectTooBusyCount++;
      Accounting.ResponseStatusCodeCountArray[5]++;

#ifdef IP_UCX
      NetTooBusy (svptr->ClientChannel,
                  svptr->ClientHostName,
                  svptr->ClientInternetAddress);
      NetDummyRead (svptr->ClientChannel);
#endif

#ifdef IP_NETLIB
      NetTooBusy (&svptr->ClientNetLibSocket,
                  svptr->ClientHostName,
                  svptr->ClientInternetAddress);
      NetDummyRead (&svptr->ClientNetLibSocket);
#endif

      /* ready for next connection */
      NetAccept (svptr);

      return;
   }

   /* 'tsvptr' is NULL if the request did not match a known service */
   if (tsvptr == NULL ||
       !ConfigAcceptClientHostName (svptr->ClientInternetAddress,
                                    svptr->ClientHostName))
   {
      /************/
      /* rejected */
      /************/

      Accounting.ConnectRejectedCount++;
      Accounting.ResponseStatusCodeCountArray[4]++;

#ifdef IP_UCX
      NetAccessDenied (svptr->ClientChannel,
                       svptr->ClientHostName,
                       svptr->ClientInternetAddress);
      NetDummyRead (svptr->ClientChannel);
#endif

#ifdef IP_NETLIB
      NetAccessDenied (&svptr->ClientNetLibSocket,
                       svptr->ClientHostName,
                       svptr->ClientInternetAddress);
      NetDummyRead (&svptr->ClientNetLibSocket);
#endif

      /* ready for next connection */
      NetAccept (svptr);

      return;
   }

   Accounting.ConnectAcceptedCount++;

   /* allocate zeroed dynamic memory for the connection thread */
   if ((rqptr = calloc (1, sizeof(struct RequestStruct))) == NULL)
   {
      /****************************/
      /* memory allocation failed */
      /****************************/

      Accounting.ConnectCallocFailCount++;
      Accounting.ResponseStatusCodeCountArray[5]++;

#ifdef IP_UCX
      NetRequestCallocFailed (svptr->ClientChannel,
                              svptr->ClientHostName,
                              svptr->ClientInternetAddress);
      NetDummyRead (svptr->ClientChannel);
#endif

#ifdef IP_NETLIB
      NetRequestCallocFailed (&svptr->ClientNetLibSocket,
                              svptr->ClientHostName,
                              svptr->ClientInternetAddress);
      NetDummyRead (&svptr->ClientNetLibSocket);
#endif

      /* ready for next connection */
      NetAccept (svptr);

      return;
   }

   /***************************/
   /* process HTTP connection */
   /***************************/

   /* once the thread structure has been created we have a client! */
   Accounting.ConnectCurrent = ++CurrentConnectCount;
   if (CurrentConnectCount > Accounting.ConnectPeak)
      Accounting.ConnectPeak = CurrentConnectCount;

   /* point at and copy some of the service's network information */
   rqptr->ServerInternetAddressPtr = tsvptr->ServerInternetAddress;
   rqptr->ServerHostNamePtr = tsvptr->ServerHostName;
   rqptr->ServerHostNameLength = tsvptr->ServerHostNameLength;
   rqptr->ServerPortStringPtr = tsvptr->ServerPortString;
   rqptr->ServerPort = tsvptr->ServerPort;
   rqptr->ServerHostPortPtr = tsvptr->ServerHostPort;

   /* copy the client's network information */
   rqptr->ClientInternetAddress32bit = svptr->ClientInternetAddress32bit;
   strcpy (rqptr->ClientInternetAddress, svptr->ClientInternetAddress);
   strcpy (rqptr->ClientHostName, svptr->ClientHostName);

#ifdef IP_UCX
   rqptr->ClientChannel = svptr->ClientChannel;
#endif

#ifdef IP_NETLIB
   rqptr->ClientNetLibSocket = svptr->ClientNetLibSocket;
#endif

   if (VMSnok (HttpdConnection (rqptr, true)))
      ConcludeProcessing (rqptr);

   /* ready for next connection */
   NetAccept (svptr);
}

/*****************************************************************************/
/*
Some browsers report a network error (e.g. Windows Netscape Navigator 2.n) if
the server does not read anything from them before returning an error message.
This routine and it's AST ensure a read is done before reporting an error
earlier that reading the request from it.  Needless-to-say, no status checking
is performed and the data just discarded (hope it works consistently :^)
*/ 

#ifdef IP_UCX

NetDummyRead (unsigned short Channel)

{
   static char  DummyBuffer [1024];
   static struct AnIOsb  DummyIOsb;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetDummyRead() Channel: %d\n", Channel);

   status = sys$qio (0, Channel,
                     IO$_READVBLK, &DummyIOsb,
                     &NetDummyReadAst, Channel,
                     DummyBuffer, sizeof(DummyBuffer), 0, 0, 0, 0);

   if (Debug) fprintf (stdout, "sys$qio() %%X%08.08X\n", status);

}
#endif

#ifdef IP_NETLIB

NetDummyRead (unsigned long *NetLibSocketPtr)

{
   static char  DummyBuffer [1024];
   static struct AnIOsb  DummyIOsb;
   static $DESCRIPTOR (DummyBufferDsc, DummyBuffer);

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "NetDummyRead() *NetLibSocketPtr: %d\n",
               *NetLibSocketPtr);

   status = netlib_read (NetLibSocketPtr, &DummyBufferDsc, 0, 0, 0, 0,
                         &DummyIOsb, &NetDummyReadAst, NetLibSocketPtr);

   if (Debug) fprintf (stdout, "netlib_read %%X%08.08X\n", status);
}
#endif

/*****************************************************************************/
/*
See commentary in NetDummyRead().
*/ 

#ifdef IP_UCX

NetDummyReadAst (unsigned short Channel)

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetDummyReadAst() Channel: %d\n", Channel);

   NetShutdownSocket (Channel);
}

#endif

#ifdef IP_NETLIB

NetDummyReadAst (unsigned long *NetLibSocketPtr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "NetDummyReadAst() *NetLibSocketPtr: %d\n",
               *NetLibSocketPtr);

   NetShutdownSocket (NetLibSocketPtr);
}

#endif

/*****************************************************************************/
/*
Perform the UCX get host name using QIO ACP functionality.  This can execute 
at AST-delivery level.  Doing this within NETLIB needs to cater for the
lowest-common-denominator package, which sometimes will not execute at AST
delivery level, therefore rendering some NETLIB packages incapable of
supporting DNS lookup!
*/ 

#ifdef IP_UCX

NetHostLookup (struct ServiceStruct *svptr) 

{
   static unsigned char ControlSubFunction [4] =
      { INETACP_FUNC$C_GETHOSTBYADDR, INETACP$C_TRANS, 0, 0 };
   static struct dsc$descriptor AddressDsc =
      { 0, DSC$K_DTYPE_T, DSC$K_CLASS_S, 0};
   static struct dsc$descriptor ControlSubFunctionDsc =
      { 4, DSC$K_DTYPE_T, DSC$K_CLASS_S, &ControlSubFunction };
   static struct dsc$descriptor ClientHostNameDsc =
      { 0, DSC$K_DTYPE_T, DSC$K_CLASS_S, 0};

   int  status;
   unsigned short  Length;
   struct  AnIOsb  IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetHostLookup()\n");

   AddressDsc.dsc$w_length = sizeof(svptr->ClientSocketName.sin_addr);
   AddressDsc.dsc$a_pointer = &svptr->ClientSocketName.sin_addr;

   ClientHostNameDsc.dsc$w_length = sizeof(svptr->ClientHostName)-1;
   ClientHostNameDsc.dsc$a_pointer = &svptr->ClientHostName;

   Length = 0;
   if (Debug) fprintf (stdout, "sys$qio() IO$_ACPCONTROL\n");
   status = sys$qiow (0, svptr->ServerChannel, IO$_ACPCONTROL,
                      &IOsb, 0, 0, &ControlSubFunctionDsc,
                      &AddressDsc, &Length, &ClientHostNameDsc, 0, 0);
   if (Debug)
      fprintf (stdout, "sys$qio() %%X%08.08X IOsb %%X%08.08X\n",
               status, IOsb.Status);
   if (Length)
      svptr->ClientHostName[Length] = '\0';
   else
      strcpy (svptr->ClientHostName, svptr->ClientInternetAddress);
}
#endif

#ifdef IP_NETLIB

NetHostLookup (struct ServiceStruct *svptr) 

{
   static struct dsc$descriptor ClientHostNameDsc =
      { 0, DSC$K_DTYPE_T, DSC$K_CLASS_S, 0};

   int  status;
   unsigned short  Length;
   struct AnIOsb  IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetHostLookup()\n");

   ClientHostNameDsc.dsc$w_length = sizeof(svptr->ClientHostName)-1;
   ClientHostNameDsc.dsc$a_pointer = &svptr->ClientHostName;

   Length = 0;
   status = netlib_address_to_name (&svptr->ClientNetLibSocket, 0,
               &svptr->ClientNetLibName.sin_x_addr.inaddr_l_addr,
               &NetLibSizeofINADDRDEF, &ClientHostNameDsc, &Length,
               &IOsb, 0, 0);
   if (Debug)
      fprintf (stdout, "netlib_address_to_name() %%X%08.08X IOsb %%X%08.08X\n",
               status, IOsb.Status);
   if (Length)
      svptr->ClientHostName[Length] = '\0';
   else
      strcpy (svptr->ClientHostName, svptr->ClientInternetAddress);
}
#endif

/****************************************************************************/
/*
*/

#ifdef IP_UCX

NetShutdownSocket (unsigned short Channel)

{
   int  status;
   struct  AnIOsb  IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetShutdownSocket() %d\n", Channel);

   status = sys$qiow (0, Channel, IO$_DEACCESS | IO$M_SHUTDOWN, &IOsb, 0, 0,
                      0, 0, 0, 0, 0, 0);
   if (Debug)
      fprintf (stdout, "sys$qiow() %%X%08.08X IOsb %%X%08.08X\n",
               status, IOsb.Status);

   status = sys$dassgn (Channel);
   if (Debug) fprintf (stdout, "sys$dassgn() %%X%08.08X\n", status);
}
#endif

#ifdef IP_NETLIB

NetShutdownSocket (unsigned long *NetLibSocketPtr)

{
   int  status;
   struct  AnIOsb  IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "NetShutdownSocket() *NetLibSocketPtr: %d\n",
               *NetLibSocketPtr);

   status = netlib_shutdown (NetLibSocketPtr, &NetLibShutdown, &IOsb, 0, 0);
   if (Debug)
      fprintf (stdout, "netlib_shutdown %%X%08.08X IOsb %%X%08.08X\n",
               status, IOsb.Status);

   status = netlib_close (NetLibSocketPtr);
   if (Debug) fprintf (stdout, "netlib_close() %%X%08.08X\n", status);
}
#endif

/****************************************************************************/
/*
Stop the server from receiving incoming requests.
*/

#ifdef IP_UCX

NetShutdownServerSocket ()

{
   register struct ServiceStruct  *svptr;

   int  status;
   struct  AnIOsb  IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetShutdownServerSocket()\n");

   for (svptr = ServiceListHead; svptr != NULL; svptr = svptr->NextPtr)
   {
      status = sys$qiow (0, svptr->ServerChannel, IO$_DEACCESS, &IOsb, 0, 0,
                         0, 0, 0, 0, 0, 0);

      if (Debug)
         fprintf (stdout, "sys$qiow() %%X%08.08X IOsb %%X%08.08X\n",
                  status, IOsb.Status);
   }
}
#endif

#ifdef IP_NETLIB

NetShutdownServerSocket ()

{
   register struct ServiceStruct  *svptr;

   int  status;
   struct  AnIOsb  IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetShutdownServerSocket()\n");

   for (svptr = ServiceListHead; svptr != NULL; svptr = svptr->NextPtr)
   {
      status = netlib_shutdown (&svptr->ServerNetLibSocket,
                                &NetLibShutdown, &IOsb, 0, 0);

      if (Debug)
         fprintf (stdout, "netlib_shutdown %%X%08.08X IOsb %%X%08.08X\n",
                  status, IOsb.Status);
   }
}
#endif

/*****************************************************************************/
/*
Unlike NetWrite() this function writes directly using a client channel/socket,
not using the full request data structure. This is for writing to the client
before any of the request data structures are created (e.g. when the server
denies access). This is a blocking function, but shouldn't introduce much
granularity as it is only used for short messages before a client thread is
set up.
*/ 

#ifdef IP_UCX

int NetDirectWrite
(
unsigned short Channel,
char *DataPtr,
int DataLength
)
{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetDirectWrite()\n");

   status = sys$qiow (0, Channel, IO$_WRITEVBLK, 0, 0, 0,
                      DataPtr, DataLength, 0, 0, 0, 0);

   if (Debug) fprintf (stdout, "sys$qio() %%X%08.08X\n", status);

   return (status);
}
#endif

#ifdef IP_NETLIB

int NetDirectWrite
(
unsigned long *NetLibSocketPtr,
char *DataPtr,
int DataLength
)
{
   static $DESCRIPTOR (DataPtrDsc, "");

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetDirectWrite()\n");

   DataPtrDsc.dsc$a_pointer = DataPtr;
   DataPtrDsc.dsc$w_length = DataLength;

   status = netlib_write (NetLibSocketPtr, &DataPtrDsc, 0, 0, 0, 0, 0);

   if (Debug) fprintf (stdout, "netlib_write() %%X%08.08X\n", status);

   return (status);
}
#endif

/*****************************************************************************/
/*
Called from NetWrite() is a response header needs to be sent before any data.
Response header has now been sent, send the data using the buffered
information about it.
*/ 

NetResponseHeaderAst (struct RequestStruct *rqptr)

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetResponseHeaderAst()\n");

   NetWrite (rqptr,
             rqptr->ResponseHeaderAstAddress,
             rqptr->ResponseHeaderDataPtr,
             rqptr->ResponseHeaderDataLength);
}

/*****************************************************************************/
/*
Write 'DataLength' bytes located at 'DataPtr' via the network to the client.
If 'AstFunctionPtr' zero then use sys$qiow(), waiting for completion. If
an AST completion address is supplied then use sys$qio().  If empty data
buffer is supplied (zero length) then declare an AST to service any AST
routine supplied.  If none then just return.

For responses generated by the server the HTTP header is a separate structure
which can be sent separately. If the HTTP method is "HEAD" only allow bytes in
the header to be sent, absorb any other, explicitly calling the AST completion
routine as necessary (this, in particular, is for scripts that don't recognise
this method).

An "Xray" returns a full response (header and body) as a plain text document
(allows diagnostics, etc.) With an Xray just output a hard-wired, plain-text
response with wait for completion (won't be happening enough for any
significant impact!)

Explicitly declares any AST routine if an error occurs. The calling function
must not do any error recovery if an AST routine has been supplied but the
associated AST routine must!  If an AST was not supplied then the return
status can be checked.
*/ 

int NetWrite
(
struct RequestStruct *rqptr,
void *AstFunctionPtr,
char *DataPtr,
int DataLength
)
{
   static char  XrayHeader [] =
"HTTP/1.0 200 Xray\r\n\
Content-Type: text/plain\r\n\
Expires: Thu, 01-Jan-1970 00:00:01 GMT\r\n\
\r\n";

#ifdef IP_NETLIB
   static $DESCRIPTOR (DataPtrDsc, "");
#endif

   int  status,
        DclAstStatus;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "NetWrite() %d %d %d %d bytes\n",
               rqptr, AstFunctionPtr, DataPtr, DataLength);
   /** if (Debug) fprintf (stdout, "|%*s|\n", DataLength, DataPtr); **/
if (Debug) fprintf (stdout, "|%*s|\n", DataLength, DataPtr);

   status = SS$_NORMAL;

   if (!rqptr->BytesTx)
   {
      /*******************************/
      /* nothing sent to client yet! */
      /*******************************/

      if (rqptr->Xray)
      {
         /*********************************/
         /* header for "Xray" of response */
         /*********************************/

#ifdef IP_UCX
         status = sys$qiow (0, rqptr->ClientChannel,
                            IO$_WRITEVBLK, &rqptr->NetWriteIOsb, 0, 0,
                            XrayHeader, sizeof(XrayHeader)-1, 0, 0, 0, 0);

         if (Debug) fprintf (stdout, "sys$qio() %%%08.08X\n", status);
#endif

#ifdef IP_NETLIB
         DataPtrDsc.dsc$a_pointer = XrayHeader;
         DataPtrDsc.dsc$w_length = sizeof(XrayHeader)-1;

         status = netlib_write (&rqptr->ClientNetLibSocket, &DataPtrDsc,
                                0, 0, &rqptr->NetWriteIOsb, 0, 0);

         if (Debug) fprintf (stdout, "netlib_write() %%X%08.08X\n", status);
#endif

         if (VMSok (status))
            rqptr->BytesTx += sizeof(XrayHeader)-1;
         else
         {
            /* if resource wait enabled only quota not waited for is ASTLM */
            if (status == SS$_EXQUOTA)
               ErrorExitVmsStatus (status, "sys$qio()", FI_LI);
            else
            if (AstFunctionPtr)
            {
               /* write failed, call AST explicitly, status in the IOsb */
               ErrorVmsStatus (rqptr, status, FI_LI);
               rqptr->NetWriteIOsb.Status = status;
               rqptr->NetWriteIOsb.Count = 0;
               if (VMSnok (DclAstStatus =
                   sys$dclast (AstFunctionPtr, rqptr, 0)))
                  ErrorExitVmsStatus (DclAstStatus, "sys$dclast()", FI_LI);
            }
            return (status);
         }
      }

      if (rqptr->ResponseHeaderPtr != NULL)
      {
         /************************/
         /* send response header */
         /************************/

         if (Debug)
            fprintf (stdout, "flushing rqptr->ResponseHeaderPtr\n|%s|\n",
                     rqptr->ResponseHeaderPtr);

         /* indicate to "send only response header" routine it's been sent */
         if (rqptr->HttpMethod == HTTP_METHOD_HEAD)
            rqptr->ResponseHeaderNewlineCount = 2;

         if (!rqptr->ResponseHeaderLength)
            rqptr->ResponseHeaderLength = strlen(rqptr->ResponseHeaderPtr);

         if (AstFunctionPtr)
         {
            /************************/
            /* AST routine, no wait */
            /************************/

            if (DataLength)
            {
               rqptr->ResponseHeaderAstAddress = AstFunctionPtr;
               rqptr->ResponseHeaderDataPtr = DataPtr;
               rqptr->ResponseHeaderDataLength = DataLength;
               AstFunctionPtr = &NetResponseHeaderAst;
            }

#ifdef IP_UCX
            status = sys$qio (0, rqptr->ClientChannel,
                              IO$_WRITEVBLK, &rqptr->NetWriteIOsb,
                              AstFunctionPtr, rqptr,
                              rqptr->ResponseHeaderPtr,
                              rqptr->ResponseHeaderLength, 0, 0, 0, 0);

            if (Debug) fprintf (stdout, "sys$qio() %%%08.08X\n", status);
#endif

#ifdef IP_NETLIB
            DataPtrDsc.dsc$a_pointer = rqptr->ResponseHeaderPtr;
            DataPtrDsc.dsc$w_length = rqptr->ResponseHeaderLength;

            status = netlib_write (&rqptr->ClientNetLibSocket, &DataPtrDsc,
                                   0, 0, &rqptr->NetWriteIOsb,
                                   AstFunctionPtr, rqptr);

            if (Debug) fprintf (stdout, "netlib_write() %%X%08.08X\n", status);
#endif

            if (VMSok (status))
               rqptr->BytesTx += rqptr->ResponseHeaderLength;
            else
            {
               /* write failed, call AST explicitly, status in the IOsb */
               ErrorVmsStatus (rqptr, status, FI_LI);
               rqptr->NetWriteIOsb.Status = status;
               rqptr->NetWriteIOsb.Count = 0;
               if (VMSnok (DclAstStatus =
                   sys$dclast (AstFunctionPtr, rqptr, 0)))
                  ErrorExitVmsStatus (DclAstStatus, "sys$dclast()", FI_LI);
            }

            /* return, the AST function will sys$qiow() the data */
            return (status);
         }
         else
         {
            /************************/
            /* no AST routine, wait */
            /************************/

#ifdef IP_UCX
            status = sys$qiow (0, rqptr->ClientChannel,
                               IO$_WRITEVBLK, &rqptr->NetWriteIOsb, 0, 0,
                               rqptr->ResponseHeaderPtr,
                               rqptr->ResponseHeaderLength, 0, 0, 0, 0);

            if (Debug) fprintf (stdout, "sys$qio() %%%08.08X\n", status);
#endif

#ifdef IP_NETLIB
            DataPtrDsc.dsc$a_pointer = rqptr->ResponseHeaderPtr;
            DataPtrDsc.dsc$w_length = rqptr->ResponseHeaderLength;

            status = netlib_write (&rqptr->ClientNetLibSocket, &DataPtrDsc,
                                   0, 0, &rqptr->NetWriteIOsb, 0, 0);

            if (Debug) fprintf (stdout, "netlib_write() %%X%08.08X\n", status);
#endif

            if (VMSok (status))
               rqptr->BytesTx += rqptr->ResponseHeaderLength;
            else
            {
               /* if resource wait enabled only quota not waited for ASTLM */
               if (status == SS$_EXQUOTA)
                  ErrorExitVmsStatus (status, "sys$qio()", FI_LI);
               else
                  return (status);
            }

            /* continue on to sys$qiow() the data */
         }
      }
   }

   /********/
   /* data */
   /********/

   if (!DataLength)
   {
      if (!AstFunctionPtr) return (status);

      /* without a real network write just fudge the status */
      rqptr->NetWriteIOsb.Status = STS$K_SUCCESS;
      rqptr->NetWriteIOsb.Count = 0;

      if (VMSnok (DclAstStatus =
          sys$dclast (AstFunctionPtr, rqptr, 0)))
         ErrorExitVmsStatus (DclAstStatus, "sys$dclast()", FI_LI);

      return (status);
   }

   if (rqptr->HttpMethod == HTTP_METHOD_HEAD)
   {
      /*****************************/
      /* send only response header */
      /*****************************/

      if (Debug) fprintf (stdout, "method HEAD\n");
      if (rqptr->ResponseHeaderNewlineCount <= 1)
      {
         /* local storage */
         register int  cnt, nlcnt;
         register char  *cptr;

         nlcnt = rqptr->ResponseHeaderNewlineCount;
         if (Debug) fprintf (stdout, "nlcnt: %d\n", nlcnt);
         cptr = DataPtr;
         cnt = DataLength;
         while (cnt--)
         {
            if (*cptr == '\n' && ++nlcnt == 2)
            {
               /* two successive end-of-lines, therefore end of header */
               if (Debug) fprintf (stdout, "end-of-header\n");
               cptr++;
               break;
            }
            else
            if (*cptr != '\r' && *cptr != '\n')
               nlcnt = 0;
            cptr++;
         }
         if ((rqptr->ResponseHeaderNewlineCount = nlcnt) == 2)
         {
            if (Debug) fprintf (stdout, "DataLength: %d\n", cptr - DataPtr);
            if (DataLength = cptr - DataPtr)
            {
               /* adjust data length to include only the HTTP header */
               DataLength = cptr - DataPtr;
            }
            else
            {
               /* no data in HTTP header left at all */
               rqptr->NetWriteIOsb.Status = STS$K_SUCCESS;
               rqptr->NetWriteIOsb.Count = 0;
               if (AstFunctionPtr)
                  if (VMSnok (DclAstStatus =
                      sys$dclast (AstFunctionPtr, rqptr, 0)))
                     ErrorExitVmsStatus (DclAstStatus, "sys$dclast()", FI_LI);
               return (SS$_NORMAL);
            }
         }
      }
      else
      {
         /* HTTP header has been completely sent, absorb anything else */
         rqptr->NetWriteIOsb.Status = STS$K_SUCCESS;
         rqptr->NetWriteIOsb.Count = 0;
         if (AstFunctionPtr)
            if (VMSnok (DclAstStatus =
                sys$dclast (AstFunctionPtr, rqptr, 0)))
               ErrorExitVmsStatus (DclAstStatus, "sys$dclast()", FI_LI);
         return (SS$_NORMAL);
      }
   }

   /*************/
   /* send data */
   /*************/

   if (AstFunctionPtr)
   {
#ifdef IP_UCX
      status = sys$qio (0, rqptr->ClientChannel,
                        IO$_WRITEVBLK, &rqptr->NetWriteIOsb,
                        AstFunctionPtr, rqptr,
                        DataPtr, DataLength, 0, 0, 0, 0);

      if (Debug) fprintf (stdout, "sys$qio() %%%08.08X\n", status);
#endif

#ifdef IP_NETLIB
      DataPtrDsc.dsc$a_pointer = DataPtr;
      DataPtrDsc.dsc$w_length = DataLength;

      status = netlib_write (&rqptr->ClientNetLibSocket, &DataPtrDsc, 0, 0,
                             &rqptr->NetWriteIOsb,
                             AstFunctionPtr, rqptr);

      if (Debug) fprintf (stdout, "netlib_write() %%X%08.08X\n", status);
#endif

      if (VMSnok (status))
      {
         /* write failed, call AST explicitly, status in the IOsb */
         rqptr->NetWriteIOsb.Status = status;
         rqptr->NetWriteIOsb.Count = 0;
         if (VMSnok (DclAstStatus =
             sys$dclast (AstFunctionPtr, rqptr, 0)))
            ErrorExitVmsStatus (DclAstStatus, "sys$dclast()", FI_LI);
      }
   }
   else
   {
#ifdef IP_UCX
      status = sys$qiow (0, rqptr->ClientChannel,
                         IO$_WRITEVBLK, &rqptr->NetWriteIOsb, 0, 0,
                         DataPtr, DataLength, 0, 0, 0, 0);

      if (Debug) fprintf (stdout, "sys$qio() %%%08.08X\n", status);
#endif

#ifdef IP_NETLIB
      DataPtrDsc.dsc$a_pointer = DataPtr;
      DataPtrDsc.dsc$w_length = DataLength;

      status = netlib_write (&rqptr->ClientNetLibSocket, &DataPtrDsc, 0, 0,
                             &rqptr->NetWriteIOsb, 0, 0);

      if (Debug) fprintf (stdout, "netlib_write() %%X%08.08X\n", status);

      /* if resource wait enabled the only quota not waited for is ASTLM */
      if (status == SS$_EXQUOTA)
         ErrorExitVmsStatus (status, "sys$qio()", FI_LI);
#endif
   }

   if (VMSok (status)) rqptr->BytesTx += DataLength;

   return (status);
}

/*****************************************************************************/
/*
Queue up a read from the client over the network. If 'AstFunctionPtr' 
is zero then no I/O completion AST routine is called.  If it is non-zero then 
the function pointed to by the parameter is called when the network write 
completes.

Explicitly declares any AST routine if an error occurs. The calling function
must not do any error recovery if an AST routine has been supplied but the
associated AST routine must!  If an AST was not supplied then the return
status can be checked.
*/ 

int NetRead
(
struct RequestStruct *rqptr,
void *AstFunctionPtr,
char *DataBuffer,
int SizeofDataBuffer
)
{
#ifdef IP_NETLIB
   static $DESCRIPTOR (DataPtrDsc, "");
#endif

   int  status,
        DclAstStatus;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetRead()\n");

   if (AstFunctionPtr)
   {
#ifdef IP_UCX
      status = sys$qio (0, rqptr->ClientChannel,
                        IO$_READVBLK, &rqptr->NetReadIOsb,
                        AstFunctionPtr, rqptr,
                        DataBuffer, SizeofDataBuffer, 0, 0, 0, 0);

      if (Debug) fprintf (stdout, "sys$qio() %%X%08.08X\n", status);
#endif

#ifdef IP_NETLIB
      DataPtrDsc.dsc$a_pointer = DataBuffer;
      DataPtrDsc.dsc$w_length = SizeofDataBuffer;

      status = netlib_read (&rqptr->ClientNetLibSocket, &DataPtrDsc, 0, 0, 0, 0,
                            &rqptr->NetReadIOsb, AstFunctionPtr, rqptr);

      if (Debug) fprintf (stdout, "netlib_read() %%X%08.08X\n", status);
#endif

      /* with resource wait enabled the only quota not waited for is ASTLM */
      if (status == SS$_EXQUOTA)
         ErrorExitVmsStatus (status, "sys$qio()", FI_LI);

      if (VMSnok (status))
      {
         /* queuing of read failed, call AST explicitly, status in the IOsb */
         ErrorVmsStatus (rqptr, status, FI_LI);
         rqptr->NetReadIOsb.Status = status;
         rqptr->NetReadIOsb.Count = 0;
         if (VMSnok (DclAstStatus =
             sys$dclast (AstFunctionPtr, rqptr, 0)))
            ErrorExitVmsStatus (DclAstStatus, "sys$dclast()", FI_LI);
      }
   }
   else
   {
#ifdef IP_UCX
      status = sys$qiow (0, rqptr->ClientChannel,
                         IO$_READVBLK, &rqptr->NetReadIOsb, 0, 0,
                         DataBuffer, SizeofDataBuffer, 0, 0, 0, 0);

      if (Debug) fprintf (stdout, "sys$qio() %%X%08.08X\n", status);
#endif

#ifdef IP_NETLIB
      DataPtrDsc.dsc$a_pointer = DataBuffer;
      DataPtrDsc.dsc$w_length = SizeofDataBuffer;

      status = netlib_read (&rqptr->ClientNetLibSocket, &DataPtrDsc, 0, 0, 0, 0,
                            &rqptr->NetReadIOsb, 0, 0);

      if (Debug) fprintf (stdout, "netlib_read() %%X%08.08X\n", status);
#endif
   }

   return (status);
}

/*****************************************************************************/
/*
This (moderately complex) function buffers multiple 'records' (usually lines 
of text) before writing them as one, larger 'record' via the network to the 
client.  It will escape HTML-forbidden characters if the appropriate flag is 
set. 

This function "double" buffers data.  This scheme allows the buffer to
"overflow" and be written via the network to the client without overwriting
that buffered data  when buffering the new data, and with processing remaining
correct for an AST-driven environment.  This single NetWriteBuffered() call per
AST routine is very  critical to the integrity of the functionality!  This
requires an AST to be supplied with every call supplying data to be output
(i.e. every call EXCEPT an initial allocation call), AND that no more output be
buffered until that AST is delivered (i.e. only one NetWriteBuffered() per
AST-context function.

Uses the thread structure's 'OutputBufferPtr', 'OutputBufferCurrentPtr' and
'OutputBufferRemaining' storage.  Call with 'OutputPtr' set to NULL to
explicitly flush the buffer at any time.  For uniformity of behaviour, an AST 
function MAY be declared with any call, if it's not employed in an actual 
network write, its explicitly declared.  An AST always sets the network write 
IOsb (it's fudged as SS$_NORMAL normal with declared ASTs to simulate an
actual AST-driven write).

Always declares the AST routine and returns a success status even if an error
occurs. The calling function should not do any error recovery but the
associated AST routine must!
*/ 

int NetWriteBuffered
(
struct RequestStruct *rqptr,
void *AstFunctionPtr,
char *OutputPtr,
int OutputLength
)
{
   int  status,
        DclAstStatus,
        CurrentLength;
   unsigned char  *TempDoubleBufferPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
       fprintf (stdout, "NetWriteBuffered() %d %d %d\n",
                OutputPtr, OutputLength, AstFunctionPtr);

   if (rqptr->OutputBufferPtr == NULL)
   {
      /**********************************************/
      /* allocate heap memory for the output buffer */
      /**********************************************/

      if (Debug) fprintf (stdout, "allocate\n");

      if ((rqptr->OutputBufferPtr =
          HeapAlloc (rqptr, OutputBufferSize+2)) == NULL)
      {
         /* heap allocation failed, fudge a network write */
         ErrorVmsStatus (rqptr, SS$_INSFMEM, FI_LI);
         rqptr->NetWriteIOsb.Status = SS$_INSFMEM;
         rqptr->NetWriteIOsb.Count = 0;
         if (AstFunctionPtr)
            if (VMSnok (DclAstStatus = sys$dclast (AstFunctionPtr, rqptr, 0)))
               ErrorExitVmsStatus (DclAstStatus, "sys$dclast()", FI_LI);
         return (SS$_NORMAL);
      }
      else
      {
         rqptr->OutputBufferCurrentPtr = rqptr->OutputBufferPtr;
         rqptr->OutputBufferRemaining = rqptr->OutputBufferSize =
            OutputBufferSize;
      }
   }

   if (OutputPtr == NULL)
   {
      /*************************/
      /* flush buffer contents */
      /*************************/

      if (Debug)
         fprintf (stdout, "flush %d bytes\n|%s|\n",
                  rqptr->OutputBufferCurrentPtr - rqptr->OutputBufferPtr,
                  rqptr->OutputBufferPtr);

      if (rqptr->OutputBufferCurrentPtr > rqptr->OutputBufferPtr)
      {
         NetWrite (rqptr, AstFunctionPtr,
                   rqptr->OutputBufferPtr,
                   rqptr->OutputBufferCurrentPtr - rqptr->OutputBufferPtr);
      }
      else
      {
         /* no network write actually occured, fudge it */
         rqptr->NetWriteIOsb.Status = STS$K_SUCCESS;
         rqptr->NetWriteIOsb.Count = 0;
         if (AstFunctionPtr)
            if (VMSnok (DclAstStatus = sys$dclast (AstFunctionPtr, rqptr, 0)))
               ErrorExitVmsStatus (DclAstStatus, "sys$dclast()", FI_LI);
      }

      rqptr->OutputBufferCurrentPtr = rqptr->OutputBufferPtr;
      rqptr->OutputBufferRemaining = OutputBufferSize;

      return (SS$_NORMAL);
   }

   /*********************************************/
   /* after here an AST must have been supplied */
   /*********************************************/

   if (!AstFunctionPtr)
      ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);

   if (Debug)
      fprintf (stdout, "[%d] %d\n|%*s|\n",
               rqptr->NetWriteBufferedEscapeHtml,
               OutputLength, OutputLength, OutputPtr);

   if (rqptr->NetWriteBufferedEscapeHtml)
   {
      /************************************/
      /* escape HTML-forbidden characters */
      /************************************/

      /* local storage */
      register int  bcnt, ocnt;
      register unsigned char  *bptr, *optr;

      if (Debug) fprintf (stdout, "escape-HTML\n");

      TempDoubleBufferPtr = NULL;
      optr = OutputPtr;
      ocnt = OutputLength;
      bptr = rqptr->OutputBufferCurrentPtr;
      bcnt = rqptr->OutputBufferRemaining;

      while (ocnt)
      {
         /* worst-case is having to escape an ampersand */
         if (bcnt < 5)
         {
            if (TempDoubleBufferPtr != NULL)
            {
               /*
                  If there should be so many HTML-escaped characters that
                  the output buffer needs flushing twice ... then don't,
                  its not designed to allow that.  Instead wriggle out of
                  it gracefully (as possible), providing an error message.
                  An AST has already been queued from the NetWrite() below.
                  This should seldom happen, if at all!
               */
               ErrorVmsStatus (rqptr, RMS$_RTB, FI_LI);
               return (SS$_NORMAL);
            }

            /* only provide the double-buffering when necessary */
            if (rqptr->OutputDoubleBufferPtr == NULL)
            {
               if (Debug) fprintf (stdout, "double-buffering\n");
               if ((rqptr->OutputDoubleBufferPtr =
                   HeapAlloc (rqptr, rqptr->OutputBufferSize+2)) == NULL)
               {
                  /* explicitly declare an AST */
                  ErrorVmsStatus (rqptr, SS$_INSFMEM, FI_LI);
                  rqptr->NetWriteIOsb.Status = SS$_INSFMEM;
                  rqptr->NetWriteIOsb.Count = 0;
                  if (VMSnok (DclAstStatus =
                      sys$dclast (AstFunctionPtr, rqptr, 0)))
                     ErrorExitVmsStatus (status, "sys$dclast()", FI_LI);
                  return (SS$_NORMAL);
               }
            }

            NetWrite (rqptr, AstFunctionPtr,
                      rqptr->OutputBufferPtr,
                      bptr - rqptr->OutputBufferPtr);

            TempDoubleBufferPtr = rqptr->OutputDoubleBufferPtr;
            rqptr->OutputDoubleBufferPtr = rqptr->OutputBufferPtr;
            rqptr->OutputBufferPtr = TempDoubleBufferPtr;

            bptr = rqptr->OutputBufferPtr;
            bcnt = OutputBufferSize;

            /* ensure any AST function is queued only once! */
            AstFunctionPtr = 0;
         }

         switch (*optr)
         {
            case '<' :
               memcpy (bptr, "&lt;", 4); bptr += 4; bcnt -= 4; optr++; break;
            case '>' :
               memcpy (bptr, "&gt;", 4); bptr += 4; bcnt -= 4; optr++; break;
            case '&' :
               memcpy (bptr, "&amp;", 5); bptr += 5; bcnt -= 5; optr++; break;
            default :
               *bptr++ = *optr++; bcnt--;
         }
         ocnt--;
      }

      rqptr->OutputBufferCurrentPtr = bptr;
      rqptr->OutputBufferRemaining = bcnt;
   }
   else
   {
      /*****************/
      /* buffer output */
      /*****************/

      if (OutputLength > rqptr->OutputBufferRemaining)
      {
         /************************************/
         /* insufficient space, flush buffer */
         /************************************/

         if (Debug)
            fprintf (stdout, "full-flush\n|%s|\n", rqptr->OutputBufferPtr);

         if (OutputLength > rqptr->OutputBufferSize)
         {
            /* record is actually too big to buffer! */
            ErrorVmsStatus (rqptr, RMS$_RTB, FI_LI);
            rqptr->NetWriteIOsb.Status = RMS$_RTB;
            rqptr->NetWriteIOsb.Count = 0;
            if (VMSnok (DclAstStatus =
                sys$dclast (AstFunctionPtr, rqptr, 0)))
               ErrorExitVmsStatus (DclAstStatus, "sys$dclast()", FI_LI);
            return (SS$_NORMAL);
         }

         /* only provide the double-buffering when necessary */
         if (rqptr->OutputDoubleBufferPtr == NULL)
         {
            if (Debug) fprintf (stdout, "double-buffering\n");
            if ((rqptr->OutputDoubleBufferPtr =
                HeapAlloc (rqptr, rqptr->OutputBufferSize+2)) == NULL)
            {
               /* declare an AST to do it */
               ErrorVmsStatus (rqptr, SS$_INSFMEM, FI_LI);
               rqptr->NetWriteIOsb.Status = SS$_INSFMEM;
               rqptr->NetWriteIOsb.Count = 0;
               if (VMSnok (DclAstStatus =
                   sys$dclast (AstFunctionPtr, rqptr, 0)))
                  ErrorExitVmsStatus (DclAstStatus, "sys$dclast()", FI_LI);
               return (SS$_NORMAL);
            }
         }

         NetWrite (rqptr, AstFunctionPtr,
                   rqptr->OutputBufferPtr,
                   rqptr->OutputBufferCurrentPtr - rqptr->OutputBufferPtr);

         TempDoubleBufferPtr = rqptr->OutputDoubleBufferPtr;
         rqptr->OutputDoubleBufferPtr = rqptr->OutputBufferPtr;
         rqptr->OutputBufferPtr = TempDoubleBufferPtr;

         rqptr->OutputBufferCurrentPtr = rqptr->OutputBufferPtr;
         rqptr->OutputBufferRemaining = OutputBufferSize;

         /* ensure any AST function is queued only once! */
         AstFunctionPtr = 0;
      }

      /* buffer new output */
      memcpy (rqptr->OutputBufferCurrentPtr, OutputPtr, OutputLength);
      rqptr->OutputBufferCurrentPtr += OutputLength;
      rqptr->OutputBufferRemaining -= OutputLength;
   }

   /******************/
   /* declare an AST */
   /******************/

   /* if a network write has not been done then explicitly declare the AST */
   if (AstFunctionPtr)
   {
      /* fudge the fact a network write has not actually been done! */
      rqptr->NetWriteIOsb.Status = SS$_NORMAL;
      rqptr->NetWriteIOsb.Count = OutputLength;
      if (VMSnok (DclAstStatus =
          sys$dclast (AstFunctionPtr, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", FI_LI);
   }

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Report that the initial memory allocation failed.
*/
 
#ifdef IP_UCX
   NetRequestCallocFailed
(
unsigned short Channel,
char *HostName,
char *InternetAddress
)
#endif
#ifdef IP_NETLIB
   NetRequestCallocFailed
(
unsigned long *NetLibSocketPtr,
char *HostName,
char *InternetAddress
)
#endif
 
{
   static $DESCRIPTOR (MessageFaoDsc,
"HTTP/1.0 500 Error\r\n\
Content-Type: text/html\r\n\
\r\n\
<HTML>\n\
<HEAD>\n\
<TITLE>!AZ 500</TITLE>\n\
</HEAD>\n\
<BODY>\n\
!AZ\n\
</BODY>\n\
</HTML>\n");

   short  Length;
   char  Buffer [256];
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   sys$fao (&MessageFaoDsc, &Length, &BufferDsc,
            MsgForNoRequest(HostName,InternetAddress,MSG_STATUS_ERROR),
            MsgForNoRequest(HostName,InternetAddress,
                            MSG_GENERAL_REQUEST_CALLOC));
   Buffer[Length] = '\0';

#ifdef IP_UCX
   NetDirectWrite (Channel, Buffer, Length);
#endif
#ifdef IP_NETLIB
   NetDirectWrite (NetLibSocketPtr, Buffer, Length);
#endif
}

/*****************************************************************************/
/*
Report that client host name has been denied access to the server.
*/
 
#ifdef IP_UCX
   NetAccessDenied
(
unsigned short Channel,
char *HostName,
char *InternetAddress
)
#endif
#ifdef IP_NETLIB
   NetAccessDenied
(
unsigned long *NetLibSocketPtr,
char *HostName,
char *InternetAddress
)
#endif
 
{
   static $DESCRIPTOR (MessageFaoDsc,
"HTTP/1.0 403 Forbidden\r\n\
Content-Type: text/html\r\n\
\r\n\
<HTML>\n\
<HEAD>\n\
<TITLE>!AZ 403</TITLE>\n\
</HEAD>\n\
<BODY>\n\
!AZ\n\
</BODY>\n\
</HTML>\n");

   short  Length;
   char  Buffer [256];
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   sys$fao (&MessageFaoDsc, &Length, &BufferDsc,
            MsgForNoRequest(HostName,InternetAddress,MSG_STATUS_ERROR),
            MsgForNoRequest(HostName,InternetAddress,MSG_GENERAL_ACCESS_DENIED));
   Buffer[Length] = '\0';

#ifdef IP_UCX
   NetDirectWrite (Channel, Buffer, Length);
#endif
#ifdef IP_NETLIB
   NetDirectWrite (NetLibSocketPtr, Buffer, Length);
#endif
}

/*****************************************************************************/
/*
Report that there is already the limit of client connections.
*/
 
#ifdef IP_UCX
   NetTooBusy
(
unsigned short Channel,
char *HostName,
char *InternetAddress
)
#endif
#ifdef IP_NETLIB
   NetTooBusy
(
unsigned long *NetLibSocketPtr,
char *HostName,
char *InternetAddress
)
#endif
 
{
   static $DESCRIPTOR (MessageFaoDsc,
"HTTP/1.0 502 Busy\r\n\
Content-Type: text/html\r\n\
\r\n\
<HTML>\n\
<HEAD>\n\
<TITLE>!AZ 502</TITLE>\n\
</HEAD>\n\
<BODY>\n\
!AZ\n\
</BODY>\n\
</HTML>\n");

   short  Length;
   char  Buffer [256];
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   sys$fao (&MessageFaoDsc, &Length, &BufferDsc,
            MsgForNoRequest(HostName,InternetAddress,MSG_STATUS_ERROR),
            MsgForNoRequest(HostName,InternetAddress,MSG_GENERAL_TOO_BUSY));
   Buffer[Length] = '\0';

#ifdef IP_UCX
   NetDirectWrite (Channel, Buffer, Length);
#endif
#ifdef IP_NETLIB
   NetDirectWrite (NetLibSocketPtr, Buffer, Length);
#endif
}

/****************************************************************************/

