/*****************************************************************************/
/*
                             (what')snew.c

CGI-compliant script to maintain and access a (simple) "database" of changes
to a web (new and revised).  The "database" is a plain-text file where each
line represents an entry.

This script uses HTTP redirection extensively to get the files and directories
specified separately ("parent" and "file" CGI variables) available as a path
info CGI variable.

This program acts as a CGI script and also as a command-line utility (data
file purging) so ensure any error messages are reported appropriately.

This script requires two HTTPd mappings.  The first should be "snew", to allow
the What's New data to be displayed.  The second "snewadd", that allows
entries to be added to the data file.  Although names other than "snew" could
be used for the utility the second must contain the string "add".


Format of Data File
-------------------

path dd/mm/yyyy d name c description-to-end-of-line

where:

  o  path             URL path to the document
  o  dd/mm/yyyy       date of the snew/change
  o  d                number of days before the entry is removed
  o  name             authenticated name of user adding entry (or ?)
  o  c                'N' (new) 'R' (revised) 'U' (updated)
  o  description...   description of the snew/change

For example:

/web/admin/reports/report1.htm 1/9/96 14 user-name N Report One
/web/admin/reports/snew/ 1/9/96 14 user-name U Folder containing new reports.
/web/technical/html3/ 2/9/96 14 user-name N Snew documentation on HTML 3


Data File Maintentance
----------------------

The data file should be purged occasionally to reduce the number of expired
entries.  The parameter to the purge qualifier is the number of days an
entry's should be before purging (and expired of course!)  This can be done
manually or included in the hypertext services (system) startup in the
following manner (here, three months old):

  $ SNEW = "$HT_EXE:SNEW"
  $ SNEW /PURGE=90

This will create a new version of the file with expired entries removed.


CGI VARIABLE NAMES (GET method)
------------------
WWW_FORM_DO             function
WWW_FORM_DIRECTORY      directory specification (URL format)   
WWW_FORM_FILE           file name (name and type only)
WWW_FORM_PARENT         parent directory
WWW_FORM_TITLE          title of page (optional, otherwise a link)
WWW_FORM_TITLEURL       the URL representing the title
WWW_PATH_INFO           the URL path component
WWW_PATH_TRANSLATED     the VMS equivalent of the URL path component
WWW_QUERY_STRING        any of the form components
WWW_REMOTE_USER         the authenticated name of the user adding the entry
WWW_SCRIPT_NAME         the name of the script being executed
WWW_SERVER_NAME         the server IP name
WWW_SERVER_PORT         the server IP port

CGI VARIABLE NAMES (POST method)
------------------
DESCRIPTION             textual description     
LIFETIME                lifetime of entry (in days)
STATUS                  new, revised, updated, reminder


BUILD
-----
See BUILD_snew.COM


VERSION HISTORY
---------------
12-AUG-97  MGD  v1.2.0, change ReadHttpInput() for HTTPd v4.3 body-only POSTs
06-MAY-97  MGD  v1.1.0, now employs POST method to add entries to database
08-SEP-96  MGD  v1.0.0, initial development
*/
/*****************************************************************************/

#ifdef __ALPHA
   char SoftwareID [] = "SNEW AXP-1.2.0";
#else
   char SoftwareID [] = "SNEW VAX-1.2.0";
#endif

#ifdef __ALPHA
#   pragma nomember_alignment
#endif

/* standard header files */
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <file.h>

/* VMS header files */
#include <descrip.h>
#include <libdtdef.h>
#include <rms.h>
#include <ssdef.h>
#include <stsdef.h>

#define VMSok(x) ((x) & STS$M_SUCCESS)
#define VMSnok(x) !(((x) & STS$M_SUCCESS))

#define boolean int
#define true 1
#define false 0

#define SNEW_NEW 1
#define SNEW_REVISED 2
#define SNEW_UPDATED 3

/* macro for a simple boolean version of case-insensitive compare */
#define strsame(str1,str2) strnsame (str1,str2,-1)

/* macro provides NULL pointer if CGI variable does not exist */
#define GetCgiVarIfExists(CharPointer,CgiVariableName) \
   CharPointer = getenv(CgiVariableName)

/* macro provides pointer to empty string even if CGI variable does not exist */
#define GetCgiVar(CharPointer,CgiVariableName) \
   if ((CharPointer = getenv(CgiVariableName)) == NULL) \
       CharPointer = ""; \
   if (Debug) fprintf (stdout, "%s |%s|\n", CgiVariableName, CharPointer);

char  AboutUrl [] = "/snew/-/about.html",
      AboutButton [] = "About <I>What's New?</I>",
      DbFileName [] = "SNEW$DATA",
      ExpiredIcon [] = "<IMG SRC=\"/snew/-/expired.gif\">",
      HelpUrl [] = "/snew/-/help.html",
      HelpButton [] = "<SUP><I>?</I></SUP>",
      NewIcon [] = "<IMG SRC=\"/snew/-/new.gif\">",
      ReminderIcon [] = "<IMG SRC=\"/snew/-/reminder.gif\">",
      RevisedIcon [] = "<IMG SRC=\"/snew/-/revised.gif\">",
      UpdatedIcon [] = "<IMG SRC=\"/snew/-/updated.gif\">",
      Utility [] = "SNEW",
      WhatsNewIcon [] = "<IMG SRC=\"/snew/-/snew.gif\">";

char  Http200Header [] =
"HTTP/1.0 200 Document follows.\r\n\
Content-Type: text/html\r\n\
\r\n";

char  Http404Header [] =
"HTTP/1.0 404 Error report follows.\r\n\
Content-Type: text/html\r\n\
\r\n";

struct SnewStruct
{
   struct SnewStruct  *PrevPtr;
   struct SnewStruct  *NextPtr;
   int  Age,
        Day,
        LifeTime,
        Month,
        Year;
   char  StatusChar;
   char  *DescriptionPtr,
         *PathPtr,
         *RemoteUserPtr;
   /* path, description and user strings are stored here and pointed to */
};

boolean  Debug,
         CgiScript,
         HttpHasBeenOutput;

int  Expired30Days,
     Expired60Days,
     Expired90Days,
     Expired180Days,
     ExpiredAgesAgo,
     EntriesInFile,
     EntriesInList,
     FormListSize = 5;

char  *CgiContentLengthPtr,
      *CgiContentTypePtr,
      *CgiFormAgePtr,
      *CgiFormDirectoryPtr,
      *CgiFormDoPtr,
      *CgiFormFilePtr,
      *CgiFormParentPtr,
      *CgiFormTitlePtr,
      *CgiFormTitleUrlPtr,
      *CgiRequestMethodPtr,
      *CgiPostDescriptionPtr,
      *CgiPostLifeTimePtr,
      *CgiPostStatusPtr,
      *CgiScriptNamePtr,
      *CgiServerNamePtr,
      *CgiServerPortPtr,
      *CgiPathInfoPtr,
      *CgiPathTranslatedPtr,
      *CgiRemoteUserPtr;
      
struct SnewStruct  *ListHeadPtr = NULL,
                   *ListTailPtr = NULL;

FILE  *HttpOut;

/* required function prototypes */
char* CopyToHtml (char*, char*, int);
char* ReadHttpInput (char*);

/*****************************************************************************/
/*
*/

main (int argc, char *argv[])

{
   register char  *cptr;

   /*********/
   /* begin */
   /*********/


   if (argc > 1)
   {
      /**********************/
      /* command-line usage */
      /**********************/

      int  cnt,
           PurgeDays;

      CgiScript = false;
      PurgeDays = 0;
      for (cnt = 1; cnt < argc; cnt++)
      {
         if (strsame (argv[cnt], "/DBUG"))
            Debug = true;
         else
         if (strsame (argv[cnt], "/PURGE="))
         {
            if (!isdigit(argv[cnt][7]))
            {
               fprintf (stdout, "%%%s-E-IVPARM, invalid parameter\n \\%s\\\n",
                        Utility, argv[cnt]);
               exit (STS$K_ERROR | STS$M_INHIB_MSG);
            }
            PurgeDays = atoi (argv[cnt]+7);
         }
         else
         {
            fprintf (stdout,
                "%%%s-E-IVQUAL, unrecognised qualifier\n \\%s\\\n",
                Utility, argv[cnt]);
            exit (STS$K_ERROR | STS$M_INHIB_MSG);
         }
      }
      if (PurgeDays) PurgeDataFile (PurgeDays);
      exit (SS$_NORMAL);
   }

   /********************/
   /* CGI-script usage */
   /********************/

   CgiScript = true;

   /* open another output stream so that the '\r' and '\n' are not filtered */
#ifdef __DECC
   if ((HttpOut = fopen ("SYS$OUTPUT", "w", "ctx=bin")) == NULL)
      exit (vaxc$errno);
#else
   if ((HttpOut = fopen ("SYS$OUTPUT", "w", "rfm=udf")) == NULL)
      exit (vaxc$errno);
#endif

   if (getenv ("SNEW$DBUG") != NULL)
   {
      Debug = true;
      system ("SHOW SYM WWW_*");
   }

   /* uses redirection extensively, might as well get these now! */
   GetCgiVar (CgiServerNamePtr, "WWW_SERVER_NAME");
   GetCgiVar (CgiServerPortPtr, "WWW_SERVER_PORT");
   GetCgiVar (CgiScriptNamePtr, "WWW_SCRIPT_NAME");
   GetCgiVar (CgiRequestMethodPtr, "WWW_REQUEST_METHOD");
   GetCgiVar (CgiPathInfoPtr, "WWW_PATH_INFO");
   GetCgiVar (CgiPathTranslatedPtr, "WWW_PATH_TRANSLATED");
   GetCgiVar (CgiRemoteUserPtr, "WWW_REMOTE_USER");
   GetCgiVar (CgiContentTypePtr, "WWW_CONTENT_TYPE");
   GetCgiVar (CgiContentLengthPtr, "WWW_CONTENT_LENGTH");

   if (CgiServerPortPtr[0])
   {
      /* if default port then empty string, otherwise specify port */
      if (!strcmp (CgiServerPortPtr, "80"))
         CgiServerPortPtr = "";
      else
      {
         static char  ServerPort [16];
         sprintf (ServerPort, ":%s", CgiServerPortPtr);
         CgiServerPortPtr = ServerPort;
      }
   }

   /* not all of these variables are used for all activities :^) */
   GetCgiVar (CgiFormAgePtr, "WWW_FORM_AGE");
   GetCgiVar (CgiFormDirectoryPtr, "WWW_FORM_DIRECTORY");
   GetCgiVar (CgiFormDoPtr, "WWW_FORM_DO");
   GetCgiVar (CgiFormFilePtr, "WWW_FORM_FILE");
   GetCgiVar (CgiFormParentPtr, "WWW_FORM_PARENT");
   GetCgiVar (CgiFormTitlePtr, "WWW_FORM_TITLE");
   GetCgiVar (CgiFormTitleUrlPtr, "WWW_FORM_TITLEURL");

   if (!strcmp(CgiRequestMethodPtr, "POST"))
   {
      /****************/
      /* add an entry */
      /****************/

      AddToDataFile ();
      exit (SS$_NORMAL);
   }

   if (CgiFormDoPtr[0] == 'c' && CgiFormDoPtr[1] == 'd')
   {
      /*********************************/
      /* change directory, REDIRECTION */
      /*********************************/

      if (!CgiFormDirectoryPtr[0])
         exit (ErrorGeneral ("Select a subdirectory from the list.",
                             __FILE__, __LINE__));

      if (CgiFormDirectoryPtr[0] == '.')
         exit (ErrorGeneral ("Cannot &quot;goto&quot; current directory.",
                             __FILE__, __LINE__));

      fprintf (HttpOut, "Location: %s%s%s/?do=menu\r\n\r\n",
               CgiScriptNamePtr, CgiFormParentPtr, CgiFormDirectoryPtr);

      exit (SS$_NORMAL);
   }

   if (CgiFormDoPtr[0] == 'l' && CgiFormDoPtr[1] == 'i')
   {
      /*******************************/
      /* list directory, REDIRECTION */
      /*******************************/

      if (!CgiFormDirectoryPtr[0])
         exit (ErrorGeneral ("Select a subdirectory from the list.",
                             __FILE__, __LINE__));

      if (CgiFormDirectoryPtr[0] == '.')
      {
         CgiFormDirectoryPtr = "";
         cptr = "*.*";
      }
      else
         cptr = "/*.*";

      fprintf (HttpOut, "Location: http://%s%s%s%s%s\r\n\r\n",
         CgiServerNamePtr, CgiServerPortPtr,
         CgiFormParentPtr, CgiFormDirectoryPtr, cptr);

      exit (SS$_NORMAL);
   }

   if (CgiFormDoPtr[0] == 'e' && CgiFormDoPtr[1] == 'n')
   {
      /*************************/
      /* add entry REDIRECTION */
      /*************************/

      if (!CgiFormFilePtr[0] && !CgiFormDirectoryPtr[0])
         exit (ErrorGeneral ("Select a file or directory from the lists.",
                             __FILE__, __LINE__));

      if (CgiFormFilePtr[0])
         fprintf (HttpOut, "Location: %s%s%s?do=ENTRY\r\n\r\n",
                  CgiScriptNamePtr, CgiFormParentPtr, CgiFormFilePtr);
      else
      if (CgiFormDirectoryPtr[0] == '.')
      {
         /* current directory is represented by the period */
         fprintf (HttpOut, "Location: %s%s?do=ENTRY\r\n\r\n",
                  CgiScriptNamePtr, CgiFormParentPtr);
      }
      else
         fprintf (HttpOut, "Location: %s%s%s/?do=ENTRY\r\n\r\n",
                  CgiScriptNamePtr, CgiFormParentPtr, CgiFormDirectoryPtr);

      exit (SS$_NORMAL);
   }

   if (CgiFormDoPtr[0] == 'E' && CgiFormDoPtr[1] == 'N')
   {
      /******************************/
      /* provide the add entry form */
      /******************************/

      SnewEntry ();
      exit (SS$_NORMAL);
   }

   /* script name should be "snewadd" */
   if (strstr (CgiScriptNamePtr, "add") != NULL)
   {
      /* if a file then it's an entry, if a directory then a menu */
      for (cptr = CgiPathInfoPtr; *cptr; cptr++);
      while (cptr > CgiPathInfoPtr && *cptr != '/') cptr--;
      while (cptr > CgiPathInfoPtr && *cptr == '/') cptr++;
      if (*cptr)
         SnewEntry ();
      else
         SnewMenu ();
      exit (SS$_NORMAL);
   }

   /* just display the What's New data */
   WhatsNew ();

   exit (SS$_NORMAL);
}

/*****************************************************************************/
/*
Display the page of "what's new" matching the specified path.
*/

WhatsNew ()

{
   register char  *cptr;

   boolean  AgeOneDay,
            AgeOneWeek,
            AgeOneMonth,
            AgeOneYear,
            AgeThreeMonths,
            AgeUnbelievable;
   int  Age,
        ExpiredCount;
   char  InTheLast [128];
   char  *ExpiredPtr,
         *StatusPtr,
         *TitlePtr,
         *TitleUrlPtr;
   struct SnewStruct  *llptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "WhatsNew()\n");

   /* if a file has been specified then get the directory from it */
   for (cptr = CgiPathInfoPtr; *cptr; cptr++);
   while (cptr > CgiPathInfoPtr && *cptr != '/') cptr--;
   if (*cptr == '/') *++cptr = '\0';

   Age = atoi(CgiFormAgePtr);
   if (Age)
      sprintf (InTheLast,
         "<I>All entries (expired or not) in the last %d days.</I>\n", Age);
   else
      InTheLast[0] = '\0';

   ReadDataFile (CgiPathInfoPtr, Age);

   SortDataList ();

   if (CgiFormTitlePtr[0])
      TitlePtr = CgiFormTitlePtr;
   else
      TitlePtr = CgiPathInfoPtr;

   if (CgiFormTitleUrlPtr[0])
      TitleUrlPtr = CgiFormTitleUrlPtr;
   else
      TitleUrlPtr = CgiPathInfoPtr;

   fprintf (HttpOut,
"%s\
<!-- SoftwareID: %s -->\n\
<HTML>\n\
<TITLE>What's New? %s</TITLE>\n\
<H1>%s <A HREF=\"%s\">%s</A></H1>\n\
%s\
<HR SIZE=3>\n\
<P>\n",
      Http200Header, SoftwareID,
      CgiPathInfoPtr,
      WhatsNewIcon,
      TitleUrlPtr, TitlePtr,
      InTheLast);
   HttpHasBeenOutput = true;

   AgeOneDay = AgeOneWeek = AgeOneMonth = AgeThreeMonths =
      AgeOneYear = AgeUnbelievable = false;
   ExpiredCount = 0;

   llptr = ListHeadPtr;
   while (llptr != NULL)
   {
      if (llptr->Age > llptr->LifeTime)
      {
         ExpiredCount++;
         ExpiredPtr = ExpiredIcon;
      }
      else
         ExpiredPtr = "";

      if (llptr->Age <= 1 && !AgeOneDay)
      {
         AgeOneDay = true;
         fprintf (HttpOut, "<H2>Since yesterday ...</H2>\n<DL>\n");
      }
      else
      if (llptr->Age > 1 && llptr->Age <= 7 && !AgeOneWeek)
      {
         AgeOneWeek = true;
         if (AgeOneDay) fprintf (HttpOut, "</DL>\n");
         fprintf (HttpOut, "<H2>In the last week ...</H2>\n<DL>\n");
      }
      else
      if (llptr->Age > 7 && llptr->Age <= 28 && !AgeOneMonth)
      {
         AgeOneMonth = true;
         if (AgeOneDay || AgeOneWeek) fprintf (HttpOut, "</DL>\n");
         fprintf (HttpOut, "<H2>In the last month ...</H2>\n<DL>\n");
      }
      else
      if (llptr->Age > 28 && llptr->Age <= 90 && !AgeThreeMonths)
      {
         AgeThreeMonths = true;
         if (AgeOneDay || AgeOneWeek || AgeOneMonth)
            fprintf (HttpOut, "</DL>\n");
         fprintf (HttpOut, "<H2>Older than one month ...</H2>\n<DL>\n");
      }
      else
      if (llptr->Age > 90 && llptr->Age <= 365 && !AgeOneYear)
      {
         AgeOneYear = true;
         if (AgeOneDay || AgeOneWeek || AgeOneMonth || AgeThreeMonths)
            fprintf (HttpOut, "</DL>\n");
         fprintf (HttpOut, "<H2>Older than three months ...</H2>\n<DL>\n");
      }
      else
      if (llptr->Age > 365 && !AgeUnbelievable)
      {
         AgeUnbelievable = true;
         if (AgeOneWeek || AgeOneMonth || AgeOneYear)
            fprintf (HttpOut, "</DL>\n");
         fprintf (HttpOut, "<H2>Unbelievably old ...</H2>\n<DL>\n");
      }

      if (llptr->StatusChar == 'M')
         StatusPtr = ReminderIcon;
      else
      if (llptr->StatusChar == 'N')
         StatusPtr = NewIcon;
      else
      if (llptr->StatusChar == 'R')
         StatusPtr = RevisedIcon;
      else
      if (llptr->StatusChar == 'U')
         StatusPtr = UpdatedIcon;
      else
         StatusPtr = "";

      fprintf (HttpOut,
"<DD><NOBR><TT>%02.02d/%02.02d/%02.02d</TT><!-- %d/%d --> \
%s %s<A HREF=\"%s\">%s</A></NOBR>\n",
         llptr->Day, llptr->Month, llptr->Year % 100,
         llptr->Age, llptr->LifeTime,
         StatusPtr, ExpiredPtr, llptr->PathPtr, llptr->DescriptionPtr);

      llptr = llptr->NextPtr;
   }

   if (!EntriesInList) fprintf (HttpOut, "<DL>\n<DD><I>(nothing new!)</I>\n");

   fprintf (HttpOut, "</DL>\n<HR SIZE=3>\n");

   if (ExpiredCount)
      fprintf (HttpOut, "%s marked entries have expired.\n<BR>", ExpiredIcon);

   if (!Age)
   {
      char  UrlTitle [256];

      if (CgiFormTitlePtr[0])
      {
         memcpy (UrlTitle, "&title=", 7);
         CopyStringToUrl (UrlTitle+7, CgiFormTitlePtr, sizeof(UrlTitle)-7,
                          __FILE__, __LINE__);
      }
      else
         UrlTitle[0] = '\0';

      fprintf (HttpOut, "<B>All entries</B> (expired or not) for last\n");

      if (Expired30Days)
         fprintf (HttpOut, "<A HREF=\"%s%s?age=30%s\">30</A>,\n",
                  CgiScriptNamePtr, CgiPathInfoPtr, UrlTitle);
      else
         fprintf (HttpOut, "30,\n");

      if (Expired60Days)
         fprintf (HttpOut, "<A HREF=\"%s%s?age=60%s\">60</A>,\n",
                  CgiScriptNamePtr, CgiPathInfoPtr, UrlTitle);
      else
         fprintf (HttpOut, "60,\n");

      if (Expired90Days)
         fprintf (HttpOut, "<A HREF=\"%s%s?age=90%s\">90</A>,\n",
                  CgiScriptNamePtr, CgiPathInfoPtr, UrlTitle);
      else
         fprintf (HttpOut, "90,\n");

      if (Expired180Days)
         fprintf (HttpOut, "<A HREF=\"%s%s?age=180%s\">180</A>\ndays.\n",
                  CgiScriptNamePtr, CgiPathInfoPtr, UrlTitle);
      else
         fprintf (HttpOut, "180\ndays.\n");

      if (ExpiredAgesAgo)
         fprintf (HttpOut,
"<A HREF=\"%s%s?age=99999%s\">(plus unbelievably old!)</A>\n<BR>",
            CgiScriptNamePtr, CgiPathInfoPtr, UrlTitle);
      else
         fprintf (HttpOut, "<BR>");
   }

   fprintf (HttpOut, "<A HREF=\"%s\">%s</A>\n</HTML>\n",
            AboutUrl, AboutButton);
}

/*****************************************************************************/
/*
If 'FilterPath' is not empty the path in the database is matched against the
filter path.  If the filter path is a leading substring of the database path
the entry is used, otherwise it is filtered out :^)  If 'FilterAge' is zero
all ages are used.  If a positive number all entries with an age less than or
equal to the filter age (expired or not) are used (path match also filters).
If a negative number all those entries with an age greater than or equal to
the filter age (expired of not) are used.  Without 'FilterAge' any expired
entries (i.e. age greater than lifetime) are filtered out.
*/

ReadDataFile
(
char *FilterPath,
int FilterAge
)
{
   static int  LibJulianDate = LIB$K_JULIAN_DATE;

   register char  *cptr, *lptr, *pptr, *zptr;

   int  status,
        Age,
        FilterPathLength,
        DescriptionLength,
        Day,
        Length,
        LifeTime,
        LineCount,
        Month,
        PathLength,
        RemoteUserLength,
        Year;
   unsigned long  CurrentJulianDate,
                  EntryJulianDate;
   unsigned long  BinaryTime [2];
   unsigned short  NumericTime [7];
   char  StatusChar;
   char  *DayPtr,
         *DescriptionPtr,
         *LifeTimePtr,
         *MonthPtr,
         *RemoteUserPtr,
         *PathPtr,
         *YearPtr;
   char  Line [256];
   struct SnewStruct  *SnewPtr;
   struct FAB  FileFab;
   struct RAB  FileRab;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ReadDataFile()\n");

   FileFab = cc$rms_fab;
   FileFab.fab$l_fna = DbFileName;
   FileFab.fab$b_fns = strlen(DbFileName);
   FileFab.fab$b_shr = FAB$M_SHRGET | FAB$M_SHRPUT;

   if (VMSnok (status = sys$open (&FileFab, 0, 0)))
      exit (ErrorVmsStatus (status, "database", DbFileName,
            __FILE__, __LINE__));

   FileRab = cc$rms_rab;
   FileRab.rab$l_fab = &FileFab;
   FileRab.rab$b_mbf = 2;
   FileRab.rab$l_rop = RAB$M_RAH;
   FileRab.rab$l_ubf = Line;
   FileRab.rab$w_usz = sizeof(Line)-1;

   if (VMSnok (status = sys$connect (&FileRab, 0, 0)))
      exit (ErrorVmsStatus (status, "database", DbFileName,
            __FILE__, __LINE__));

   if (VMSnok (status =
       lib$cvt_from_internal_time (&LibJulianDate, &CurrentJulianDate, 0)))
      exit (ErrorVmsStatus (status, "lib$cvt_from_internal_time()", "",
                            __FILE__, __LINE__));
   memset (&NumericTime, 0, sizeof(NumericTime));

   FilterPathLength = strlen(FilterPath);
   LineCount = Expired30Days = Expired60Days =
      Expired90Days = Expired180Days = ExpiredAgesAgo = 0;

   while (VMSok (status = sys$get (&FileRab, 0, 0)))
   {
      Line[FileRab.rab$w_rsz] = '\0';
      if (Debug) fprintf (stdout, "Line |%s|\n", Line);
      LineCount++;
      EntriesInFile++;
      lptr = Line;
      while (isspace(*lptr)) lptr++;

      /* ignore comments */
      if (*lptr == '!' || *lptr == '#') continue;

      /************/
      /* get path */
      /************/

      while (isspace(*lptr)) lptr++;
      if (*lptr != '/') exit (DataFileProblem ("path", LineCount));
      PathPtr = lptr;
      while (*lptr && !isspace(*lptr)) lptr++;
      if (!isspace(*lptr)) exit (DataFileProblem ("path", LineCount));
      PathLength = lptr - PathPtr;

      if (FilterPath[0])
      {
         /******************/
         /* filter on path */
         /******************/

         if (Debug)
            fprintf (stdout, "PathPtr %d |%*.*s| FilterPath %d |%s|\n",
                     PathLength, PathLength, PathLength, PathPtr,
                     FilterPathLength, FilterPath);
         if (PathLength < FilterPathLength) continue;
         cptr = FilterPath;
         pptr = PathPtr;
         while (*cptr && *pptr && tolower(*cptr) == tolower(*pptr))
         {
            cptr++;
            pptr++;
         }
         if (*cptr) continue;
         if (cptr > FilterPath)
         {
            cptr--;
            pptr--;
         }
         if (*cptr != '/' || *pptr != '/') continue;
      }

      /****************/
      /* date and age */
      /****************/

      /* date (dd/mm/yyyy) */
      while (isspace(*lptr)) lptr++;
      if (!isdigit(*lptr)) exit (DataFileProblem ("date", LineCount));
      DayPtr = lptr;
      while (isdigit(*lptr)) lptr++;
      if (*lptr != '/') exit (DataFileProblem ("date", LineCount));
      MonthPtr = ++lptr;
      while (isdigit(*lptr)) lptr++;
      if (*lptr != '/') exit (DataFileProblem ("date", LineCount));
      YearPtr = ++lptr;
      while (isdigit(*lptr)) lptr++;
      if (!isspace(*lptr)) exit (DataFileProblem ("date", LineCount));

      /* get LifeTime (in days) */
      while (isspace(*lptr)) lptr++;
      if (!isdigit(*lptr)) exit (DataFileProblem ("lifetime", LineCount));
      LifeTimePtr = lptr;
      while (isdigit(*lptr)) lptr++;

      Day = atol(DayPtr);
      Month = atol(MonthPtr);
      Year = atol(YearPtr);
      LifeTime = atol(LifeTimePtr);

      if (Year >= 100)
         NumericTime[0] = Year;
      else
      if (Year >= 96 && Year <= 99)
         NumericTime[0] = Year = Year + 1900;
      else
         NumericTime[0] = Year = Year + 2000;
      NumericTime[1] = Month;
      NumericTime[2] = Day;

      if (VMSnok (status =
          lib$cvt_vectim (&NumericTime, &BinaryTime)))
         exit (ErrorVmsStatus (status, "lib$cvt_vectim()", "",
                               __FILE__, __LINE__));

      if (VMSnok (status =
          lib$cvt_from_internal_time (&LibJulianDate, &EntryJulianDate,
                                      &BinaryTime)))
         exit (ErrorVmsStatus (status, "lib$cvt_from_internal_time()", "",
                               __FILE__, __LINE__));

      if (Debug)
         fprintf (stdout, "CurrentJulianDate: %d EntryJulianDate: %d\n",
                  CurrentJulianDate, EntryJulianDate);

      if ((Age = CurrentJulianDate - EntryJulianDate) < 0)
         exit (DataFileProblem ("date", LineCount));

      if (Age > LifeTime)
      {
         if (Age > 0 && Age <= 30)
            Expired30Days++;
         else
         if (Age > 30 && Age <= 60)
            Expired60Days++;
         else
         if (Age > 60 && Age <= 90)
            Expired90Days++;
         else
         if (Age > 90 && Age <= 180)
            Expired180Days++;
         else
            ExpiredAgesAgo++;
      }

      if (FilterAge)
      {
         /*****************/
         /* filter on age */
         /*****************/

         /* only those less than the specified age */
         if (FilterAge > 0 && Age > FilterAge)
            continue;

         /* only those greater than the specified age */
         if (FilterAge < 0 && Age <= -FilterAge)
            continue;
      }
      else
      {
         /*****************************/
         /* filter on whether expired */
         /*****************************/

         if (Age > LifeTime) continue;
      }

      /**********************************/
      /* user name, status, description */
      /**********************************/

      /* get authenticated remote user name used to add the entry */
      while (isspace(*lptr)) lptr++;
      RemoteUserPtr = lptr;
      while (*lptr && !isspace(*lptr)) lptr++;
      RemoteUserLength = lptr - RemoteUserPtr;

      /* status: 'M'nemonic (reminder) 'N'ew, 'R'evised, 'U'pdated */
      while (isspace(*lptr)) lptr++;
      StatusChar = toupper(*lptr);
      if (StatusChar != 'M' &&
          StatusChar != 'N' &&
          StatusChar != 'R' &&
          StatusChar != 'U')
          exit (DataFileProblem ("status", LineCount));
      while (*lptr && !isspace(*lptr)) lptr++;
      while (isspace(*lptr)) lptr++;

      /* get description */
      while (isspace(*lptr)) lptr++;
      if (!*lptr || *lptr == '\n')
         exit (DataFileProblem ("description", LineCount));
      DescriptionPtr = lptr;
      while (*lptr && *lptr != '\n') lptr++;
      DescriptionLength = lptr - DescriptionPtr;

      /**********************/
      /* add to linked-list */
      /**********************/

      Length = sizeof(struct SnewStruct) +
               PathLength + 1 +
               DescriptionLength + 1 +
               RemoteUserLength + 1;
      if ((SnewPtr = calloc (1, Length)) == NULL)
         exit (ErrorVmsStatus (vaxc$errno, "calloc()", "",
                               __FILE__, __LINE__));

      if (ListHeadPtr == NULL)
      {
         SnewPtr->PrevPtr = SnewPtr->NextPtr = NULL;
         ListHeadPtr = SnewPtr;
      }
      else
      {
         SnewPtr->PrevPtr = ListTailPtr;
         ListTailPtr->NextPtr = SnewPtr;
      }
      ListTailPtr = SnewPtr;

      SnewPtr->PathPtr = (char*)SnewPtr + sizeof(struct SnewStruct);
      SnewPtr->DescriptionPtr = SnewPtr->PathPtr + PathLength + 1;
      SnewPtr->RemoteUserPtr = SnewPtr->DescriptionPtr + DescriptionLength + 1;
      memcpy (SnewPtr->PathPtr, PathPtr, PathLength);
      SnewPtr->PathPtr[PathLength] = '\0';
      memcpy (SnewPtr->DescriptionPtr, DescriptionPtr, DescriptionLength);
      SnewPtr->DescriptionPtr[DescriptionLength] = '\0';
      memcpy (SnewPtr->RemoteUserPtr, RemoteUserPtr, RemoteUserLength);
      SnewPtr->RemoteUserPtr[RemoteUserLength] = '\0';

      SnewPtr->StatusChar = StatusChar;

      SnewPtr->Day = Day;
      SnewPtr->Month = Month;
      SnewPtr->Year = Year;
      SnewPtr->LifeTime = atol(LifeTimePtr);
      SnewPtr->Age = Age;

      EntriesInList++;

      if (Debug)
         fprintf (stdout,
"%d. SnewPtr: %d PrevPtr: %d NextPtr: %d \
Day: %d Month: %d Year: %d LifeTime: %d Age: %d\n\
PathPtr |%s|\n\
DescPtr |%s|\n\
RemoPtr |%s|\n",
            EntriesInList,
            SnewPtr, SnewPtr->PrevPtr, SnewPtr->NextPtr,
            SnewPtr->Day, SnewPtr->Month, SnewPtr->Year,
            SnewPtr->LifeTime, SnewPtr->Age,
            SnewPtr->PathPtr, SnewPtr->DescriptionPtr, SnewPtr->RemoteUserPtr);
   }

   sys$close (&FileFab, 0, 0);

   if (VMSnok (status) && status != RMS$_EOF)
      exit (ErrorVmsStatus (status, "database", DbFileName,
            __FILE__, __LINE__));
}

/*****************************************************************************/
/*
When raeding the data file report any formatting problem within it.  This
program acts as a CGI script and also as a command-line utility so ensure any
error messages are reported appropriately.
*/

DataFileProblem
(
char *Explanation,
int LineNumber
)
{
   char  String [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DataFileProblem()\n");

   if (CgiScript)
   {
      sprintf (String, "Database format problem with \"%s\" at line %d.",
               Explanation, LineNumber);                                  
      ErrorGeneral (String, __FILE__, __LINE__);
   }
   else
   {
      fprintf (stdout, "%%%s-E-DATABASE, format \"%s\" at line %d.",
               Utility, Explanation, LineNumber);                                  
      exit (STS$K_ERROR | STS$M_INHIB_MSG);
   }
}

/*****************************************************************************/
/*
Append an entry to the data file.
*/

AddToDataFile ()

{
   register char  *cptr;

   int  status;
   unsigned long  BinaryTime [2];
   unsigned short  NumericTime [7];
   char  Description [256],
         DirectoryUrl [256],
         Line [256];
   struct FAB  FileFab;
   struct RAB  FileRab;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AddToDataFile()\n");

   if (!CgiRemoteUserPtr[0])
      exit (ErrorGeneral (
"Server authentication is not enabled, or path not access-controlled.",
            __FILE__, __LINE__));

   if ((CgiPostDescriptionPtr = ReadHttpInput ("BDESCRIPTION")) == NULL)
      exit (ErrorGeneral ("Description not included.", __FILE__, __LINE__));
   if (!CgiPostDescriptionPtr[0])
      exit (ErrorGeneral ("Supply a description.", __FILE__, __LINE__));

   if ((CgiPostLifeTimePtr = ReadHttpInput ("BLIFETIME")) == NULL)
      exit (ErrorGeneral ("Life-time not included.", __FILE__, __LINE__));

   if ((CgiPostStatusPtr = ReadHttpInput ("BSTATUS")) == NULL)
      exit (ErrorGeneral ("Status not included.", __FILE__, __LINE__));

   if (!CgiPathInfoPtr[0] || !CgiPostLifeTimePtr[0] || !CgiPostStatusPtr[0])
      exit (ErrorGeneral ("All required information not supplied.",
                          __FILE__, __LINE__));

   StringCopyToHtml (Description, CgiPostDescriptionPtr, sizeof(Description),
                     __FILE__, __LINE__);

   sys$gettim (&BinaryTime);
   sys$numtim (&NumericTime, &BinaryTime);

   FileFab = cc$rms_fab;
   FileFab.fab$b_fac = FAB$M_PUT;
   FileFab.fab$l_fna = DbFileName;
   FileFab.fab$b_fns = strlen(DbFileName);
   FileFab.fab$b_shr = FAB$M_SHRGET | FAB$M_SHRPUT;

   if (VMSnok (status = sys$open (&FileFab, 0, 0)))
      exit (ErrorVmsStatus (status, "database", DbFileName,
            __FILE__, __LINE__));

   FileRab = cc$rms_rab;
   FileRab.rab$l_fab = &FileFab;
   FileRab.rab$b_mbf = 2;
   FileRab.rab$l_rop = RAB$M_WBH | RAB$M_EOF;

   if (VMSnok (status = sys$connect (&FileRab, 0, 0)))
      exit (ErrorVmsStatus (status, "database", DbFileName,
            __FILE__, __LINE__));

   FileRab.rab$w_rsz = sprintf (FileRab.rab$l_rbf = Line,
      "%s %d/%d/%d %s %s %c %s",
      CgiPathInfoPtr, 
      NumericTime[2], NumericTime[1], NumericTime[0],
      CgiPostLifeTimePtr,
      CgiRemoteUserPtr,
      toupper(CgiPostStatusPtr[0]),
      Description);

   status = sys$put (&FileRab, 0, 0);

   sys$close (&FileFab, 0, 0);

   if (VMSnok (status))
      exit (ErrorVmsStatus (status, "database", DbFileName,
            __FILE__, __LINE__));

   StringCopy (DirectoryUrl, CgiPathInfoPtr, sizeof(DirectoryUrl),
               __FILE__, __LINE__);
   for (cptr = DirectoryUrl; *cptr; cptr++);
   while (cptr > DirectoryUrl && *cptr != '/') cptr--;
   if (*cptr == '/') *++cptr = '\0';

   fprintf (HttpOut,
"%s\
<!-- SoftwareID: %s -->\n\
<H1>SUCCESS!</H1>\n\
<P>Report generated by server.\n\
<P>Entry for <A HREF=\"%s%s\">%s</A> added.\n",
      Http200Header, SoftwareID,
      CgiScriptNamePtr, DirectoryUrl, CgiPathInfoPtr);
   HttpHasBeenOutput = true;
}

/*****************************************************************************/
/*
Read all entries from the data file.  No path is specified so none are
filtered out, but all expired entries older than the 'PurgeAge' parameter will
not be in the list.  Then open a new version of the data file and write all
entries into that.
*/

PurgeDataFile (int PurgeAge)

{
   int  status;
   unsigned long  BinaryTime [2];
   unsigned short  NumericTime [7];
   char  ExpandedFileName [256],
         Line [256];
   struct SnewStruct  *llptr;
   struct FAB  FileFab;
   struct NAM  FileNam;
   struct RAB  FileRab;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "PurgeDataFile()\n");

   ReadDataFile ("", PurgeAge);

   SortDataList ();

   FileFab = cc$rms_fab;
   FileFab.fab$b_fac = FAB$M_PUT;
   FileFab.fab$l_fna = DbFileName;
   FileFab.fab$b_fns = strlen(DbFileName);
   FileFab.fab$l_fop = FAB$M_SUP;
   FileFab.fab$l_nam = &FileNam;
   FileFab.fab$b_shr = FAB$M_SHRGET | FAB$M_SHRPUT;
   FileFab.fab$b_rat = FAB$M_CR;
   FileFab.fab$b_rfm = FAB$C_VAR;

   FileNam = cc$rms_nam;
   FileNam.nam$l_esa = ExpandedFileName;
   FileNam.nam$b_ess = sizeof(ExpandedFileName)-1;

   if (VMSnok (status = sys$create (&FileFab, 0, 0)))
      exit (ErrorVmsStatus (status, "database", DbFileName,
            __FILE__, __LINE__));

   FileRab = cc$rms_rab;
   FileRab.rab$l_fab = &FileFab;
   FileRab.rab$b_mbf = 2;
   FileRab.rab$l_rop = RAB$M_WBH;

   if (VMSnok (status = sys$connect (&FileRab, 0, 0)))
      exit (ErrorVmsStatus (status, "database", DbFileName,
            __FILE__, __LINE__));

   sys$gettim (&BinaryTime);
   sys$numtim (&NumericTime, &BinaryTime);

   FileRab.rab$w_rsz = sprintf (FileRab.rab$l_rbf = Line,
      "!Created %d/%d/%d %02.02d:%02.02d by %s, %d entries.",
      NumericTime[2], NumericTime[1], NumericTime[0],
      NumericTime[3], NumericTime[4], SoftwareID,
      EntriesInList);
                    
   if (VMSnok (status = sys$put (&FileRab, 0, 0)))
   {
      sys$close (&FileFab, 0, 0);
      if (VMSnok (status))
         exit (ErrorVmsStatus (status, "database", DbFileName,
               __FILE__, __LINE__));
   }

   llptr = ListHeadPtr;
   while (llptr != NULL)
   {
      FileRab.rab$w_rsz = sprintf (FileRab.rab$l_rbf = Line,
         "%s %d/%d/%d %d %s %c %s",
         llptr->PathPtr, 
         llptr->Day, llptr->Month, llptr->Year,
         llptr->LifeTime,
         llptr->RemoteUserPtr,
         llptr->StatusChar,
         llptr->DescriptionPtr);

      if (VMSnok (status = sys$put (&FileRab, 0, 0)))
      {
         sys$close (&FileFab, 0, 0);
         if (VMSnok (status))
            exit (ErrorVmsStatus (status, "database", DbFileName,
                  __FILE__, __LINE__));
      }

      llptr = llptr->NextPtr;
   }

   FileRab.rab$w_rsz = sprintf (FileRab.rab$l_rbf = Line,
      "!New entries follow.");

   if (VMSnok (status = sys$put (&FileRab, 0, 0)))
   {
      sys$close (&FileFab, 0, 0);
      if (VMSnok (status))
         exit (ErrorVmsStatus (status, "database", DbFileName,
               __FILE__, __LINE__));
   }

   sys$close (&FileFab, 0, 0);
}

/*****************************************************************************/
/*
Cruddy old bubble sort!  Organise all entries in the linked list into order
by increasing age and within equal ages by increasing first character of
description.
*/

SortDataList ()

{
   register struct SnewStruct  *llptr1, *llptr2, *llptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "\nSortDataList()\n");

   if (Debug) DebugDataList ();

   llptr1 = ListHeadPtr;
   while (llptr1 != NULL)
   {
      llptr2 = llptr1->NextPtr;
      while (llptr2 != NULL)
      {
         if (Debug)
         {
            fprintf (stdout,
"\n\
llptr1: %d PrevPtr: %d NextPtr: %d \
Day: %d Month: %d Year: %d LifeTime: %d Age: %d\n\
PathPtr |%s|\n\
DescPtr |%s|\n\
llptr2: %d PrevPtr: %d NextPtr: %d \
Day: %d Month: %d Year: %d LifeTime: %d Age: %d\n\
PathPtr |%s|\n\
DescPtr |%s|\n",
               llptr1, llptr1->PrevPtr, llptr1->NextPtr,
               llptr1->Day, llptr1->Month, llptr1->Year,
               llptr1->LifeTime, llptr1->Age,
               llptr1->PathPtr, llptr1->DescriptionPtr,
               llptr2, llptr2->PrevPtr, llptr2->NextPtr,
               llptr2->Day, llptr2->Month, llptr2->Year,
               llptr2->LifeTime, llptr2->Age,
               llptr2->PathPtr, llptr2->DescriptionPtr);
         }

         if (llptr1->Age > llptr2->Age ||
             (llptr1->Age == llptr2->Age &&
              tolower(*llptr1->DescriptionPtr) >
              tolower(*llptr2->DescriptionPtr)))
         {
            if (Debug) fprintf (stdout, "\nSWAPPING!\n");

            llptr = llptr2->PrevPtr;
            llptr2->PrevPtr = llptr1->PrevPtr;
            llptr1->PrevPtr = llptr;
            if (llptr1->PrevPtr != NULL)
               llptr1->PrevPtr->NextPtr = llptr1;
            if (llptr2->PrevPtr != NULL)
               llptr2->PrevPtr->NextPtr = llptr2;

            llptr = llptr2->NextPtr;
            llptr2->NextPtr = llptr1->NextPtr;
            llptr1->NextPtr = llptr;
            if (llptr1->NextPtr != NULL)
               llptr1->NextPtr->PrevPtr = llptr1;
            if (llptr2->NextPtr != NULL)
               llptr2->NextPtr->PrevPtr = llptr2;

            if (llptr1 == ListHeadPtr)
               ListHeadPtr = llptr2;
            else
            if (llptr2 == ListHeadPtr)
               ListHeadPtr = llptr1;

            if (llptr1 == ListTailPtr)
               ListTailPtr = llptr2;
            else
            if (llptr2 == ListTailPtr)
               ListTailPtr = llptr1;

            llptr = llptr1;
            llptr1 = llptr2;
            llptr2 = llptr;
         }
         else
            llptr2 = llptr2->NextPtr;
      }

      llptr1 = llptr1->NextPtr;
   }

   if (Debug) DebugDataList ();
}

/*****************************************************************************/
/*
List database entries.
*/

DebugDataList ()

{
   register struct SnewStruct  *llptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "\nDebugDataList()\n");

   llptr = ListHeadPtr;
   while (llptr != NULL)
   {
       fprintf (stdout,
"\n\
llptr: %d PrevPtr: %d NextPtr: %d \
Day: %d Month: %d Year: %d LifeTime: %d Age: %d\n\
PathPtr |%s|\n\
DescPtr |%s|\n",
          llptr, llptr->PrevPtr, llptr->NextPtr,
          llptr->Day, llptr->Month, llptr->Year,
          llptr->LifeTime, llptr->Age,
          llptr->PathPtr, llptr->DescriptionPtr);
       llptr = llptr->NextPtr;
   }
}

/*****************************************************************************/
/*
The menu allows files and directories to be selected for adding "what's new"
entries against.
*/

SnewMenu ()

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SnewMenu()\n");

   fprintf (HttpOut,
"%s\
<!-- SoftwareID: %s -->\n\
<HTML>\n\
<TITLE>What's New ... select from %s</TITLE>\n\
<H1>\n\
What's New <A HREF=\"%s\">%s</A>\n\
</H1>\n\
<H2>\n\
Select from <A HREF=\"%s\">%s</A>\n\
</H2>\n\
<HR SIZE=5>\n",
      Http200Header, SoftwareID,
      CgiPathInfoPtr,
      HelpUrl, HelpButton,
      CgiPathInfoPtr, CgiPathInfoPtr,
      CgiScriptNamePtr, CgiPathInfoPtr);
   HttpHasBeenOutput = true;

   SelectDirectory (CgiPathInfoPtr, CgiPathTranslatedPtr);

   SelectFile (CgiPathInfoPtr, CgiPathTranslatedPtr);

   fprintf (HttpOut,
"<HR SIZE=5>\n\
</HTML>\n");
}

/****************************************************************************/
/*
In an HTML <SELECT>..<OPTION>..</SELECT> construct list, the names of all the
".DIR" files in the specified directory.
*/ 

SelectDirectory
(
char *DirectoryUrl,
char *Directory
)
{
   register char  *cptr;

   int  status;
   char  FileName [256],
         ExpandedFileSpec [256];
   struct FAB  SearchFileFab;
   struct NAM  SearchFileNam;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SelectDirectory() |%s|\n", Directory);

   /* initialize the file access block */
   SearchFileFab = cc$rms_fab;
   SearchFileFab.fab$l_dna = Directory;
   SearchFileFab.fab$b_dns = strlen(Directory);
   SearchFileFab.fab$l_fna = "*.DIR;0";
   SearchFileFab.fab$b_fns = 7;
   SearchFileFab.fab$l_fop = FAB$V_NAM;
   SearchFileFab.fab$l_nam = &SearchFileNam;

   /* initialize the file name block */
   SearchFileNam = cc$rms_nam;
   SearchFileNam.nam$l_esa = ExpandedFileSpec;
   SearchFileNam.nam$b_ess = sizeof(ExpandedFileSpec)-1;
   SearchFileNam.nam$l_rsa = FileName;
   SearchFileNam.nam$b_rss = sizeof(FileName)-1;

   if (VMSnok (status = sys$parse (&SearchFileFab, 0, 0)))
      exit (ErrorVmsStatus (status, DirectoryUrl, Directory,
                            __FILE__, __LINE__));

   fprintf (HttpOut,
"<FORM METHOD=GET ACTION=\"%s\">\n\
<INPUT TYPE=hidden NAME=parent VALUE=\"%s\">\n\
<INPUT TYPE=submit VALUE=\" Subdirectory ... \">\n\
<SELECT SIZE=%d NAME=directory>\n\
<OPTION VALUE=\".\" SELECTED>(current)\n",
      CgiScriptNamePtr, CgiPathInfoPtr, FormListSize);

   while (VMSok (status = sys$search (&SearchFileFab, 0, 0)))
   {
      SearchFileNam.nam$l_ver[SearchFileNam.nam$b_ver] = '\0';
      if (Debug) fprintf (stdout, "FileName |%s|\n", FileName);

      *SearchFileNam.nam$l_type = '\0';
      for (cptr = SearchFileNam.nam$l_name; *cptr; cptr++)
         *cptr = tolower(*cptr);
      fprintf (HttpOut, "<OPTION VALUE=\"%s\">%s\n",
               SearchFileNam.nam$l_name, SearchFileNam.nam$l_name);
      for (cptr = SearchFileNam.nam$l_name; *cptr; cptr++)
         *cptr = toupper(*cptr);
      *SearchFileNam.nam$l_type = '.';
   }

   if (VMSnok (status) && status != RMS$_FNF && status != RMS$_NMF)
      exit (ErrorVmsStatus (status, DirectoryUrl, ExpandedFileSpec,
            __FILE__, __LINE__));

   fprintf (HttpOut,
"</SELECT>\n\
<INPUT TYPE=radio NAME=do VALUE=entry CHECKED>add entry\n\
<INPUT TYPE=radio NAME=do VALUE=cd>goto\n\
<INPUT TYPE=radio NAME=do VALUE=list>list\n\
<A HREF=\"%s#AddDirectoryEntry\">%s</A>\n\
</FORM>\n",
      HelpUrl, HelpButton);
}

/****************************************************************************/
/*
In an HTML <SELECT>..<OPTION>..</SELECT> construct, list the names of all the
non-".DIR" files in the specified directory.
*/ 

SelectFile
(
char *DirectoryUrl,
char *Directory
)
{
   register char  *cptr;

   int  status,
        FileCount;
   char  FileName [256],
         ExpandedFileSpec [256];
   struct FAB  SearchFileFab;
   struct NAM  SearchFileNam;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SelectFile() |%s|\n", Directory);

   /* initialize the file access block */
   SearchFileFab = cc$rms_fab;
   SearchFileFab.fab$l_dna = Directory;
   SearchFileFab.fab$b_dns = strlen(Directory);
   SearchFileFab.fab$l_fna = "*.*;0";
   SearchFileFab.fab$b_fns = 5;
   SearchFileFab.fab$l_fop = FAB$V_NAM;
   SearchFileFab.fab$l_nam = &SearchFileNam;

   /* initialize the file name block */
   SearchFileNam = cc$rms_nam;
   SearchFileNam.nam$l_esa = ExpandedFileSpec;
   SearchFileNam.nam$b_ess = sizeof(ExpandedFileSpec)-1;
   SearchFileNam.nam$l_rsa = FileName;
   SearchFileNam.nam$b_rss = sizeof(FileName)-1;

   if (VMSnok (status = sys$parse (&SearchFileFab, 0, 0)))
      exit (ErrorVmsStatus (status, Directory, Directory, __FILE__, __LINE__));

   FileCount = 0;

   while (VMSok (status = sys$search (&SearchFileFab, 0, 0)))
   {
      SearchFileNam.nam$l_ver[SearchFileNam.nam$b_ver] = '\0';
      if (Debug) fprintf (stdout, "FileName |%s|\n", FileName);

      if (!strncmp (SearchFileNam.nam$l_type, ".DIR;", 5)) continue;

      if (!FileCount++)
      {
          fprintf (HttpOut,
"<FORM METHOD=GET ACTION=\"%s\">\n\
<INPUT TYPE=hidden NAME=do VALUE=entry>\n\
<INPUT TYPE=hidden NAME=parent VALUE=\"%s\">\n\
<INPUT TYPE=submit VALUE=\" Document ... \">\n\
<SELECT SIZE=%d NAME=file>\n",
             CgiScriptNamePtr, CgiPathInfoPtr, FormListSize);
      }

      *SearchFileNam.nam$l_ver = '\0';
      for (cptr = SearchFileNam.nam$l_name; *cptr; cptr++)
         *cptr = tolower(*cptr);
      fprintf (HttpOut, "<OPTION VALUE=\"%s\">%s\n",
               SearchFileNam.nam$l_name, SearchFileNam.nam$l_name);
      for (cptr = SearchFileNam.nam$l_name; *cptr; cptr++)
         *cptr = toupper(*cptr);
      *SearchFileNam.nam$l_ver = ';';
   }

   if (VMSnok (status) && status != RMS$_FNF && status != RMS$_NMF)
      exit (ErrorVmsStatus (status, Directory, ExpandedFileSpec,
            __FILE__, __LINE__));

   if (FileCount)
   {
      fprintf (HttpOut,
"</SELECT>\n\
<A HREF=\"%s#Document\">%s</A>\n\
</FORM>\n",
         HelpUrl, HelpButton,
         HelpUrl, HelpButton);
   }
}

/*****************************************************************************/
/*
Provide a form allowing the user to input a description, select the lifetime
for the entry and then submit to add the entry to the data file.
*/

SnewEntry ()

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SnewEntry()\n");

   if (!CgiPathInfoPtr[0])
      exit (ErrorGeneral ("Supply a file or directory path.",
                          __FILE__, __LINE__));

   fprintf (HttpOut,
"%s\
<!-- SoftwareID: %s -->\n\
<HTML>\n\
<TITLE>What's New ... entry for %s</TITLE>\n\
<H1>\n\
What's New <A HREF=\"%s\">%s</A>\n\
</H1>\n\
<H2>\n\
Entry for <A HREF=\"%s\">%s</A>\n\
</H2>\n\
<HR SIZE=5>\n\
<FORM METHOD=POST ACTION=\"%s%s\">\n\
<INPUT TYPE=hidden NAME=do VALUE=add>\n\
<NOBR><INPUT TYPE=text NAME=description SIZE=60> <I>(description\
<A HREF=\"%s#EntryDescription\">%s</A>)</I></NOBR>\n\
<BR><INPUT TYPE=radio NAME=status VALUE=new CHECKED>new\n\
<INPUT TYPE=radio NAME=status VALUE=updated>updated\n\
<INPUT TYPE=radio NAME=status VALUE=revised>revised\n\
<INPUT TYPE=radio NAME=status VALUE=mnemonic>reminder\n\
<A HREF=\"%s#EntryType\">%s</A>\n\
<BR>Removed after\n\
<SELECT NAME=LifeTime>\n\
<OPTION VALUE=7>7 days\n\
<OPTION VALUE=14 SELECTED>14 days\n\
<OPTION VALUE=28>1 month\n\
<OPTION VALUE=90>3 months\n\
<OPTION VALUE=180>6 months\n\
<OPTION VALUE=365>1 year\n\
<OPTION VALUE=99999>never\n\
</SELECT>\n\
<A HREF=\"%s#EntryRemovedAfter\">%s</A>\n\
<P><INPUT TYPE=submit VALUE=\" add entry \">\n\
<INPUT TYPE=reset VALUE=\" reset \">\n\
<A HREF=\"%s#EntryCheck\">%s</A>\n\
</FORM>\n\
<P><HR SIZE=3>\n\
</HTML>\n",
      Http200Header, SoftwareID,
      CgiPathInfoPtr,
      HelpUrl, HelpButton,
      CgiPathInfoPtr, CgiPathInfoPtr,
      CgiScriptNamePtr, CgiPathInfoPtr,
      HelpUrl, HelpButton,
      HelpUrl, HelpButton,
      HelpUrl, HelpButton,
      HelpUrl, HelpButton);
   HttpHasBeenOutput = true;
}

/****************************************************************************/
/*
Copies 'SrcPtr' string to 'DstPtr' with destination string overflow error
reporting.  Returns number of characters copied.
*/ 

int StringCopy
(
char *DstPtr,
char *SrcPtr,
int SizeOfDst,
char *SourceFileName,
int SourceLineNumber
)
{
   register char  *cptr, *sptr, *zptr;

   zptr = (sptr = DstPtr) + SizeOfDst;
   for (cptr = SrcPtr; *cptr && sptr < zptr; *sptr++ = *cptr++);
   if (sptr > zptr)
      exit (ErrorGeneral ("String too small.",
                          SourceFileName, SourceLineNumber));
   *sptr = '\0';
   return (sptr - DstPtr);
}

/*****************************************************************************/
/*
Copy string from 'cptr' to 'sptr', escaping forbidden HTML characters.  A 
maximum of 'scnt'-1 characters are copied, with the resulting string always 
terminated.  The count of the number of characters copied is returned 
(excluding the terminating null).
*/ 

int StringCopyToHtml
(
register char *sptr,
register char *cptr,
register int scnt,
char *SourceFileName,
int SourceLineNumber
)
{
   register char  *StartPtr;

   if (scnt <= 0)
      exit (ErrorGeneral ("String too small.",
                          SourceFileName, SourceLineNumber));
   StartPtr = sptr;
   while (*cptr && scnt--)
   {
      switch (*cptr)
      {
         case '\"' :
            if (scnt >= 6)
            {
               *sptr++ = '&'; *sptr++ = 'q'; *sptr++ = 'u'; *sptr++ = 'o'; 
               *sptr++ = 't'; *sptr++ = ';';
               scnt -= 6;
               cptr++;
            }
            else
               scnt = 0;
            break;
         case '<' :
            if (scnt >= 4)
            {
               *sptr++ = '&'; *sptr++ = 'l'; *sptr++ = 't'; *sptr++ = ';'; 
               scnt -= 4;
               cptr++;
            }
            else
               scnt = 0;
            break;
         case '>' :
            if (scnt >= 4)
            {
               *sptr++ = '&'; *sptr++ = 'g'; *sptr++ = 't'; *sptr++ = ';'; 
               scnt -= 4;
               cptr++;
            }
            else
               scnt = 0;
            break;
         case '&' :
            if (scnt >= 5)
            {
               *sptr++ = '&'; *sptr++ = 'a'; *sptr++ = 'm'; *sptr++ = 'p'; 
               *sptr++ = ';';
               scnt -= 5;
               cptr++;
            }
            else
               scnt = 0;
            break;
         default :
            *sptr++ = *cptr++;
            scnt--;
      }
   }
   if (scnt <= 0)
      exit (ErrorGeneral ("String too small.",
                          SourceFileName, SourceLineNumber));
   *sptr = '\0';
   return (sptr - StartPtr);
}

/*****************************************************************************/
/*
Copy text from one string to another, converting characters forbidden to 
appear as plain-text in an HTTP URL back into plain-text.  For example the 
'?', '+', '&', etc., are converted from "%nn" hexadecimal escaped characters. 
Returns number of characters copied.
*/ 

int StringCopyFromUrl
(
char *DstPtr,
char *SrcPtr,
int SizeOfDst,
char *SourceFileName,
int SourceLineNumber
)
{
   register char  c;
   register char  *cptr, *sptr, *zptr;

   zptr = (sptr = DstPtr) + SizeOfDst;

   while (*cptr && sptr < zptr)
   {
      if (*cptr == '%')
      {
         /* an escaped character ("%xx" where xx is a hexadecimal number) */
         cptr++;
         if (*cptr) cptr++;
         if (*cptr) cptr++;
         c = *cptr;
         *cptr = '\0';
         *sptr++ = (unsigned char) strtol (cptr-2, NULL, 16);
         *cptr = c;
      }
      else
         *sptr++ = *cptr++;
   }
   if (sptr > zptr)
      exit (ErrorGeneral ("String too small.",
                          SourceFileName, SourceLineNumber));
   *sptr = '\0';

   return (sptr - DstPtr);
}

/*****************************************************************************/
/*
Copy text from one string to another, converting characters forbidden to 
appear as plain-text text in an HTTP URL.  For example the '?', '+', '&', etc.  
Convert these to "%nn" hexadecimal escaped characters.  Checks for string
overflow (after the event unfortunately, but only a small chance of inelegant
application crash!)
*/ 

int CopyStringToUrl
( 
char *DstPtr,
char *SrcPtr,
int  SizeOfDst,
char *SourceFileName,
int SourceLineNumber
)
{
   register char  *dptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   sptr = SrcPtr;
   zptr = (dptr = DstPtr) + SizeOfDst;
   while (*sptr && dptr < zptr)
   {  
      switch (*sptr)
      {
         case '\t' : memcpy (dptr, "%09", 3); dptr += 3; sptr++; break;
         case ' ' : memcpy (dptr, "%20", 3); dptr += 3; sptr++; break;
         case '!' : memcpy (dptr, "%21", 3); dptr += 3; sptr++; break;
         case '\"' : memcpy (dptr, "%22", 3); dptr += 3; sptr++; break;
         case '#' : memcpy (dptr, "%23", 3); dptr += 3; sptr++; break;
         case '$' : memcpy (dptr, "%24", 3); dptr += 3; sptr++; break;
         case '%' : memcpy (dptr, "%25", 3); dptr += 3; sptr++; break;
         case '&' : memcpy (dptr, "%26", 3); dptr += 3; sptr++; break;
         case '\'' : memcpy (dptr, "%27", 3); dptr += 3; sptr++; break;
         case '(' : memcpy (dptr, "%28", 3); dptr += 3; sptr++; break;
         case ')' : memcpy (dptr, "%29", 3); dptr += 3; sptr++; break;
         case '*' : memcpy (dptr, "%2a", 3); dptr += 3; sptr++; break;
         case '+' : memcpy (dptr, "%2b", 3); dptr += 3; sptr++; break;
         case ',' : memcpy (dptr, "%2c", 3); dptr += 3; sptr++; break;
         case '-' : memcpy (dptr, "%2d", 3); dptr += 3; sptr++; break;
         case '.' : memcpy (dptr, "%2e", 3); dptr += 3; sptr++; break;
         case '/' : memcpy (dptr, "%2f", 3); dptr += 3; sptr++; break;
         case ':' : memcpy (dptr, "%3a", 3); dptr += 3; sptr++; break;
         case ';' : memcpy (dptr, "%3b", 3); dptr += 3; sptr++; break;
         case '<' : memcpy (dptr, "%3c", 3); dptr += 3; sptr++; break;
         case '=' : memcpy (dptr, "%3d", 3); dptr += 3; sptr++; break;
         case '>' : memcpy (dptr, "%3e", 3); dptr += 3; sptr++; break;
         case '?' : memcpy (dptr, "%3f", 3); dptr += 3; sptr++; break;
         case '[' : memcpy (dptr, "%5b", 3); dptr += 3; sptr++; break;
         case '\\' : memcpy (dptr, "%5c", 3); dptr += 3; sptr++; break;
         case ']' : memcpy (dptr, "%5d", 3); dptr += 3; sptr++; break;
         case '^' : memcpy (dptr, "%5e", 3); dptr += 3; sptr++; break;
         case '_' : memcpy (dptr, "%5f", 3); dptr += 3; sptr++; break;
         case '{' : memcpy (dptr, "%7b", 3); dptr += 3; sptr++; break;
         case '|' : memcpy (dptr, "%7c", 3); dptr += 3; sptr++; break;
         case '}' : memcpy (dptr, "%7d", 3); dptr += 3; sptr++; break;
         case '~' : memcpy (dptr, "%7e", 3); dptr += 3; sptr++; break;
         default : if (isprint(*sptr)) *dptr++ = *sptr++; else sptr++;
      }
   }
   if (dptr >= zptr)
      exit (ErrorGeneral ("String too small.",
                          SourceFileName, SourceLineNumber));

   *dptr = '\0';
   return (dptr-DstPtr);
}

/*****************************************************************************/
/*
Useful, self-contained routine to read the HTTP stream and then search and/or
return certain portions of it.

If a field name is supplied the in-memory stream is searched (case-insensitive)
for that field name. The values are copied to malloc()ed memory and the pointer
returned. If not found a NULL pointer is returned.

If the HTTP method is not POST or the body is not
"application/x-www-form-urlencoded" a NULL pointer is returned.

The field name "*" returns a pointer the the complete, undecoded body.
This pointer should not be free()ed unless the routines will not be called
again!

Successive calls using an empty string "" returns each body field, decoded.  A
NULL is again returned when the last line has been read.  After any complete
sequence the entire sequence may be read again.  Field name searches can be
interspursed with the successive reads.

All field memory pointers returned can subsequently be free()ed as desired.
*/

char *ReadHttpInput (char *FieldNamePtr)

{
   boolean  FormUrlEncoded;
   static int  ContentLength,
               HttpBodyLength;
   static char  *ContentTypePtr,
                *HttpBodyPtr,
                *HttpBodyNextPtr;

   register char  c;
   register char  *cptr, *sptr, *zptr;

   int  ContentCount,
        FieldNameLength,
        LineLength;
   char  *FieldValuePtr;
   char  HttpInputLine [2048];
   FILE  *HttpInput;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ReadHttpInput()\n");

   if (HttpBodyPtr == NULL)
   {
      if ((HttpInput = fopen ("HTTP$INPUT", "r")) == NULL)
         return (vaxc$errno);

      /* these two global storage must exist! */
      ContentTypePtr = CgiContentTypePtr;
      ContentLength = atoi(CgiContentLengthPtr);

      if (strsame (ContentTypePtr, "application/x-www-form-urlencoded"))
         FormUrlEncoded = true;
      else
         FormUrlEncoded = false;

      HttpBodyPtr = NULL;
      HttpBodyLength = 0;

      /**********************/
      /* read the HTTP body */
      /**********************/

      ContentCount = 0;
      while (ContentCount < ContentLength)
      {
         if (fgets (HttpInputLine, sizeof(HttpInputLine), HttpInput) == NULL)
            break;
         LineLength = strlen(HttpInputLine);
         if ((HttpBodyPtr =
             realloc (HttpBodyPtr, HttpBodyLength+LineLength+1)) == NULL)
         exit (vaxc$errno);
         memcpy (HttpBodyPtr+HttpBodyLength, HttpInputLine, LineLength+1);
         HttpBodyLength += LineLength;

         ContentCount += LineLength;
         if (Debug)
            fprintf (stdout, "%d/%d/%d |%s|\n",
                     LineLength, ContentCount, ContentLength, HttpInputLine);
      }

      HttpBodyNextPtr = HttpBodyPtr;

      if (Debug) fprintf (stdout, "HttpBodyPtr |%s|\n", HttpBodyPtr);

      fclose (HttpInput);
   }

   if (FieldNamePtr == NULL) return (NULL);
   if (!strcmp (FieldNamePtr, "*")) return (HttpBodyPtr);

   /**************/
   /* find field */
   /**************/

   if (!FormUrlEncoded) return (NULL);

   if (Debug) fprintf (stdout, "FieldNamePtr |%s|\n", FieldNamePtr);

   FieldNameLength = strlen(FieldNamePtr)-1;
   sptr = NULL;

   if (*FieldNamePtr)
      cptr = HttpBodyPtr;
   else
   if (!*(cptr = HttpBodyNextPtr))
   {
      cptr = HttpBodyNextPtr = HttpBodyPtr;
      return (NULL);
   }

   while (*cptr)
   {
      if (FieldNamePtr[1] &&
          (cptr[FieldNameLength] != '=' ||
           !strnsame (cptr, FieldNamePtr+1, FieldNameLength)))
      {
         /* look for the next field name */
         while (*cptr && *cptr != '&') cptr++;
         if (*cptr) cptr++;
         continue;
      }  

      /* found the field (or an empty fieldname) */
      if (FieldNamePtr[1]) cptr += FieldNameLength + 1;
      sptr = cptr;
      while (*sptr && *sptr != '&') sptr++;
      if (!FieldNamePtr[1])
         if (*sptr)
            HttpBodyNextPtr = sptr + 1;
         else
            HttpBodyNextPtr = sptr;
      break;
   }

   /* return if we didn't find it */
   if (sptr == NULL) return (sptr);

   if ((FieldValuePtr = malloc (sptr-cptr+1)) == NULL)
      exit (vaxc$errno);
   memcpy (FieldValuePtr, cptr, sptr - cptr);
   FieldValuePtr[sptr-cptr] = '\0';
   if (Debug) fprintf (stdout, "FieldValuePtr |%s|\n", FieldValuePtr);

   /* no need to decode a header field */
   if (toupper(FieldNamePtr[0]) == 'H') return (FieldValuePtr);

   /*********************/
   /* decode body field */
   /*********************/

   cptr = sptr = FieldValuePtr;
   while (*cptr)
   {
      if (*cptr == '\r' || *cptr == '\n')
      {
         cptr++;
         continue;
      }
      if (*cptr == '+')
      {
         cptr++;
         *sptr++ = ' ';
         continue;
      }
      if (*cptr == '%')
      {
         cptr++;
         while (*cptr && (*cptr == '\r' || *cptr == '\n')) cptr++;
         c = 0;
         if (*cptr >= '0' && *cptr <= '9')
            c = (*cptr - (int)'0') << 4;
         else
         if (toupper(*cptr) >= 'A' && toupper(*cptr) <= 'F')
            c = (toupper(*cptr) - (int)'A' + 10) << 4;
         else
         {
            free (FieldValuePtr);
            return (NULL);
         }
         cptr++;
         while (*cptr && (*cptr == '\r' || *cptr == '\n')) cptr++;
         if (*cptr >= '0' && *cptr <= '9')
            c += (*cptr - (int)'0');
         else
         if (toupper(*cptr) >= 'A' && toupper(*cptr) <= 'F')
            c += (toupper(*cptr) - (int)'A' + 10);
         else
         {
            free (FieldValuePtr);
            return (NULL);
         }
         if (*cptr) cptr++;
         *sptr++ = c;
         continue;
      }

      /* else non-encoded character */
      *sptr++ = *cptr++;
   }
   *sptr = '\0';

   if (Debug) fprintf (stdout, "FieldValuePtr |%s|\n", FieldValuePtr);

   return (FieldValuePtr);
}

/*****************************************************************************/
/*
*/

ErrorGeneral
(
char *Text,
char *SourceFileName,
int SourceLineNumber
)
{
   register char  *cptr;

   /*********/
   /* begin */
   /*********/

   /* 
      The source file format provided by the "__FILE__" macro will be
      "device:[directory]name.type;ver".  Reduce that to "name.type".
   */
   for (cptr = SourceFileName; *cptr && *cptr != ';'; cptr++);
   if (*cptr)
   {
      while (*cptr != '.') cptr--;
      *cptr-- = '\0';
   }
   while (*cptr != ']') cptr--;
   cptr++;

   if (!CgiScript)
   {
      fprintf (stdout,
"%%%s-E-WHERE, module: %s line: %d\n\
-%s-E-WHAT, %s",
               Utility, cptr, SourceLineNumber, Utility, Text);
      exit (STS$K_ERROR | STS$M_INHIB_MSG);
   }

   if (!HttpHasBeenOutput) fputs (Http404Header, HttpOut);
   fprintf (HttpOut,
"<!-- SoftwareID: %s Module: %s Line: %d -->\n\
<H1>ERROR!</H1>\n\
<P>Reported by server.\n\
<P>%s\n",
   SoftwareID, cptr, SourceLineNumber, Text);

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
*/

ErrorVmsStatus
(
int StatusValue,
char *Text,
char *HiddenText,
char *SourceFileName,
int SourceLineNumber
)
{
   static char  Message [256];
   static $DESCRIPTOR (MessageDsc, Message);

   register char  *cptr;
   int  status;
   short int  Length;

   /*********/
   /* begin */
   /*********/

   if (VMSok (status = sys$getmsg (StatusValue, &Length, &MessageDsc, 1, 0))) 
   {
      Message[Length] = '\0';
      Message[0] = toupper(Message[0]);
   }
   else
      exit (status);

   /* 
      The source file format provided by the "__FILE__" macro will be
      "device:[directory]name.type;ver".  Reduce that to "name.type".
   */
   for (cptr = SourceFileName; *cptr && *cptr != ';'; cptr++);
   if (*cptr)
   {
      while (*cptr != '.') cptr--;
      *cptr-- = '\0';
   }
   while (*cptr != ']') cptr--;
   cptr++;

   if (!CgiScript)
   {
      fprintf (stdout, "%%%s-E-WHERE, module: %s line: %d\n",
               Utility, cptr, SourceLineNumber);
      exit (StatusValue);
   }

   if (!HttpHasBeenOutput) fputs (Http404Header, HttpOut);
   fprintf (HttpOut,
"<!-- SoftwareID: %s Module: %s Line: %d -->\n\
<H1>ERROR!</H1>\n\
<P>Reported by server.\n\
<P>%s ... <TT>%s</TT>\n\
<!-- %%X%08.08X \"%s\" -->\n",
   SoftwareID, cptr, SourceLineNumber, Message, Text, StatusValue, HiddenText);

   return (SS$_NORMAL);
}

/****************************************************************************/
/*
Does a case-insensitive, character-by-character string compare and returns 
true if two strings are the same, or false if not.  If a maximum number of 
characters are specified only those will be compared, if the entire strings 
should be compared then specify the number of characters as 0.
*/ 

boolean strnsame
(
register char *sptr1,
register char *sptr2,
register int  count
)
{
   while (*sptr1 && *sptr2)
   {
      if (toupper (*sptr1++) != toupper (*sptr2++)) return (false);
      if (count)
         if (!--count) return (true);
   }
   if (*sptr1 || *sptr2)
      return (false);
   else
      return (true);
}

/****************************************************************************/
