/*****************************************************************************/
/*
                                Spinneret.c

Given a URL for an HTML document, will retrieve all or selected components 
comprising that document, i.e. a Web "spider" or "crawler".

The import message is a message placed at the top of each HTML page retrieved.
The import link is a URL pointing to some local page explaining the document's
imported status, etc.  If the import message and link are to be explicitly
specified then the message must include a $FAO "!AZ" directive to include the
link URL.  Import messages and links may be suppressed by explcilty specifying
them as empty strings or with a leading space (to allow the definition of such
"empty" logicals).

All files from a single "crawl" are placed in the one directory.  That is,
directory hierarchies at the original source are not reproduced.  To help
ensure uniqueness of files generated the file naming scheme derives the file
name from the original "path".  This path may be shortened to the VMS file name
limit of 39 characters (by removing characters from the start of the path). 
The length of the original path is prepended as two digits (leading zero if
necessary) leaving a possible 37 characters in the file name, plus any
dot-separated file type considered separately.

The /EXERCISE qualifier allows this utility to be used as a server exerciser,
traversing the specified path, getting all documents, but not actually creating
any files from them.  Use something like:

  $ SPINNERET /EXERCISE /ALL "http://full.host.name:port/"


LOGICAL NAMES
-------------
SPINNERET_ALWAYS_GET      same as /ALWAYS_GET qualifier
SPINNERET_EXTERNAL_LINKS  if logical defined get external links
SPINNERET_INPORT_LINK     same as /IMPORT_LINK qualifier
SPINNERET_IMPORT_MESSAGE  same as /IMPORT_MESSAGE qualifier
SPINNERET_PROXY           same as /PROXY qualifier


QUALIFIERS and PARAMETERS
-------------------------
/[NO]ALL                all subdocuments of the original specified
/ALWAYS_GET=            comma-separated list of file types (e.g. "GIF,JPG")
/DBUG                   turns on all "if (Debug)" statements
/DEPTH=                 integer, depth of subdocuments to be retrieved
/EXCLUDE=               comma-separated list of file types (e.g. "TXT,HTML")
/[NO]EXERCISE           do not create nay files, exercise the server
/[NO]EXTERNAL_LINKS     get external links (from other than root host)
/IMPORT_LINK=           link pointed to by import message (URL)
/IMPORT_MESSAGE=        message at top of HTML page stating document imported
/INCLUDE=               comma-separated list of file types (e.g. "TXT,HTML")
/PROXY=                 proxy server ("domain.host[:port]", defaults to 8080)
/RANGE=                 integer, distance of documents to be retrieved
/[NO]VERBOSE            output lots of status message (NO default)


BUILD DETAILS
-------------
See BUILD_SPINNERET.COM procedure.


VERSION HISTORY
---------------
07-DEC-97  MGD  v1.3.0, server exerciser,
                        bugfix; non-proxy request sys$fao()
20-NOV-96  MGD  v1.2.0, much overdue maintenance and refinement
23-APR-96  MGD  v1.1.0, add proxy server capability (after intro. of same)
28-AUG-94  MGD  v1.0.0, initial development
*/
/*****************************************************************************/

#ifdef __ALPHA
char SoftwareID [] = "SPINNERET AXP-1.3.0";
#else
char SoftwareID [] = "SPINNERET VAX-1.3.0";
#endif

/* standard C header files */
#include <stdio.h>
#include <ctype.h>
#include <time.h>
#include <errno.h>
#include <types.h>
#include <file.h>

/* VMS-related header files */
#include <brkdef.h>
#include <descrip.h>
#include <iodef.h>
#include <jpidef.h>
#include <lnmdef.h>
#include <prcdef.h>
#include <psldef.h>
#include <ssdef.h>
#include <stsdef.h>
#include <syidef.h>
#include <rms.h>
#include <unixio.h>

/* Internet-related header files */

#ifdef UCX_TCPIP

#include <socket.h>
#include <in.h>
#include <netdb.h>
#include <inet.h>
#include <ucx$inetdef.h>

#define IP_errno vaxc$errno

#define IP_close_socket close

#else

#include "multinet_root:[multinet.include.sys]types.h"
#include "multinet_root:[multinet.include.sys]socket.h"
#include "multinet_root:[multinet.include.netinet]in.h"
#include "multinet_root:[multinet.include.vms]inetiodef.h"
#include "multinet_root:[multinet.include]errno.h"
#include "multinet_root:[multinet.include]netdb.h"

#define IP_errno vmserrno

#define IP_close_socket socket_close

#endif

#define boolean int
#define true 1
#define false 0

#define VMSok(x) ((x) & STS$M_SUCCESS)
#define VMSnok(x) (!((x) & STS$M_SUCCESS))
#define VMSwarning(x) (((x) & 0x7) == STS$M_WARNING)
#define VMSerror(x) (((x) & 0x7) == STS$M_ERROR)
#define VMSinfo(x) (((x) & 0x7) == STS$M_INFO)
#define VMSfatal(x) (((x) & 0x7) == STS$M_SEVERR)

#define HttpPort 80
#define ProxyHttpPort 8080

#define ConnectFailRetries 3

#define ImportLinkDefault "/web/$IMPORTED$.HTML"
#define ImportMessageDefault \
   "<P><I>[<A HREF=\"!AZ\">Imported</A> from the Internet WWW]</I><P>"

struct SourceUrlStruct
{
   boolean  HrefEncountered,
            InsideAnchorTag,
            InsideFrameTag,
            InsideImgTag,
            IsAnHtmlFile,
            IsATextFile,
            LinkObtained,
            PartialLink;

   int  BufferCount,
        ClientSocket,
        ServerPort,
        ServerSocket,
        SubdocumentDepth;

   unsigned short  ClientChannel;

   char  *BufferPtr;

   char  FileType [256],
         HostName [256],
         Protocol [256],
         VmsFileName [256],
         Uri [1024],
         Url [1024];

   struct FAB  FileFab;
   struct RAB  FileRab;
   struct NAM  FileNam;

   struct hostent  RemoteHostEntry;
   struct hostent  *RemoteHostEntryPtr;
   struct sockaddr_in  SocketName;
};

/*
   This data structure is a bit shonky!  The 'Url' member is defined
   as only having a single character but in fact will be as many as
   required to accomodate the URL.  It is done this way so it can be
   accessed as string.  The structure is dynamically allocated using
   "malloc (strlen (Url) + 1 + sizeof (struct NoteUrlStruct*))".
*/
struct NoteUrlStruct
{
   struct NoteUrlStruct  *NextPtr;
   char  Url [1];
};

char Utility [] = "SPINNERET";

/* fall-back, if the file type can't be determined from the "Content-Type:" */
char HtmlTypeList [] = "HTML,HTM";
char TextTypeList [] = "TEXT,TXT,ASCII,CONF,LOG,PS";

boolean  AllRootSubdocuments,
         Debug,
         DoConfirm,
         DoShowHelp,
         ExerciseOnly,
         GetExternalLinks,
         Verbose;

int  Depth,
     DocumentRange,
     ProxyServerPort,
     Range;

char  AlwaysGetTheseFileTypes [256],
      CommandLine [256],
      ExcludeFileTypes [256],
      ImportLink [256],
      ImportMessage [256],
      IncludeFileTypes [256],
      ProxyServer [256],
      RootHostName [256],
      RootUri [256],
      UrlParameter [256];

struct NoteUrlStruct  *ObtainedHead = NULL,
                      *ObtainedTail = NULL;

struct NoteUrlStruct  *DoNotGetHead = NULL,
                      *DoNotGetTail = NULL;

/* required prototypes */
char* SysGetMsg (int);

/*****************************************************************************/
/*
*/

int main ()

{
   register char  *cptr, *sptr;

   int  status;
   char  ScratchUri [256];
   struct SourceUrlStruct  Source;

   if (VMSnok (status = ParseCommandLine ()))
      exit (status);

   if (DoShowHelp) exit (ShowHelp ());

   if (!ProxyServer[0])
   {
      cptr = getenv ("SPINNERET_EXTERNAL");
      if (cptr != NULL) GetExternalLinks = true;
   }

   if (!AlwaysGetTheseFileTypes[0])
   {
      cptr = getenv ("SPINNERET_ALWAYS_GET");
      if (cptr != NULL) strcpy (AlwaysGetTheseFileTypes, cptr);
   }

   if (!ImportLink[0])
   {
      cptr = getenv ("SPINNERET_IMPORT_LINK");
      if (cptr != NULL) strcpy (ImportLink, cptr);
   }
   if (!ImportLink[0]) strcpy (ImportLink, ImportLinkDefault);
   if (ImportLink[0] == ' ') ImportLink[0] = '\0';

   if (!ImportMessage[0])
   {
      cptr = getenv ("SPINNERET_IMPORT_MESSAGE");
      if (cptr != NULL) strcpy (ImportMessage, cptr);
   }
   if (!ImportMessage[0]) strcpy (ImportMessage, ImportMessageDefault);
   if (ImportMessage[0] == ' ') ImportMessage[0] = '\0';

   if (!ProxyServer[0])
   {
      cptr = getenv ("SPINNERET_PROXY");
      if (cptr != NULL) strcpy (ProxyServer, cptr);
   }

   if (ProxyServer[0])
   {
      for (cptr = ProxyServer; *cptr && *cptr != ':'; cptr++)
      if (*cptr == ':')
      {
         *cptr++ = '\0';
         ProxyServerPort = atol(cptr);
      }
      if (!ProxyServerPort) ProxyServerPort = ProxyHttpPort;
   }

   ParseUrl (UrlParameter, Source.Protocol,
             Source.HostName, &Source.ServerPort, ScratchUri);
   strcpy (RootHostName, Source.HostName);
   if (Debug) fprintf (stdout, "RootHostName |%s|\n", RootHostName);

   memset (&Source, 0, sizeof(struct SourceUrlStruct));

   ProcessUrl (&Source, UrlParameter, ScratchUri);

   exit (status);
}

/*****************************************************************************/
/*
The URL represents a link.  Parse the URL into its component parts.  Make a 
decision (or ask the user) if this link should be retrieved.  Recursively 
retrieve the link.
*/ 

ProcessUrl
(
struct SourceUrlStruct *suptr,
char *Url,
char *NewUrl
)
{
   static boolean  GetNoMore = false;
   static char  LinkReply [16] = "";
   static $DESCRIPTOR (SourceUrlFaoDsc, "!AZ//!AZ:!UL!AZ");

   register char  *cptr;

   int  status,
        UrlLength;
   unsigned short  Length;
   char  Uri [256];
   struct SourceUrlStruct  Source;
   $DESCRIPTOR (SourceUrlDsc, Source.Url);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ProcessUrl() |%s|\n", Url);

   if (ImportLink[0] && strsame (Url, ImportLink, -1)) return;

   memset (&Source, 0, sizeof(struct SourceUrlStruct));

   ParseUrl (Url, Source.Protocol, Source.HostName, &Source.ServerPort, Uri);
   if (NewUrl == NULL) strcpy (RootHostName, Source.HostName);
   if (Debug) fprintf (stdout, "RootHostName |%s|\n", RootHostName);

   /* copy any non-supplied components from the current source */
   if (!Source.Protocol[0]) strcpy (Source.Protocol, suptr->Protocol);
   if (!Source.HostName[0]) strcpy (Source.HostName, suptr->HostName);
   if (!Source.ServerPort) Source.ServerPort = suptr->ServerPort;
   if (!Source.ServerPort) Source.ServerPort = HttpPort;

   if (ParseUri (Uri, suptr->Uri, Source.Uri))
   {
      /* absolute */
      Source.PartialLink = false;
      Source.SubdocumentDepth = 0;
      DocumentRange++;
   }
   else
   {
      /* relative (partial) */
      Source.PartialLink = true;
      Source.SubdocumentDepth = suptr->SubdocumentDepth + 1;
   }
   if (!RootUri[0]) GetRootUri (Source.Uri);

   if (IgnoreThis (Source.Uri)) return;

   status = sys$fao (&SourceUrlFaoDsc, &Length, &SourceUrlDsc,
                     Source.Protocol,
                     Source.HostName, Source.ServerPort,
                     Source.Uri);
   if (VMSnok (status)) exit (status);
   Source.Url[Length] = '\0';

   VmsFileNameFromUri (Source.Uri, Source.VmsFileName, Source.FileType);
   if (!Source.VmsFileName[0]) VmsFileNameFromHostName (&Source);
   if (NewUrl != NULL) strcpy (NewUrl, Source.VmsFileName);
   for (cptr = Source.VmsFileName; *cptr; cptr++) *cptr = toupper(*cptr);

   if (GetThisUrl (&Source))
   {
      /************************************************/
      /* check if this link has already been obtained */
      /************************************************/

      if (AlreadyObtained (&Source))
      {
         if (Verbose)
            fprintf (stdout, "[%d.%d][Already obtained: \"%s\"]\n",
                     DocumentRange, Source.SubdocumentDepth, Source.Url);

         suptr->LinkObtained = true;
         if (!Source.PartialLink) DocumentRange--;
         return;
      }

      if (DoConfirm)
      {
         /**********************************/
         /* check if this link is required */
         /**********************************/

         /* an 'A' in the link reply means "get all" */
         if (LinkReply[0] != 'A')
         {
            if (GetNoMore)
               LinkReply[0] = 'N';
            else
            {
               fprintf (stdout,
                  "[%d.%d] Get \"%s\"? (Yes,No,Quit,All,noMore) [Y]: ",
                  DocumentRange, Source.SubdocumentDepth, Source.Url);
               fgets (LinkReply, sizeof(LinkReply), stdin);
               LinkReply[0] = toupper(LinkReply[0]);
            }
            if (LinkReply[0] == 'Q') exit (SS$_NORMAL);
            if (LinkReply[0] == 'M') GetNoMore = true;
            if (!(LinkReply[0] == '\n' ||
                  LinkReply[0] == 'Y' ||
                  LinkReply[0] == 'A'))
            {
               UpdateDoNotGet (&Source);
               if (!Source.PartialLink) DocumentRange--;
               return;
            }
         }
      }

      /**********************************/
      /* recursively process the source */
      /**********************************/

      while (!(status = RetrieveUrl (&Source)));
      if (status > 0) suptr->LinkObtained = true;
      if (!Source.PartialLink) DocumentRange--;
      if (DocumentRange)
         fprintf (stdout, "[%d.%d][Continuing: \"%s\"]\n",
                  DocumentRange, suptr->SubdocumentDepth, suptr->VmsFileName);

      return;
   }
   else
   {
      if (!Source.PartialLink) DocumentRange--;
      return;
   }
}

/*****************************************************************************/
/*
Decide whether this link should be retrieved.  Reasons for not include having 
already obtained it, specifically excluded file type, not a spcifically 
included file type, not HTTP URL.  If /CONFIRM was used then leave it up to 
the user to have the final say.

Return true if it should be retrieved, false if not.
*/ 

boolean GetThisUrl (struct SourceUrlStruct *suptr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "GetThisUrl()\n");

   if (!GetExternalLinks && !strsame (suptr->HostName, RootHostName, -1))
   {
      if (Verbose)
         fprintf (stdout, "[%d.%d][External host: \"%s\"]\n",
                  DocumentRange, suptr->SubdocumentDepth, suptr->Url);
      return (false);
   }

   if (!strsame (suptr->Protocol, "http:", -1))
   {
      if (Verbose)
         fprintf (stdout, "[%d.%d][Not HTTP: \"%s\"]\n",
                  DocumentRange, suptr->SubdocumentDepth, suptr->Url);
      return (false);
   }

   if (DoNotGet (suptr))
   {
      if (Verbose)
         fprintf (stdout, "[%d.%d][Do Not Get: \"%s\"]\n",
                  DocumentRange, suptr->SubdocumentDepth, suptr->Url);
      return (false);
   }

   /*
      If this file type is not always retrieved regardless
      then check if this type should be retrieved this time!
   */

   if (!AlwaysGetTheseFileTypes[0] ||
       !InCommaList (suptr->FileType+1, AlwaysGetTheseFileTypes))
   {
      /***************************************/
      /* check if this file type is included */
      /***************************************/

      if (IncludeFileTypes[0] &&
          !InCommaList (suptr->FileType+1, IncludeFileTypes))
      {
         if (Verbose)
            fprintf (stdout, "[%d.%d][Not included: \"%s\"]\n",
                     DocumentRange, suptr->SubdocumentDepth, suptr->Url);
         return (false);
      }

      /***************************************/
      /* check if this file type is excluded */
      /***************************************/

      if (ExcludeFileTypes[0] &&
          InCommaList (suptr->FileType+1, ExcludeFileTypes))
      {
         if (Verbose)
            fprintf (stdout, "[%d.%d][Excluded: \"%s\"]\n",
                     DocumentRange, suptr->SubdocumentDepth, suptr->Url);
         return (false);
      }

      /*
         Checking just the URI without the host etc., is not a foolproof
         method for establishing whether this is part of a document hierarchy.
         However it is practical, as some servers direct the client to a
         different host (name, possibly) for the same hierarchy.
      */
      if (!AllRootSubdocuments || !SameRootUri (suptr->Uri))
      {
         if (suptr->PartialLink)
         {
            /**************************************************/
            /* check if this link is within subdocument depth */
            /**************************************************/
      
            if (suptr->SubdocumentDepth > Depth)
            {
               if (Verbose)
                  fprintf (stdout, "[%d.%d][Too deep: \"%s\"]\n",
                           DocumentRange, suptr->SubdocumentDepth, suptr->Url);
               return (false);
            }
            else;
         }
         else
         {
            /***************************************/
            /* check if this link is within range */
            /***************************************/

            if (DocumentRange > Range)
            {
               if (Verbose)
                  fprintf (stdout, "[%d.%d][Too far: \"%s\"]\n",
                           DocumentRange, suptr->SubdocumentDepth, suptr->Url);
               return (false);
            }
         }
      }
   }

   return (true);
}

/*****************************************************************************/
/*
This overly-long function is recursively called to process each URL.  It 
connects with the server, opens the output file, retrieves each network record 
from the server, if a "binary" file writes the records to the output file in 
block mode, if a "text" file breaks that record up into lines and writes each 
line/record to the output file.

Then if its an HTML file rewinds and looks through each line for links to 
other documents which it accesses and if required retrieves recursively.
*/ 

int RetrieveUrl (struct SourceUrlStruct *suptr)

{
   static $DESCRIPTOR (NonProxyFaoDsc, "!AZ !AZ HTTP/1.0\r\n\r\n");
   static $DESCRIPTOR (ProxyFaoDsc, "!AZ !AZ//!AZ:!UL!AZ HTTP/1.0\r\n\r\n");

   register int  cnt;
   register char  *cptr, *lptr, *rptr, *sptr;

   boolean  HttpHeader;
   int  status,
        ByteCount,
        ConnectFailCount,
        ContentLength,
        LineCount,
        RecordCount,
        RecordLength,
        StatusCode;
   unsigned short  Length;
   char  *ContentTypePtr,
         *HttpMethodPtr;
   char  ErrorReply [16],
         Line [1024],
         Record [2048];
   $DESCRIPTOR (LineDsc, Line);

   /*********/
   /* begin */
   /*********/

   fprintf (stdout, "[%d.%d][Retrieving: \"%s\"]\n",
            DocumentRange, suptr->SubdocumentDepth, suptr->Url);

   /*********************/
   /* connect to server */
   /*********************/

   for (ConnectFailCount = ConnectFailRetries; ConnectFailCount--;)
   {
      if (ProxyServer[0])
         fprintf (stdout, "[%d.%d][Connecting: %s:%d (proxy: %s:%d)]\n",
                  DocumentRange, suptr->SubdocumentDepth,
                  suptr->HostName, suptr->ServerPort,
                  ProxyServer, ProxyServerPort);
      else
         fprintf (stdout, "[%d.%d][Connecting: %s:%d]\n",
                  DocumentRange, suptr->SubdocumentDepth,
                  suptr->HostName, suptr->ServerPort);

      if (VMSok (status = ConnectToServer (suptr))) break;

      if (!status)
      {
         fprintf (stdout, "%%%s-E-CONNECT, failed\n", Utility);
         status = STS$K_ERROR;
      }
      else
         fprintf (stdout, "%%%s-E-CONNECT, failed\n-%s\n",
                  Utility, SysGetMsg(status)+1);
      sleep (5);
   }
   if (!ConnectFailCount) return (-1);

   HttpMethodPtr = "GET";

   /* send request to server */
   if (ProxyServer[0])
      status = sys$fao (&ProxyFaoDsc, &Length, &LineDsc,
                        suptr->Protocol, HttpMethodPtr, suptr->HostName,
                        suptr->ServerPort, suptr->Uri);
   else
      status = sys$fao (&NonProxyFaoDsc, &Length, &LineDsc,
                        HttpMethodPtr, suptr->Uri);
   if (VMSnok (status)) exit (status);
   Line[Length] = '\0';
   WriteData (suptr, Line, Length);

   /*********************************/
   /* loop reading data from source */
   /*********************************/

   fprintf (stdout, "\r[%d.%d][Received: 0 bytes]",
            DocumentRange, suptr->SubdocumentDepth);

   if (suptr->BufferPtr != NULL) exit (STS$K_ERROR);

   ByteCount = RecordCount = ContentLength = suptr->BufferCount = 0;
   HttpHeader = true;

   while (VMSok (status =
          ReadData (suptr, Record, sizeof(Record), &RecordLength)))
   {
      if (!RecordLength) break;

      if (Debug)
         fprintf (stdout, "realloc() %d + %d\n",
                  suptr->BufferCount, RecordLength);

      if ((suptr->BufferPtr =
           realloc (suptr->BufferPtr, suptr->BufferCount+RecordLength+1))
           == NULL)
         exit (vaxc$errno);

      memcpy (suptr->BufferPtr+suptr->BufferCount, Record, RecordLength);
      suptr->BufferCount += RecordLength;
      ByteCount += RecordLength;

      if (HttpHeader)
      {
         suptr->BufferPtr[suptr->BufferCount] = '\0';         
         for (rptr = suptr->BufferPtr; *rptr && HttpHeader; rptr++)
         {
            if (*rptr != '\n') continue;
            if (rptr[1] == '\n' || (rptr[1] == '\r' && rptr[2] == '\n'))
            {
               HttpHeader = false;
               rptr++;
               if (*rptr == '\r') rptr++;
               rptr++;
               ByteCount -= rptr - suptr->BufferPtr;
               /* find the content-length in the header */
               ContentLength = 0;
               for (rptr = suptr->BufferPtr; *rptr; rptr++)
               {
                   if (toupper (*rptr) != 'C') continue;
                   if (!strsame (rptr, "Content-Length:", 15)) continue;
                   if (Debug) fprintf (stdout, "Content-Length: |%s|\n", rptr);
                   rptr += 15;
                   while (*rptr && isspace(*rptr)) rptr++;
                   if (!isdigit(*rptr)) break;
                   ContentLength = atol (rptr);
               }
            }
         }         
      }

      if (ContentLength)
         fprintf (stdout, "\r[%d.%d][Received: %d of %d bytes (%d%%)]",
                  DocumentRange, suptr->SubdocumentDepth,
                  ByteCount, ContentLength, ByteCount * 100 / ContentLength);
      else
         fprintf (stdout, "\r[%d.%d][Received: %d bytes]",
                  DocumentRange, suptr->SubdocumentDepth, ByteCount);
   }

   /****************************/
   /* end get source data loop */
   /****************************/

   if (ContentLength)
      fprintf (stdout, "\r[%d.%d][Received: %d of %d bytes (%d%%)]\n",
               DocumentRange, suptr->SubdocumentDepth,
               ByteCount, ContentLength, ByteCount * 100 / ContentLength);
   else
      fprintf (stdout, "\r[%d.%d][Received: %d bytes]\n",
               DocumentRange, suptr->SubdocumentDepth, ByteCount);

   /**************************/
   /* disconnect from server */
   /**************************/

   DisconnectFromServer (suptr);

   rptr = suptr->BufferPtr;
   cnt = suptr->BufferCount;

   HttpHeader = true;
   LineCount = 0;
   ContentTypePtr = "";

   while (cnt)
   {
      /**********************/
      /* process the header */
      /**********************/

      if (Debug) fprintf (stdout, "cnt: %d rptr: %d\n", cnt, rptr);
      sptr = rptr;
      while (cnt)
      {
         cnt--;
         if (rptr[0] == '\r' && rptr[1] == '\n')
            *rptr = '\0';
         else
         if (*rptr == '\n')
         {
            *rptr++ = '\0';
            break;
         }
         rptr++;
      }

      LineCount++;
      if (Debug) fprintf (stdout, "sptr |%s|\n", sptr); 
      if (!*sptr)
      {
         HttpHeader = false;
         break;
      }

      if (LineCount == 1)
      {
         for (cptr = sptr; *cptr && !isspace(*cptr); cptr++);
         while (isspace(*cptr)) cptr++;
         StatusCode = atol(cptr);
         if (*cptr != '2' || ByteCount < ContentLength)
         {
            if (ExerciseOnly)
            {
               if (ByteCount < ContentLength)
                  fprintf (stdout,
                           "%%%s-W-INTERRUPTED, transfer interrupted?%c\n",
                           Utility, 7);
               else
                  fprintf (stdout, "%%%s-W-RESPONSE, %d \n \\%s\\\n",
                           Utility, StatusCode, sptr);
               return (-1);
            }

            if (ByteCount < ContentLength)
               fprintf (stdout, "%%%s-W-INTERRUPTED, transfer interrupted?%c\n",
                        Utility, 7);
            else
               fprintf (stdout, "%%%s-W-RESPONSE, %d error\n \\%s\\\n%c",
                        Utility, StatusCode, sptr, 7);
            fprintf (stdout, "Retry? (Yes,No,Quit,Ignore) [Y]: ");
            fgets (ErrorReply, sizeof(ErrorReply), stdin);
            ErrorReply[0] = toupper(ErrorReply[0]);
            if (ErrorReply[0] == 'Q') exit (SS$_NORMAL);
            free (suptr->BufferPtr);
            suptr->BufferPtr = NULL;
            if (ErrorReply[0] == 'N') return (-1);
            if (ErrorReply[0] != 'I') return (0);
         }
      }
      else
      {
         if (toupper(sptr[0]) == 'C' &&
             strsame (sptr, "Content-Type:", 13))
         {
            for (cptr = sptr+13; *cptr && isspace(*cptr); cptr++);
            ContentTypePtr = cptr;
            if (Debug)
               fprintf (stdout, "ContentTypePtr |%s|\n", ContentTypePtr);
         }
      }
   }

   UpdateObtained (suptr);

   /********************/
   /* process the body */
   /********************/

   if (strsame (ContentTypePtr, "text/html", -1))
      suptr->IsATextFile = suptr->IsAnHtmlFile = true;
   else
   if (strsame (ContentTypePtr, "text/", 5))
      suptr->IsATextFile = true;
   else
   if (InCommaList (suptr->FileType+1, HtmlTypeList))
      suptr->IsATextFile = suptr->IsAnHtmlFile = true;
   else
   if (InCommaList (suptr->FileType+1, TextTypeList))
      suptr->IsATextFile = true;

   if (suptr->IsAnHtmlFile)
   {
      /**************************/
      /* HTML file (find links) */
      /**************************/

      if (VMSnok (status = CreateFile (suptr)))
      {
         fprintf (stdout, "%%%s-E-CREATE, file create failed\n-%s\n",
                  Utility, SysGetMsg(status)+1);
         exit (status | STS$M_INHIB_MSG);
      }
      if (VMSnok (status = StampHtmlFile (suptr)))
         exit (status);

      while (cnt)
      {
         sptr = NULL;
         while (cnt)
         {
            cnt--;
            if (sptr == NULL) sptr = rptr;
            if (rptr[0] == '\r' && rptr[1] == '\n')
               *rptr = '\0';
            else
            if (*rptr == '\n')
            {
               *rptr++ = '\0';
               break;
            }
            rptr++;
         }

         /* find and recursively process any links in this line */
         if (sptr != NULL) ProcessLine (suptr, sptr);
      }

      if (!ExerciseOnly) sys$close (&suptr->FileFab, 0, 0);
   }
   else
   {
      /*****************/
      /* non-HTML file */
      /*****************/

      if (VMSnok (status = CreateFile (suptr)))
      {
         fprintf (stdout, "%%%s-E-CREATE, file create failed\n-%s\n",
                  Utility, SysGetMsg(status)+1);
         exit (status | STS$M_INHIB_MSG);
      }

      while (cnt)
      {
         if (Debug) fprintf (stdout, "cnt: %d\n", cnt);
         if (cnt > 32767)
         {
            suptr->FileRab.rab$l_rbf = rptr;
            suptr->FileRab.rab$w_rsz = 32767;
            rptr += 32767;
            cnt -= 32767;
            if (ExerciseOnly)
               status = SS$_NORMAL;
            else
               status = sys$put (&suptr->FileRab, 0, 0);
         }
         else
         {
            suptr->FileRab.rab$l_rbf = rptr;
            suptr->FileRab.rab$w_rsz = cnt;
            cnt -= cnt;
            if (ExerciseOnly)
               status = SS$_NORMAL;
            else
               status = sys$put (&suptr->FileRab, 0, 0);
         }
         if (Debug) fprintf (stdout, "sys$put() %%X%08.08X\n", status);
      }

      if (!ExerciseOnly) sys$close (&suptr->FileFab, 0, 0);
   }

   free (suptr->BufferPtr);
   suptr->BufferPtr = NULL;

   return (1);
}

/*****************************************************************************/
/*
The opening tag (e.g. <A) may be on a separate line to the reference parameter 
(e.h. HREF=), but the reference itself cannot be broken across lines (e.g. 
cannot be HREF="/this/is/\r\nthe/reference").  Spaces may exists between the 
parameter, the equate symbol and the parameter value (e.g. HREF = "example").
*/ 

int ProcessLine
(
struct SourceUrlStruct *suptr,
char *SrcLine
)
{
   register int  cnt;
   register char  *cptr, *lptr, *lzptr, *sptr, *zptr;

   int  status;
   char  Line [8192],
         NewUrl [1024],
         Url [1024];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ProcessLine() |%s|\n", SrcLine);

   /*********************/
   /* loop through line */
   /*********************/

   lzptr = (lptr = Line) + sizeof(Line);
   cptr = SrcLine;
   while (*cptr)
   {
      if (suptr->InsideAnchorTag || suptr->InsideFrameTag)
      {
         /*******************/
         /* inside "<A" tag */
         /*******************/

         if (*cptr == '>')
         {
            if (suptr->InsideAnchorTag)
               suptr->InsideAnchorTag = false;
            else
               suptr->InsideFrameTag = false;
            if (lptr < lzptr) *lptr++ = *cptr++;
            continue;
         }
         else
         if ((toupper(cptr[0]) == 'H' &&
              toupper(cptr[1]) == 'R' &&
              toupper(cptr[2]) == 'E' &&
              toupper(cptr[3]) == 'F' &&
              (cptr[4] == '=' || isspace(cptr[4]))) ||
            ((toupper(cptr[0]) == 'S' &&
              toupper(cptr[1]) == 'R' &&
              toupper(cptr[2]) == 'C' &&
              (cptr[3] == '=' || isspace(cptr[3])))))
          {  
            /* get the URL */
            suptr->HrefEncountered = true;
            while (*cptr && *cptr != '=' && *cptr != '>' && lptr < lzptr)
               *lptr++ = *cptr++;
            while (*cptr && *cptr != '>' &&
                   (*cptr == '=' || isspace(*cptr)) && lptr < lzptr)
                *lptr++ = *cptr++;
            if (*cptr == '\"' && lptr < lzptr) *lptr++ = *cptr++;

            zptr = (sptr = Url) + sizeof(Url);
            while (*cptr && !isspace(*cptr) &&
                   *cptr != '\"'  && *cptr != '\'' /* script */ &&
                   *cptr != '#' && *cptr != '>' &&
                   sptr < zptr)
                *sptr++ = *cptr++;
            if (sptr >= zptr) exit ((StringOverflowError (__LINE__)));
            *sptr = '\0';
            if (Debug) fprintf (stdout, "Url |%s|\n", Url);

            if (Url[0])
            {
               ProcessUrl (suptr, Url, NewUrl);
               if (suptr->LinkObtained)
                  sptr = NewUrl;
               else
                  sptr = Url;
               while (*sptr && lptr < lzptr) *lptr++ = *sptr++;
            }
            else
            if (*cptr == '#')
               suptr->LinkObtained = true;

            continue;
         }
         else
         {
            if (lptr < lzptr) *lptr++ = *cptr++;
            continue;
         }
      }

      if (suptr->InsideImgTag)
      {
         /*******************/
         /* inside "<IMG" tag */
         /*******************/

         if (*cptr == '>')
         {
            suptr->InsideImgTag = false;
            if (lptr < lzptr) *lptr++ = *cptr++;
            continue;
         }
         else
         if (toupper(cptr[0]) == 'S' &&
             toupper(cptr[1]) == 'R' &&
             toupper(cptr[2]) == 'C' &&
             (cptr[3] == '=' || isspace(cptr[3])))
         {
            /* get the URL */
            while (*cptr && *cptr != '=' && *cptr != '>' && lptr < lzptr)
               *lptr++ = *cptr++;
            while (*cptr && *cptr != '>' &&
                   (*cptr == '=' || isspace(*cptr)) && lptr < lzptr)
                *lptr++ = *cptr++;

            if (*cptr == '\"' && lptr < lzptr) *lptr++ = *cptr++;
            zptr = (sptr = Url) + sizeof(Url);
            while (*cptr && !isspace(*cptr) &&
                   *cptr != '\"'  && *cptr != '\'' /* script */ &&
                   *cptr != '>' && sptr < zptr)
                *sptr++ = *cptr++;
            if (sptr >= zptr) exit ((StringOverflowError (__LINE__)));
            *sptr = '\0';
            if (Debug) fprintf (stdout, "Url |%s|\n", Url);

            ProcessUrl (suptr, Url, NewUrl);

            if (suptr->LinkObtained)
               sptr = NewUrl;
            else
               sptr = Url;
            while (*sptr && lptr < lzptr) *lptr++ = *sptr++;
            continue;
         }
         else
         {
            if (lptr < lzptr) *lptr++ = *cptr++;
            continue;
         }
      }

      /********************************/
      /* check for tag, ignore if not */
      /********************************/

      if (cptr[0] == '<')
      {
         if (cptr [1] == '/' &&
             toupper(cptr[2]) == 'A' &&
             cptr[3] == '>')
         {
            /***************************/
            /* end anchor tag ("</A>") */
            /***************************/

            /* link not obtained, introduce a caret */
            if (suptr->HrefEncountered && !suptr->LinkObtained &&
                lptr < lzptr) *lptr++ = '^';
            suptr->LinkObtained = suptr->HrefEncountered = false;

            if (lptr < lzptr) *lptr++ = *cptr++;
            if (lptr < lzptr) *lptr++ = *cptr++;
            if (lptr < lzptr) *lptr++ = *cptr++;
            if (lptr < lzptr) *lptr++ = *cptr++;
            continue;
         }
         else
         if (toupper(cptr[1]) == 'A' &&
             (isspace(cptr[2]) || !cptr[2]))
         {
            /**************************/
            /* anchor tag ("<A ...>") */
            /**************************/

            suptr->InsideAnchorTag = true;
            suptr->HrefEncountered = suptr->LinkObtained = false;
            if (lptr < lzptr) *lptr++ = *cptr++;
            if (lptr < lzptr) *lptr++ = *cptr++;
            continue;
         }
         else
         if (toupper(cptr[1]) == 'I' &&
             toupper(cptr[2]) == 'M' &&
             toupper(cptr[3]) == 'G' &&
             (isspace(cptr[4]) || !cptr[4]))
         {
            /***************************/
            /* image tag ("<IMG ...>") */
            /***************************/

            suptr->InsideImgTag = true;
            if (lptr < lzptr) *lptr++ = *cptr++;
            if (lptr < lzptr) *lptr++ = *cptr++;
            if (lptr < lzptr) *lptr++ = *cptr++;
            if (lptr < lzptr) *lptr++ = *cptr++;
            continue;
         }
         else
         if (toupper(cptr[1]) == 'F' &&
             toupper(cptr[2]) == 'R' &&
             toupper(cptr[3]) == 'A' &&
             toupper(cptr[4]) == 'M' &&
             toupper(cptr[5]) == 'E' &&
             (isspace(cptr[6]) || !cptr[6]))
         {
            /*****************************/
            /* frame tag ("<FRAME ...>") */
            /*****************************/

            suptr->InsideFrameTag = true;
            suptr->HrefEncountered = suptr->LinkObtained = false;
            if (lptr < lzptr) *lptr++ = *cptr++;
            if (lptr < lzptr) *lptr++ = *cptr++;
            if (lptr < lzptr) *lptr++ = *cptr++;
            if (lptr < lzptr) *lptr++ = *cptr++;
            if (lptr < lzptr) *lptr++ = *cptr++;
            if (lptr < lzptr) *lptr++ = *cptr++;
            continue;
         }
         else
         {
            if (lptr < lzptr) *lptr++ = *cptr++;
            continue;
         }
      }

      if (lptr < lzptr) *lptr++ = *cptr++;
   }

   if (lptr < lzptr) *lptr = '\0';
   if (Debug) fprintf (stdout, "Line |%s|\n", Line);

   if (lptr >= lzptr) exit ((StringOverflowError (__LINE__)));

   suptr->FileRab.rab$l_rbf = Line;
   suptr->FileRab.rab$w_rsz = lptr - Line;
   if (ExerciseOnly)
      status = SS$_NORMAL;
   else
      status = sys$put (&suptr->FileRab, 0, 0);
   if (Debug) fprintf (stdout, "sys$put() %%X%08.08X\n", status);
   if (VMSnok (status)) exit (status);
}

/*****************************************************************************/
/*
Create a VMS file name from the last part of the URI, making sure it is RMS 
compliant (i.e: URIs often contain Unix style names, containing all sorts of 
characters/combinations, munge these to something acceptable to RMS; e.g: 
"/directory/name_version-2.0.html" to "32$DIRECTORY$NAME_VERSION-2$0.HTML")
*/ 

VmsFileNameFromUri
(
char *Uri,
char *VmsFileName,
char *FileType
)
{
   register char  *cptr, *sptr, *zptr;
   int  UriLength;
   char  BitBucket [256],
         Scratch [256];

   if (Debug) fprintf (stdout, "VmsFileNameFomrUri() |%s|\n", Uri);

   if (VmsFileName == NULL) VmsFileName = BitBucket;
   if (FileType == NULL) FileType = BitBucket;

   if (strsame (Uri, ImportLink, -1))
   {
      strcpy (VmsFileName, Uri);
      strcpy (FileType, ".HTML");
      return;
   }

   /* maximum length of a VMS file NAME is 39 characters (type is also 39!) */
   for (cptr = Uri; *cptr; cptr++);
   if ((UriLength = cptr - Uri) > 255) exit ((StringOverflowError (__LINE__)));
   while (cptr > Uri && *cptr != '.' && *cptr != '/') cptr--;
   if (*cptr == '.') UriLength = cptr - Uri;
   if (UriLength > 37)
      cptr = Uri + UriLength - 37;
   else
      cptr = Uri;

   /* munge the name so it is acceptable to RMS as a file name */
   zptr = NULL;
   sptr = Scratch;
   while (*cptr)
   {
      /* replace multiple periods with dollar symbols */
      if (*cptr == '.')
      {
         zptr = sptr;
         *sptr++ = '$';
      }
      else
      if (*cptr == '/')
      {
         zptr = NULL;
         *sptr++ = '$';
      }
      else
      {
         /* substitute unacceptable characters with dollar symbols */
         if (isalnum(*cptr) || *cptr == '-' || *cptr == '_' || *cptr == '$')
            *sptr++ = *cptr;
         else
            *sptr++ = '$';
      }
      cptr++;
   }
   *sptr = '\0';

   if (zptr == NULL)
   {
      /* if there was no period at all then the file type is HTML */
      if (UriLength > 32)
      {
         /* too big, move it all up to accomodate the ".html" extension */
         *sptr = '\0';
         cptr = Scratch + UriLength - 32;
         sptr = Scratch;
         while (*cptr) *sptr++ = *cptr++;
      }
      strcpy (sptr, ".HTML");
      sptr += 5;
      strcpy (FileType, ".HTML");
   }
   else
   {
      /* replace the last dollar substituted period with a real period */
      *zptr = '.';
      if (FileType != NULL) strcpy (FileType, zptr);
   }

   sprintf (VmsFileName, "%02.02d%s", UriLength, Scratch);

   if (Debug) fprintf (stdout, "VmsFileName |%s|\n", VmsFileName);
}

/*****************************************************************************/
/*
Separates a URL into its component protocol, port, host and URI.  For any 
components not present it either nulls or zeros the respective variable.
*/ 

int ParseUrl
(
char *Url,
char *Protocol,
char *Host,
int *PortPtr,
char *Uri
)
{
   register char  *sptr, *uptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ParseUrl() |%s|->", Url);

   *Protocol = *Host = *Uri = '\0';
   *PortPtr = 0;

   /* get the protocol (e.g. "http:" from "http://host:port/uri") */
   uptr = Url;
   zptr = (sptr = Protocol) + 256;
   while (*uptr && *uptr != ':' && *uptr != '/' && sptr < zptr)
      *sptr++ = *uptr++;
   if (sptr >= zptr) exit ((StringOverflowError (__LINE__)));
   *sptr = '\0';
   if (uptr[0] == ':')
   {
      if (sptr < zptr) *sptr++ = *uptr++;
      if (sptr >= zptr) exit ((StringOverflowError (__LINE__)));
      *sptr = '\0';
   }
   else
   {
      /* no protocol was specifically supplied */
      Protocol[0] = '\0';
      uptr = Url;
   }

   if (uptr[0] == '/' && uptr[1] == '/')
   {
      /* get the host/port (e.g. "host:port" from "http://host:port/uri") */
      uptr += 2;
      zptr = (sptr = Host) + 256;
      while (*uptr && *uptr != '/' && *uptr != ':' && sptr < zptr)
         *sptr++ = *uptr++;
      if (sptr >= zptr) exit ((StringOverflowError (__LINE__)));
      *sptr = '\0';
      if (*uptr == ':')
      {
         /* specific port number was supplied */
         if (Debug) fprintf (stdout, "Port uptr |%s|\n", uptr);
         *PortPtr = atol (uptr+1);
         while (*uptr && *uptr != '/') uptr++;
      }
   }

   /* get the URI (e.g. "uri" from "http://host:port/uri") */
   zptr = (sptr = Uri) + 256;
   while (*uptr && sptr < zptr) *sptr++ = *uptr++;
   if (sptr >= zptr) exit ((StringOverflowError (__LINE__)));
   *sptr = '\0';

   if (Debug)
      fprintf (stdout, "|%s|%s|%d|%s|\n", Protocol, Host, *PortPtr, Uri);
}

/*****************************************************************************/
/*
Create a new URI from a context URI (the current absolute URI) and a supplied 
URI, that can be absoloute or partial.
*/ 

boolean ParseUri
(
register char *Uri,
register char *ContextUri,
register char *NewUri
)
{
   register char  *cptr, *sptr, *uptr, *zptr;
   boolean  AbsoluteLink;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ParseUri() |%s|->", Uri);

   if (Uri[0] == '/')
   {
      if (Uri[1] == '/')
      {
         /***************************/
         /* "consecutive slash" URI */
         /***************************/

         zptr = (sptr = NewUri) + 256;
         uptr = Uri;
         cptr = ContextUri;
         while (uptr[0] == '/' && uptr[1] == '/')
         {
            if (*cptr == '/')
            {
               if (sptr < zptr) *sptr++ = *cptr++;
               while (*cptr && *cptr != '/' && sptr < zptr) *sptr++ = *cptr++;
               if (sptr >= zptr) exit ((StringOverflowError (__LINE__)));
               *sptr = '\0';
            }
            else
               if (sptr < zptr) *sptr++ = *uptr;
            uptr++;
         }
         /* append the parameter URI to new URI */
         while (*uptr && sptr < zptr) *sptr++ = *uptr++;
         if (sptr >= zptr) exit ((StringOverflowError (__LINE__)));
         *sptr = '\0';
         if (Debug) fprintf (stdout, "(consecutive) NewUri |%s|\n", NewUri);
      }
      else
      {
         /****************/
         /* absolute URI */
         /****************/

         zptr = (sptr = NewUri) + 256;
         for (cptr = Uri; *cptr && sptr < zptr; *sptr++ = *cptr++);
         if (sptr >= zptr) exit ((StringOverflowError (__LINE__)));
         *sptr = '\0';
         if (Debug) fprintf (stdout, "(full) NewUri |%s|\n", NewUri);
      }
      AbsoluteLink = true;
   }
   else
   if (Uri[0])
   {
      /***************/
      /* partial URI */
      /***************/

      uptr = NULL;
      zptr = (sptr = NewUri) + 256;
      for (cptr = ContextUri; *cptr && sptr < zptr; *sptr++ = *cptr++)
         if (*cptr == '/') uptr = sptr;
      /* reset the new URI pointer to the last slash from the context URI */
      if (uptr != NULL) sptr = uptr+1;
      /* append the parameter URI to new URI */
      for (cptr = Uri; *cptr && sptr < zptr; *sptr++ = *cptr++);
      if (sptr >= zptr) exit ((StringOverflowError (__LINE__)));
      *sptr = '\0';
      if (Debug) fprintf (stdout, "(partial) NewUri |%s|\n", NewUri);
      AbsoluteLink = false;
   }
   else
   {
      /************/
      /* null URI */
      /************/

      NewUri[0] = '\0';
      AbsoluteLink = true;
   }

   /*****************************************/
   /* process any "../" and "./" constructs */
   /*****************************************/

   uptr = NewUri;
   while (*uptr)
   {
      if (*uptr == '/')
      {
         for (cptr = uptr+1; *cptr && *cptr != '/'; cptr++);
         if (*cptr)
         {
            if (cptr[1] == '.' && cptr[2] == '.' && cptr[3] == '/')
            {
               sptr = cptr + 3;
               cptr = uptr;
               while (*sptr) *cptr++ = *sptr++;
               *cptr = '\0';
               uptr = NewUri;
            }
            else
            if (cptr[1] == '.' && cptr[2] == '/')
            {
               sptr = cptr + 2;
               while (*sptr) *cptr++ = *sptr++;
               *cptr = '\0';
               uptr = NewUri;
            }
         }
      }
      uptr++;
   }

   if (Debug) fprintf (stdout, "NewUri %d |%s|\n", AbsoluteLink, NewUri);
   return (AbsoluteLink);
}

/*****************************************************************************/
/*
Generate a VMS file name using the host name.  Replace the periods between the 
host name components with dollar symbols and append a file type of ".HTML".
*/ 

int VmsFileNameFromHostName (struct SourceUrlStruct *suptr)

{
   register char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "VmsFileNameFromHostName()\n");

   sptr = suptr->VmsFileName;
   for (cptr = suptr->HostName; *cptr; cptr++)
      if (*cptr == '.') *sptr++ = '$'; else *sptr++ = *cptr;
   /* ensure the file name is no longer than 40 characters */
   suptr->VmsFileName[39] = '\0';
   strcat (suptr->VmsFileName, ".html");
   strcpy (suptr->FileType, ".html");

   if (Debug) fprintf (stdout, "|%s|%s|", suptr->HostName, suptr->VmsFileName);
}

/*****************************************************************************/
/*
Removes anything including and following the final slash in a URL.
*/ 

GetRootUri (char *Uri)

{
   register char  *rptr, *uptr;

   if (Debug) fprintf (stdout, "GetRootUri() |%s|\n", Uri);

   rptr = NULL;
   for (uptr = Uri; *uptr; uptr++)
      if (*uptr == '/') rptr = uptr;

   if (rptr == NULL)
      strcpy (RootUri, Uri);
   else
   {
      rptr++;
      strcpy (RootUri, Uri, rptr-Uri);
      RootUri[rptr-Uri] = '\0';
   }

   if (Debug) fprintf (stdout, "RootUri |%s|\n", RootUri);
}

/*****************************************************************************/
/*
*/ 

boolean SameRootUri (char *Uri)

{
   static int  RootUriLength = 0;

   register char  c;

   if (Debug) fprintf (stdout, "SameRootUri |%s|%s", Uri, RootUri);

   if (!RootUriLength) RootUriLength = strlen(RootUri);

   c = Uri[RootUriLength];
   Uri[RootUriLength] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", Uri);

   if (!strcmp (RootUri, Uri))
   {
      Uri[RootUriLength] = c;
      return (true);
   }
   else
   {
      Uri[RootUriLength] = c;
      return (false);
   }
}

/*****************************************************************************/
/*
*/

boolean IgnoreThis (char *Uri)

{
   register char  *cptr, *uptr;

   if (Debug) fprintf (stdout, "IgnoreThis() |%s|\n", Uri);

   cptr = NULL;
   for (uptr = Uri; *uptr; uptr++)
      if (*uptr == '/') cptr = uptr;
   if (cptr == NULL)
      cptr = uptr;
   else
      cptr++;

   if (strsame (cptr, "mailto:", 7)) return (true);
   if (strsame (cptr, "news:", 5)) return (true);
   return (false);
}

/*****************************************************************************/
/*
If the character string in 'String' is found in a comma-separated list of 
strings in 'List' then return, true else return false.  Example of the list of 
strings: "string1,string2,string3,etc".  Obviously a comma cannot be part of 
the searched for 'String'.  The match is case-insensitive.
*/ 

boolean InCommaList
(
char *String,
char *List
)
{
   register char  *lptr, *sptr;

   if (Debug) fprintf (stdout, "InCommaList() |%s|%s|\n", String, List);

   lptr = List;
   while (*lptr)
   {
      sptr = String;
      while (*sptr && *lptr != ',')
      {
         if (toupper(*sptr) != toupper(*lptr)) break;
         sptr++;
         lptr++;
      }
      if (!*sptr && (!*lptr || *lptr == ',')) return (true);
      while (*lptr && *lptr != ',') lptr++;
      if (*lptr) lptr++;
   }
   return (false);
}

/*****************************************************************************/
/*
Keeps track of all URLs obtained.  It checks the supplied source URL against 
this database and if already obtained returns true.  If not in the database it 
returns false. 

The database comprises a linked list of URLs.
*/ 

boolean AlreadyObtained (struct SourceUrlStruct *suptr)

{
   register struct NoteUrlStruct  *nuptr;
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AlreadyObtained() |%s|\n", suptr->Url);

   nuptr = ObtainedHead;
   while (nuptr != NULL)
   {
      if (Debug) fprintf (stdout, "old nuptr %d |%s|\n", nuptr, nuptr->Url);
      if (strsame (nuptr->Url, suptr->Url, -1))
         break;
      nuptr = nuptr->NextPtr;
   }
   if (nuptr == NULL)
      return (false);
   else
      return (true);
}

/*****************************************************************************/
/*
Update the list of obtained URLs.

The database comprises a linked list of URLs.
*/ 

UpdateObtained (struct SourceUrlStruct *suptr)

{
   register struct NoteUrlStruct  *nuptr;
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "UpdateObtained() |%s|\n", suptr->Url);

   nuptr = malloc (strlen(suptr->Url)+1 + sizeof(struct NoteUrlStruct*));
   if (nuptr == NULL)
   {
      status = vaxc$errno;
      fprintf (stdout, "%%%s-E-OBTAINED, database update failed\n-%s\n",
               Utility, SysGetMsg(status)+1);
      exit (status | STS$M_INHIB_MSG);
   }
   if (ObtainedHead == NULL)
      ObtainedHead = ObtainedTail = nuptr;
   else
   {
      ObtainedTail->NextPtr = nuptr;
      ObtainedTail = nuptr;
   }
   nuptr->NextPtr = NULL;
   strcpy (nuptr->Url, suptr->Url);
   if (Debug) fprintf (stdout, "new nuptr %d |%s|\n", nuptr, nuptr->Url);
}

/*****************************************************************************/
/*
Keeps track of all URLs obtained.  It checks the supplied source URL against 
this database and if already obtained returns true.  If not in the database it 
returns false. 

The database comprises a linked list of URLs.
*/ 

boolean DoNotGet (struct SourceUrlStruct *suptr)

{
   register struct NoteUrlStruct  *nuptr;
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DoNotGet() |%s|\n", suptr->Url);

   nuptr = DoNotGetHead;
   while (nuptr != NULL)
   {
      if (Debug) fprintf (stdout, "old nuptr %d |%s|\n", nuptr, nuptr->Url);
      if (strsame (nuptr->Url, suptr->Url, -1))
         break;
      nuptr = nuptr->NextPtr;
   }
   if (nuptr == NULL)
      return (false);
   else
      return (true);
}

/*****************************************************************************/
/*
Update the list of obtained URLs.

The database comprises a linked list of URLs.
*/ 

UpdateDoNotGet (struct SourceUrlStruct *suptr)

{
   register struct NoteUrlStruct  *nuptr;
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "UpdateDoNotGet() |%s|\n", suptr->Url);

   nuptr = malloc (strlen(suptr->Url)+1 + sizeof(struct NoteUrlStruct*));
   if (nuptr == NULL)
   {
      status = vaxc$errno;
      fprintf (stdout, "%%%s-E-OBTAINED, database update failed\n-%s\n",
               Utility, SysGetMsg(status)+1);
      exit (status | STS$M_INHIB_MSG);
   }
   if (DoNotGetHead == NULL)
      DoNotGetHead = DoNotGetTail = nuptr;
   else
   {
      DoNotGetTail->NextPtr = nuptr;
      DoNotGetTail = nuptr;
   }
   nuptr->NextPtr = NULL;
   strcpy (nuptr->Url, suptr->Url);
   if (Debug) fprintf (stdout, "new nuptr %d |%s|\n", nuptr, nuptr->Url);
}

/*****************************************************************************/
/*
Open the output file.  "Text" files (e.g. .HTML, .PS, .TXT) are opened as
STM_LF format files, binary files in UDF format.
*/ 

int CreateFile (struct SourceUrlStruct *suptr)

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CreateFile() |%s|\n", suptr->VmsFileName);

   /***************/
   /* create file */
   /***************/

   fprintf (stdout, "[%d.%d][Creating: %s]\n",
            DocumentRange, suptr->SubdocumentDepth, suptr->VmsFileName);

   if (ExerciseOnly) return (SS$_NORMAL);

   suptr->FileFab = cc$rms_fab;
   suptr->FileFab.fab$b_fac = FAB$M_PUT | FAB$M_GET | FAB$M_UPD | FAB$M_BRO;
   suptr->FileFab.fab$l_fna = suptr->VmsFileName;  
   suptr->FileFab.fab$b_fns = strlen(suptr->VmsFileName);
   /* deferred write performance option */
   suptr->FileFab.fab$l_fop = FAB$M_DFW | FAB$M_SQO;
   if (suptr->IsATextFile)
   {
      suptr->FileFab.fab$b_rat = FAB$M_CR;
      suptr->FileFab.fab$b_rfm = FAB$C_STMLF;
   }
   else
   {
      suptr->FileFab.fab$b_rat = 0;
      suptr->FileFab.fab$b_rfm = FAB$C_UDF;
   }

   if (VMSnok (status = sys$create (&suptr->FileFab, 0, 0)))
   {
      if (Debug) fprintf (stdout, "sys$create %%X%08.08X\n", status);
      return (status);
   }

   suptr->FileRab = cc$rms_rab;
   suptr->FileRab.rab$l_fab = &suptr->FileFab;
   /* 2 buffers, write behind performance option */
   suptr->FileRab.rab$b_mbf = 2;
   suptr->FileRab.rab$l_rop = RAB$M_WBH;

   if (VMSnok (status = sys$connect (&suptr->FileRab, 0, 0)))
   {
      if (Debug) fprintf (stdout, "sys$connect %%X%08.08X\n", status);
      sys$close (&suptr->FileFab, 0, 0);
      return (status);
   }

   return (status);
}

/*****************************************************************************/
/*
*/ 

StampHtmlFile (struct SourceUrlStruct *suptr)

{
   static int  ImportMessageLength;
   static $DESCRIPTOR (FromFaoDsc, "%!AZ-I-FROM, \"!AZ\"");
   static $DESCRIPTOR (ByFaoDsc, "%!AZ-I-BY, !AZ");
   static $DESCRIPTOR (DateFaoDsc, "%!AZ-I-DATE, !11%D");

   int  status;
   unsigned short  Length;
   char  Line [256];
   $DESCRIPTOR (LineDsc, Line);
   $DESCRIPTOR (ImportFaoDsc, ImportMessage);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "StampHtmlFile() |%s|\n", suptr->VmsFileName);

   if (ExerciseOnly) return;

   if (!suptr->IsAnHtmlFile) return;
 
   suptr->FileRab.rab$l_rbf = "<!--";
   suptr->FileRab.rab$w_rsz = 4;
   sys$put (&suptr->FileRab, 0, 0);

   suptr->FileRab.rab$l_rbf = Line;

   sys$fao (&FromFaoDsc, &Length, &LineDsc, Utility, suptr->Url);
   Line[suptr->FileRab.rab$w_rsz = Length] = '\0';
   sys$put (&suptr->FileRab, 0, 0);

   sys$fao (&ByFaoDsc, &Length, &LineDsc, Utility, SoftwareID);
   Line[suptr->FileRab.rab$w_rsz = Length] = '\0';
   sys$put (&suptr->FileRab, 0, 0);

   sys$fao (&DateFaoDsc, &Length, &LineDsc, Utility, 0);
   Line[suptr->FileRab.rab$w_rsz = Length] = '\0';
   sys$put (&suptr->FileRab, 0, 0);

   suptr->FileRab.rab$l_rbf = "-->";
   suptr->FileRab.rab$w_rsz = 3;
   sys$put (&suptr->FileRab, 0, 0);

   if (ImportMessage[0])
   {
      if (!ImportMessageLength) ImportMessageLength = strlen(ImportMessage);
      ImportFaoDsc.dsc$w_length = ImportMessageLength;
      if (VMSnok (status =
          sys$fao (&ImportFaoDsc, &Length, &LineDsc, ImportLink)))
         exit (status);
      suptr->FileRab.rab$l_rbf = Line;
      Line[suptr->FileRab.rab$w_rsz = Length] = '\0';
      sys$put (&suptr->FileRab, 0, 0);
   }

   /* blank line */
   suptr->FileRab.rab$l_rbf = "";
   suptr->FileRab.rab$w_rsz = 0;
   sys$put (&suptr->FileRab, 0, 0);
}

/*****************************************************************************/
/*
*/ 

int ConnectToServer (struct SourceUrlStruct *suptr)

{
   int  status;
   unsigned short  Length;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "ConnectToServer() |%s:%d|\n",
              suptr->HostName, suptr->ServerPort);

   /* get the remote host details */
   if (ProxyServer[0])
      suptr->RemoteHostEntryPtr = gethostbyname (ProxyServer);
   else
      suptr->RemoteHostEntryPtr = gethostbyname (suptr->HostName);
   if (suptr->RemoteHostEntryPtr == NULL)
   {
      if (Debug) fprintf (stdout, "gethostbyname() %%X%08.08X\n", IP_errno);
      return (IP_errno);
   }
   suptr->RemoteHostEntry = *suptr->RemoteHostEntryPtr;
   if (ProxyServer[0])
      strcpy (ProxyServer, suptr->RemoteHostEntry.h_name);
   else
      strcpy (suptr->HostName, suptr->RemoteHostEntry.h_name);
   suptr->SocketName.sin_family = suptr->RemoteHostEntry.h_addrtype;
   if (ProxyServer[0])
      suptr->SocketName.sin_port = htons (ProxyServerPort);
   else
      suptr->SocketName.sin_port = htons (suptr->ServerPort);
   suptr->SocketName.sin_addr =
      *((struct in_addr *)suptr->RemoteHostEntry.h_addr);

   if ((suptr->ClientSocket = socket (AF_INET, SOCK_STREAM, 0)) < 0)
   {
      if (Debug) fprintf (stdout, "socket() %%X%08.08X\n", IP_errno);
      return (IP_errno);
   }
   if (connect (suptr->ClientSocket,
                &suptr->SocketName,
                sizeof(suptr->SocketName)))
   {
      status = IP_errno;
      if (Debug) fprintf (stdout, "connect() %%X%08.08X\n", status);
      close (suptr->ClientSocket);
      return (status);
   }

#  ifdef UCX_TCPIP
      suptr->ClientChannel = vaxc$get_sdc (suptr->ClientSocket);
#  else
      /* multinet */
      suptr->ClientChannel = suptr->ClientSocket;
#  endif

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
*/
 
int DisconnectFromServer (struct SourceUrlStruct *suptr)

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DisconnectFromServer()\n");

   shutdown (suptr->ClientSocket, 2);
   IP_close_socket (suptr->ClientSocket);

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
*/ 
 
int ReadData
(
struct SourceUrlStruct *suptr,
char *DataBuffer,
int SizeOfDataBuffer,
int *DataLengthPtr
)
{
   int  status;
   struct {
      unsigned short  Status;
      unsigned short  Count;
      char *Buffer;
   } IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ReadData()\n");

   *DataLengthPtr = 0;
   DataBuffer[0] = '\0';

#  ifdef UCX_TCPIP
      status = sys$qiow (0, suptr->ClientChannel, IO$_READVBLK, &IOsb, 0, 0,
                         DataBuffer, SizeOfDataBuffer, 0, 0, 0, 0);
#  else
      /* multinet */
      status = sys$qiow (0, suptr->ClientChannel, IO$_RECEIVE, &IOsb, 0, 0,
                         DataBuffer, SizeOfDataBuffer, 0, 0, 0, 0);
#  endif

   if (Debug)
      fprintf (stdout, "sys$qiow() %%X%08.08X IOsb: %%X%08.08X Count: %d\n",
               status, IOsb.Status, IOsb.Count);
   if (VMSnok (status)) return (status);
   if (VMSnok (IOsb.Status)) return (IOsb.Status);

   DataBuffer[*DataLengthPtr = IOsb.Count] = '\0';
   if (Debug) fprintf (stdout, "recv() %d\n", *DataLengthPtr);

   return (status);
}

/*****************************************************************************/
/*
*/
 
int WriteData
(
struct SourceUrlStruct *suptr,
char *DataPtr,
int DataLength
)
{
   int  status;
   struct {
      unsigned short  Status;
      unsigned short  Count;
      char *Buffer;
   } IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "WriteData() %d |%s|\n", DataLength, DataPtr);

   if (DataLength < 0) DataLength = strlen(DataPtr);

#  ifdef UCX_TCPIP
      status = sys$qiow (0, suptr->ClientChannel, IO$_WRITEVBLK, &IOsb, 0, 0,
                         DataPtr, DataLength, 0, 0, 0, 0);
#  else
      /* multinet */
      status = sys$qiow (0, suptr->ClientChannel, IO$_SEND, &IOsb, 0, 0,
                         DataPtr, DataLength, 0, 0, 0, 0);
#  endif

   if (Debug)
      fprintf (stdout, "sys$qiow() %%X%08.08X IOsb: %%X%08.08X Count: %d\n",
               status, IOsb.Status, IOsb.Count);
   if (VMSok (status) && VMSnok (IOsb.Status)) status = IOsb.Status;
   if (VMSnok (status)) return (status);

   return (status);
}

/*****************************************************************************/
/*
*/
 
int StringOverflowError (int SourceLineNumber)
 
{
   fprintf (stdout, "%%%s-E-STRINGOVF, string overflow at line %d\n",
            Utility, SourceLineNumber);
   return (STS$K_ERROR | STS$M_INHIB_MSG);
}

/*****************************************************************************/
/*
*/
 
char* SysGetMsg (int StatusValue)
 
{
   static char  Message [256];
   short int  Length;
   $DESCRIPTOR (MessageDsc, Message);
 
   sys$getmsg (StatusValue, &Length, &MessageDsc, 0, 0);
   Message[Length] = '\0';
   if (Debug) fprintf (stdout, "SysGetMsg() |%s|\n", Message);
   return (Message);
}
 
/****************************************************************************/
/*
Does a case-insensitive, character-by-character string compare and returns 
true if two strings are the same, or false if not.  If a maximum number of 
characters are specified only those will be compared, if the entire strings 
should be compared then specify the number of characters as 0.
*/ 
 
boolean strsame
(
register char *sptr1,
register char *sptr2,
register int  count
)
{
   while (*sptr1 && *sptr2)
   {
      if (toupper (*sptr1++) != toupper (*sptr2++)) return (false);
      if (count)
         if (!--count) return (true);
   }
   if (*sptr1 || *sptr2)
      return (false);
   else
      return (true);
}
 
/****************************************************************************/

int ShowHelp ()

{
   fprintf (stdout,
"%%%s-I-HELP, Basic usage of the Spinneret Utility (%s)\n\
\n\
Defn: \"the body part used by an arachnid to extrude gossamer\"\n\
Spinneret allows the retrieval and local compilation of documents from the\n\
Internet WWW.  Files comprising the document(s) are created in the local\n\
directory.  URLs within the files are modified for that single directory.\n\
\n\
$ SPINNERET \"url\" [qualifiers]\n\
\n\
/[NO]ALL(d) /ALWAYS_GET=string /CONFIRM /DEPTH=integer /EXCLUDE=string\n\
/[NO]EXERCISE /[NO]EXTERNAL /HELP /IMPORT_LINK=string /IMPORT_MESSAGE=string\n\
/INCLUDE=string /PROXY=string /RANGE=integer /[NO(d)]VERBOSE\n\
\n\
$ SPINNERET \"http://host/file.type\"  !retrieve document with this URL\n\
$ SPINNERET \"url\" /ALL               !get all files from child directories\n\
$ SPINNERET \"url\" /ALWAYS=GIF,XBM    !always retrieve this file type(s)\n\
$ SPINNERET \"url\" /NOEXTERNAL        !do not retrieve from external hosts\n\
$ SPINNERET \"url\" /INCLUDE=HTML,HTM  !only retrieve this file type(s)\n\
$ SPINNERET \"url\" /PROXY=host        !domain.host[:port] (default 8080)\n\
$ SPINNERET \"url\" /RANGE=integer     !get other documents up to 'integer' away\n\
$ SPINNERET \"url\" /DEPTH=integer     !from child directories to 'integer' deep\n\
$ SPINNERET \"url\" /CONFIRM           !ask before retrieving any file\n\
\n",
   Utility, SoftwareID);
   return (SS$_NORMAL);
}
 
/****************************************************************************/
/*
This function allows images activated by a "foreign verb" to behave in a way 
that approximates the CLI$ (Command Line Interpreter) utility calls.  Get the 
entire command Line following the verb that activated the image.  The command 
Line is returned in uppercase, space compressed (i.e. maximum of one space 
between text elements, trimmed of leading and trailing spaces).  Returns a 
warning status if there were no parameters/qualifiers on the command Line.
The variable CommandLine is global.
*/ 
 
int ParseCommandLine ()
 
{
   int  status;
   unsigned short  Length;
   unsigned long  Flags = 0;
   struct dsc$descriptor_s 
          CommandLineDsc = { sizeof(CommandLine)-1, DSC$K_DTYPE_T,
                             DSC$K_CLASS_S, CommandLine };
 
   /* get the entire command Line following the verb */
   if (VMSnok (status = lib$get_foreign (&CommandLineDsc, 0, &Length, &Flags)))
      return (status);
   CommandLine[Length] = '\0';
 
   if (ParseCommand (CommandLine))
      return (SS$_NORMAL);
   else
      return (STS$K_ERROR | STS$M_INHIB_MSG);
}
 
/****************************************************************************/
/*
This function allows images activated by a "foreign verb" to behave in a way 
that approximates the CLI$ (Command Line Interpreter) utility calls.  Quoted 
strings are always indicated by being parsed to include a single leading 
quote.
*/ 
 
boolean ParseCommand (char *CommandLine)
 
{
   register int  QuoteCount = 0;
   register char  *cptr, *eptr;
   boolean  CommandLineOK = true;
   char  Entity [256] = "";
 
   /* set up any argument defaults */
   ParseCommandEntity (NULL);
 
   cptr = CommandLine;
   eptr = Entity;
 
   for (;;)
   {
      if (*cptr == '\"')
      {
         QuoteCount++;
         *eptr++ = *cptr++;
         continue;
      }
 
      if (QuoteCount & 1 && *cptr)
      {
         /* inside quoted text, copy all characters as literals */
         *eptr++ = *cptr++;
         continue;
      }
 
      if (*cptr == '/' || isspace (*cptr) || !*cptr)
      {
         if (isspace (*cptr))
         {
            /* span the white space */
            while (*cptr && isspace (*cptr)) cptr++;
            if (*cptr == '=')
            {
               /* part of a qualifier, continue to get the value */
               *eptr++ = *cptr++;
               /* span any intervening white space */
               while (*cptr && isspace (*cptr)) cptr++;
               continue;
            }
         }
 
         if (Entity[0])
         {
            *eptr = '\0';
            if (!ParseCommandEntity (Entity)) CommandLineOK = false;
         }
 
         /* if end of command Line then break from loop */
         if (!*cptr) break;
 
         /* start of new entity */
         eptr = Entity;
         /* if start of qualifier ensure slash is copied */
         if (*cptr == '/') *eptr++ = *cptr++;
 
         continue;
      }
 
      /* any other character, just copy, ensure upper case */
      *eptr++ = toupper(*cptr++);
   }
 
   return (CommandLineOK);
}
 
/*****************************************************************************/
/*
Get a string value from a qualifier, e.g. '/EXAMPLE=TEST'.
*/
 
boolean ParseCommandString
(
char *Entity,
char *String,
boolean Qualifier,
boolean ReportErrors,
boolean EnsureUpperCase
)
{
   register int  QuoteCount = 0;
   register char  *eptr, *sptr;
 
   if (Debug) fprintf (stdout, "ParseCommandString()\nEntity: '%s'\n", Entity);
 
   eptr = Entity;
 
   if (Qualifier)
   {
      /* scan down to equate symbol */
      while (*eptr && *eptr != '=') eptr++;
      if (*eptr) eptr++;
      if (!*eptr)
      {
         if (ReportErrors)
         {
            fprintf (stdout,
            "%%%s-E-VALREQ, missing qualifier or keyword value\n \\%s\\\n",
            Utility, Entity+1);
         }
         return (false);
      }
   }
 
   sptr = String;
   while (*eptr)
   {
      if (*eptr == '\"')
      {
         if (QuoteCount & 1)
         {
            /* are inside quotes, check for escaped quotes ("") */
            if (*++eptr != '\"')
            {
               /* now outside quotes */
               QuoteCount++;
            }
            /* drop thru to character copy */
         }
         else
         {
            /* now inside quotes */
            QuoteCount++;
            eptr++;
            continue;
         }
      }
 
      if (EnsureUpperCase)
         *sptr++ = toupper(*eptr++);
      else
         *sptr++ = *eptr++;
   }
   *sptr = '\0';
 
   if (Debug) fprintf (stdout, "String: '%s'\n", String);
 
   return (true);
}
 
/*****************************************************************************/
/*
Get an integer value from a qualifier, e.g. '/EXAMPLE=99'.
*/
 
boolean ParseCommandInteger
(
char *Entity,
int *IntegerPtr,
int Base,
boolean ReportErrors
)
{
   register char  *eptr;
   char  *sptr;
 
   if (Debug)
      fprintf (stdout, "ParseCommandInteger() '%s' Base: %d\n", Entity, Base);
 
   for (eptr = Entity; *eptr && *eptr != '='; eptr++);
   if (*eptr) eptr++;
   if (*eptr)
   {
      *IntegerPtr = strtol (eptr, &sptr, Base);
      if (sptr > eptr && !*sptr)
         return (true);
      else
      {
         if (ReportErrors)
         {
            fprintf (stdout,
            "%%%s-E-BADVALUE, '%s' is an invalid keyword value\n",
            Utility, eptr);
         }
         return (false);
      }
   }
   else
   {
      if (ReportErrors)
      {
         fprintf (stdout,
         "%%%s-E-VALREQ, missing qualifier or keyword value\n \\%s\\\n",
         Utility, Entity+1);
      }
      return (false);
   }
}
 
/*****************************************************************************/
/*
A single command Line "entity" has been parsed, check if its recognised.  This 
function is the one modified for the individual requirements of each program.
*/
 
boolean ParseCommandEntity (char *Entity)
 
{
   if (Entity == NULL)
   {
      /* set up any argument defaults */
      Debug = DoConfirm = DoShowHelp = GetExternalLinks = Verbose = false;
      AllRootSubdocuments = true;
      AlwaysGetTheseFileTypes[0] = ExcludeFileTypes[0] = ImportLink[0] =
         ImportMessage[0] = IncludeFileTypes[0] = ProxyServer[0] = '\0';
      Depth = Range = 1;
      return (true);
   }
 
   if (Debug) fprintf (stdout, "ParseCommandEntity() Entity: '%s'\n", Entity);
 
   if (Entity[0] == '/')
   {
      if (strsame (Entity, "/ALL", 4))
         return (AllRootSubdocuments = true);
      if (strsame (Entity, "/NOALL", 6))
      {
         AllRootSubdocuments = false;
         return (true);
      }

      if (strsame (Entity, "/ALWAYS_GET=", 4))
         return (ParseCommandString (Entity, AlwaysGetTheseFileTypes,
                                     true, true, true));

      if (strsame (Entity, "/CONFIRM", 4))
         return (DoConfirm = true);

      /* turns on all "if (Debug)" statements */
      if (strsame (Entity, "/DBUG", -1))
         return (Debug = true);

      if (strsame (Entity, "/DEPTH=", 4))
         return (ParseCommandInteger (Entity, &Depth, 10, true));

      if (strsame (Entity, "/EXCLUDE=", 4))
         return (ParseCommandString (Entity, ExcludeFileTypes,
                                     true, true, true));

      if (strsame (Entity, "/EXERCISE", 4))
         return (ExerciseOnly = true);
      if (strsame (Entity, "/NOEXERCISE", 6))
      {
         ExerciseOnly = false;
         return (true);
      }

      if (strsame (Entity, "/EXTERNAL", 4))
         return (GetExternalLinks = true);
      if (strsame (Entity, "/NOEXTERNAL", 6))
      {
         GetExternalLinks = false;
         return (true);
      }

      if (strsame (Entity, "/IMPORT_LINK=", 12))
         return (ParseCommandString (Entity, ImportLink, true, true, true));

      if (strsame (Entity, "/IMPORT_MESSAGE=", 12))
         return (ParseCommandString (Entity, ImportMessage,
                                     true, true, true));

      if (strsame (Entity, "/INCLUDE=", 4))
         return (ParseCommandString (Entity, IncludeFileTypes,
                                     true, true, true));

      if (strsame (Entity, "/HELP", 4))
         return (DoShowHelp = true);

      if (strsame (Entity, "/PROXY=", 4))
         return (ParseCommandString (Entity, ProxyServer, true, true, true));

      if (strsame (Entity, "/RANGE=", 4))
         return (ParseCommandInteger (Entity, &Range, 10, true));

      if (strsame (Entity, "/VERBOSE", 4))
         return (Verbose = true);
      if (strsame (Entity, "/NOVERBOSE", 6))
      {
         Verbose = false;
         return (true);
      }

      fprintf (stdout, "%%%s-E-IVQUAL, unrecognised qualifier\n \\%s\\\n",
               Utility, Entity+1);
      return (false);
   }
 
   if (!UrlParameter[0])
      return (ParseCommandString (Entity, UrlParameter, false, true, false));

   fprintf (stdout, "%%%s-E-MAXPARM, too many parameters\n \\%s\\\n",
            Utility, Entity);
   return (false);
}
   
/*****************************************************************************/

